%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% CosieClass.oz ; This file is part of the OMEGA system
%
%   major updates: 26.10.1999,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA Project                                                        
%   AG Siekmann/FB nformatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)
% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   %% The class of Cosie constraint solvers, Version 2.0
   %%      jzimmer 19.1.99 
import
   Property
   System(show:Show
%	  showInfo: ShowInfo
	  printInfo gcDo)
   Browser(browse: Browse)
   Space
   Search(one: SearchOne)
   Open
   Explorer
%  Compiler
   DaVinci at 'x-oz://contrib/DaVinci.ozf'
   
%   Graph   at 'x-ozlib://mathweb/cs/cosie/Graph.ozf'
   Printer('class': PrinterClass) at 'x-ozlib://mathweb/cs/share/Printer.ozf'
   Sets at 'x-ozlib://mathweb/cs/share/Sets.ozf'
   
   ContextTreeClass('class': ContextTree)
   at 'x-ozlib://mathweb/cs/share/ContextTreeClass.ozf'
   
   RIDomain(isConstraint:     IsConstraint
	    isPureConstraint: IsPureConstraint
	    isGround:         IsGround
	    solver:           DomainSolver
	    getConstants:     GetConstants
	    oz2PostDict:      Oz2PostDict
	    )
   at 'x-ozlib://mathweb/cs/cosie/RIDomain.ozf'

   Share(car:              Car
	 cdr:              Cdr
%	 setDifference:    SetDifference
	 removeConstraint: RemoveConstraint
	 applyRenaming:    ApplyRenaming
	 myMember:         MyMember
	)
   at 'x-ozlib://mathweb/cs/share/Share.ozf'

   Translate
   at 'x-ozlib://mathweb/cs/share/Translate.ozf'
   

export
   'class': Cosie
   


define
   Version = 2.2     %% separated assumptions; new search procedure; search trace
   {Property.put 'print.depth' 100}
   {Property.put 'print.width' 100}
   
   fun{RenameSymbolicConsts CStores Renamings}
      SymbolicConsts = {Map CStores fun{$ CS}
				       node(consts: {Flatten {Map CS.context GetConstants}}
					    cstore: CS)
				    end}
      PairSets = {Sets.pairSets SymbolicConsts}
      %CommonConsts = {Sets.mapIntersection SymbolicConsts}
   in
      {Show pairSets(PairSets)}
      {Show symbConsts(SymbolicConsts)}
      Renamings = {Map CStores fun{$ _} {NewDictionary} end}
      
%       if CommonConsts == nil then
% 	 CStores
%       else


      {ForAll PairSets proc{$ [S1 S2]}
			  case S1
			  of I1#node(consts: C1 ...) then
			     case S2
			     of I2#node(consts: C2 ...) then
				CommonConsts = {Sets.intersection C1 C2}
			     in
				{Show commonConsts(CommonConsts)}
				{ForAll CommonConsts
				 proc{$ Const}
				    {Dictionary.put
				     {Nth Renamings I2}
				     Const
				     {VirtualString.toAtom Const#'Renamed'#I1}}
				 end}
			     else skip end
			  else skip end
		       end}
      
      {Show {Map Renamings Dictionary.entries}}
      {List.mapInd CStores fun{$ I CStore}
			      {Show store(CStore)}
			      cstore(constraints: {ApplyRenaming {List.nth Renamings I} CStore.constraints}
				     assumptions: {ApplyRenaming {List.nth Renamings I} CStore.assumptions}
				     context:     {ApplyRenaming {List.nth Renamings I} CStore.context}
				    )
			   end}
   end


   
   fun{List2VS Xs}
      case Xs
      of nil then ""
      [] X|Xr then
	 X#{List2VS Xr}
      end
   end
   
   fun{ReReName Cons Renamings}
      Naming = {NewDictionary}
   in
      {ForAll Renamings proc{$ Ren}
			   {ForAll {Dictionary.entries Ren}
			    proc{$ Old#New}
			       {Dictionary.put Naming New Old}
			    end}
			end}
      {Show name({Dictionary.entries Naming})}
      {ApplyRenaming Naming Cons}
   end

   fun{MergeLists S1 S2}
      %{Show merge_lists(S1 S2)}
      case S1
      of nil then S2
      [] S|Sr then
	 if {MyMember S S2} then
	    {MergeLists Sr S2}
	 else
	    S|{MergeLists Sr S2}
	 end
      else S1 end
   end
   
   fun{MergeConstraintStores CStores Context Assumptions Constraints}
      case CStores
      of nil then
	 cstore(context: Context assumptions: Assumptions constraints: Constraints)
      [] CS|Cr then
	 {MergeConstraintStores Cr
	  {MergeLists Context CS.context}
	  {MergeLists Assumptions CS.assumptions}
	  {MergeLists Constraints CS.constraints}
	 }
      else
	 cstore(context: Context assumptions: Assumptions constraints: Constraints)
      end
   end
   
   fun{InsertCons Hyps Ass Cons ConLists}
      case ConLists
      of nil then [cstore(context: Hyps assumptions: Ass constraints: Cons)]
      [] cstore(context: Hs assumptions: As constraints: Cs)|Lr then
	 if {Sets.equal Hs Hyps} then
	    cstore(context: Hyps assumptions: Ass constraints: Cons)|Lr
	 else
	    cstore(context: Hs assumptions: As constraints: Cs)|{InsertCons Hyps Ass Cons Lr}
	 end
      else [cstore(context: Hyps assumptions: Ass constraints: Cons)] end
   end

   
   fun{GetConstraintsIn Lines EVars}
      {Flatten {Map Lines fun{$ Line}
			     case Line
			     of _#F then
				if {IsConstraint F EVars} then
				   [Line]
				else nil end
			     else nil end
			  end}}
   end
   
   fun{FilterHyps CStore EVars}
      case CStore
      of cstore(context: Hyps assumptions: Ass constraints: Cons) then
	 cstore(context: {GetConstraintsIn Hyps EVars}
		assumptions: Ass
		constraints: Cons)
      else CStore end
   end
   
	 
%===============================================
   %%===============================================
   %%  The class of all Constraint solvers for inequalities and equations over the reals.
   %%                   --         -           -                -
   %%===============================================
   /**The class of all Constraint Solvers for Inequalities and Equations over the reals: */

   class Cosie from BaseObject

      prop
	 locking
	 
      attr
	 contextTree: nil      %% the context tree

	 normalizedCons: nil
	 detVars:        nil   %% the list of determined variables
	 
	 assumptions:    nil   %% all assumption constraints
	 constraints:    nil   %% all goal constraints
	 entailed:       nil   %% all entailment checks

	 metaConstraints: nil  %% the meta constraints (e.g. not-in(x D) )
	 eigenVariables:  nil  %% the Eigenvariable conditions

	 searchTrace: nil       %% the trace of the search procedure
	 messages:    nil       %% all important messages received
	 problemName: 'UNKNOWN' %% the name of the planning problem (for the term printer)
	 vinci:       nil       %% the DaVinci object showing the context tree
	 app:         nil       %% the XApplication (for output)
	 traceFile:   nil       %% the traceFile if open
	 printer:     nil       %% the TermPrinter
	 

	 %%-----------------
	 %% initialize the constraint solver with
	 %% an application manager and a trader URL.
	 %%-----------------
      meth init(App TraderURL)
	 app <- App
	 {@app.show 'CoSIE-service version '#Version#'requested from '#TraderURL}
      end


      %%---------------
      %% shutdown of the CoSIE service (e.g. close windows...)
      %%---------------
      meth terminate
	 Cosie, unshowConstraints(_)
      end
      
      %%-----------------
      %% reset the Solver with hyps Hyps
      %%-----------------
      /** resets CoSIE for planning a new problem */
			 %% @1=ProblemName
      meth reset(ProblemName $)
	 contextTree <- {New ContextTree init}  %% open a new context tree

	 normalizedCons <- nil
	 
	 assumptions <- nil           %% no assumption constraints
	 constraints <- nil           %% no goal constraints
	 entailed    <- nil           %% no entailment checks
	 metaConstraints <- nil       %% no meta-constraints
	 eigenVariables <- nil        %% no Eigenvariables 
	 
	 searchTrace <- nil           %% no search trace
	 messages <- nil              %% no messages

	 detVars <- nil
         
	 Cosie, stopTrace(_)           %% close the trace file if open
	 
	 problemName <- ProblemName   %% store the problem name 
	 
	 if @printer \= nil then      %% reset the TermPrinter if it's initialized.
	    Title = 'CoSIE constraint store(s) for '#ProblemName
	 in
	    {@printer newTitle(Title)} 
	    {@printer clear}
	 else skip end
	 
	 {@app.showInfo 'CoSIE has been reseted for planning '#ProblemName}
	 'T'
      end
      
      %%-----------------
      %% browse the object
      %%-----------------
      meth browse($)
	 {@contextTree browse}
	 'T'
      end
      
      %%-----------------
      %% Show the object
      %%-----------------
      meth show($)
	 {Show 'constrains:'#@constraints}
	 {Show 'assumptions:'#@assumptions}
	 {Show 'entailed:'#@entailed}
	 {@contextTree show}
	 'T'
      end
      
      %%-----------------
      %% Open a constraint printer window
      %%-----------------
      /** opens a constraint printer window */
      meth showConstraints($)
	 Title = 'CoSIE constraint store(s) for '#@problemName
      in
 	 if @printer \= nil then
	    {@printer close}
	 else skip end
	 
	 printer <- {New PrinterClass init(Title)}
	 
	 if @contextTree \= nil then
	    Cosie, PrintConstraints
	 else skip end
	 
	 'T'
      end

      %%-----------------
      %% close the constraint printer
      %%-----------------
      /** closes the constraint printer window */
      meth unshowConstraints($)
	 if @printer \= nil then
	    {@printer close}
	    printer <- nil
	 else skip end
	 'T'
      end
      
      %%-----------------
      %% opens the DaVinci browser.
      %%-----------------
      /** opens a daVinci browser for the ContextTree */
      meth showContextTree($)
	 if @vinci == nil then %% Stream braucht man im Moment nicht
	    vinci <- {New DaVinci.daVinciClass init(_)}
	    {@vinci send(nil)}
	    Cosie, daVinci
	    'T'
	 else 'NIL' end
      end
      
      %%-----------------
      %% close the DaVinci browser.
      %%-----------------
      /** closes a daVinci browser for the ContextTree */
      meth unshowContextTree($)
	 if @vinci \= nil then
	    {@vinci close}
	    vinci <- nil
	    'T'
	 else 'NIL' end
      end
	 
      %%-----------------
      %% sends a graph representation of the context tree to
      %% daVinci
      %%-----------------
      /** explicitly shows the context tree in the daVinci browser */
      meth daVinci
	 if @contextTree \= nil then
	    VS = {@contextTree daVinci($)}
	 in
%	    {Browse vs#VS}
	    if @vinci \= nil then
	       {@vinci graph(VS)}
	    else skip end
	 else skip end
      end

      %%-----------------
      %% start tracing the messsages in an oz file with the given name.
      %%-----------------
      /** starts tracing certain messages in a file FileName.oz that can be
        compiled and run (helpful for debugging without proof planning) */
				      %% @1=FileName
      meth startTrace(FileName $)
	 if @traceFile == nil then
	    {Show 'startTrace'#FileName}
	    traceFile <- {New Open.file init(name:FileName flags:[write create])}
	    {@traceFile write(vs:"functor\nimport\n")}
	    {@traceFile write(vs:"   Application(exit)
   Property
   System(showInfo: ShowInfo)
   CosieClass(\'class\': Cosie) at \'x-ozlib://mathweb/cs/cosie/CosieClass.ozf\'
   XApplication at \'x-ozlib://mathweb/share/XApplication.ozf\'\n")}
	    {@traceFile write(vs:"define\n   local Cos={New Cosie init({New XApplication.manager init} 'Lim.oz')} Result Sol\n   in\n")}
	    {@traceFile write(vs:"      {Cos reset('"#@problemName#"' _)}\n\n")}
	    {@traceFile write(vs:"      %{Cos showConstraints(_)}\n\n")}
            {@traceFile write(vs:"      %{Property.put 'debug.out' true}\n")}
            {@traceFile write(vs:"      %{Property.put 'debug.props' true}\n\n")}
	    'T'
	 else
	    'NIL'
	 end
      end

   
      %%-----------------
      %% stop tracing messages.
      %%-----------------
      /** stops tracing messages */
      meth stopTrace($)
	 if @traceFile == nil then
	    'T'
	 else 
	    {@traceFile write(vs:"      %{Cos unshowConstraints(_)}\n")}
	    {@traceFile write(vs:"      %{Cos findSolution(Sol)}\n")}
	    {@traceFile write(vs:"      %{ShowInfo Sol}\n")}
	    {@traceFile write(vs:"   end\n   %{Application.exit 0}\nend")}
	    {@traceFile close}
	    traceFile <- nil
	    'T'
	 end
      end
      
      /** private method: write Trace information into trace file*/
      meth Trace(VS)
	 if @traceFile \= nil then
	    {@traceFile write(vs:VS)}
	 else skip end
      end
      
      %%-----------------
      %% computes the reflection of the constraint store, i.e. the
      %% constraints collected from all context nodes.
      %%-----------------
      /** gives a reflection of the current constraint stores, i.e. a conjunction of implications that represents 
the constraints collected from all context nodes.
*/
      meth reflect($)
	 CStores = {@contextTree getAllCStores($)} % note: a constraint store has the
	                                                    % form  cstore(context: [...] constraints:[..])
	 Implications = {Map CStores fun{$ CStore}
					ContextConstraints = {Map
							      {GetConstraintsIn CStore.context @eigenVariables}
							      Cdr}
					Constraints = {Map CStore.constraints Cdr}
				     in
					if ContextConstraints == nil then
					   Constraints
					else
					   '->'(ContextConstraints Constraints)
					end
				     end}
      in
	 {Show Implications}
	 {@contextTree show}
	 {Translate.list2PostConjunct Implications false Oz2PostDict}
      end
      
      %%-----------------
      %% gives a POST representation of the list of Eigenvariables accumulated so far.
      %%-----------------
      /** gives a list of the current Eigenvariables */
      meth eigenVars($)
	 B = {Translate.list2Post {Map @eigenVariables Car} false Oz2PostDict}
      in
	 B
	 %'TRUE'
      end
      
      %%-----------------
      %% gives a string representation of an association list with for the determined
      %% variables. The list has the form: ((Var1 . Term1) ... (Varn . Termn))
      %%-----------------
      /** give a list of the variables currently determined by any constraint store*/
      meth getDetVars($)      
	 if @contextTree == nil then
	    {@app.show 'The constraint-solver has not been initialized yet.'}
	    'NIL'
	 else
            {@app.showInfo 'detvars: '#@detVars}
	    if @detVars == nil then
               'NIL'
            else
               @detVars
            end
	 end
      end

      %%-----------------
      %% gives the trace of the search procedure of CoSIE, i.e. a list of messages of the form
      %% (Theorem (premises) (conclusions))
      %%-----------------
      /** gives the trace of the search procedure */
      meth getSearchTrace($)   
	 if @searchTrace == nil then 
	    {@app.show 'There is no search trace.'}
	    'NIL'
	 else
	    @searchTrace
	 end
      end

      
      %%-----------------
      %% Searches for a solution (instantiations) for the problem variables.
      %% Returns a post representation of the solution, if found,
      %% 'NIL' otherwise.
      %%-----------------
      /** tries to find a solution for the accumulated constraints */
      meth findSolution(explore: Explore <= false $)
	 %% First, collect all constraint stores from the leaf nodes of context tree
	 %%
	 AllCStores = {@contextTree getAllCStores($)} % note: a constraint store has the
	                                                   %  form cstore(context: [...]
	                                                   %              assumptions: [..]
	                                                   %              constraints: [..])
	 RightCStores = {Map AllCStores fun{$ Store}
					   {FilterHyps Store @eigenVariables}
					end}
	 
	 Renamings
  	             % then rename the symbolic constants occuring in different contexts.
	 RenamedCStores = {RenameSymbolicConsts RightCStores Renamings}
	             % finally, merge all constraint stores to one list of constraints.
	 SearchStore = {MergeConstraintStores RenamedCStores nil nil nil}	 
	 
	 Assumptions = {Append SearchStore.context SearchStore.assumptions}
	 Constraints = SearchStore.constraints
	 
	             % compute the unary search procedure for the root computation space
	 SearchProcedure = proc{$ Sol}
			      SymbolicSolver = {New DomainSolver init}
			   in
			      {ForAll Assumptions
			       proc{$ C}                       %% first propagate all constraints...
				  case C
				  of nil then skip
				  [] _#Con then
				     {Show tellAssumption(Con)}
				     {SymbolicSolver tellAssumption(Con)}
				     %{System.gcDo}				     
				  else skip end
			       end}
			      {ForAll Constraints
			       proc{$ C}                       %% first propagate all constraints...
				  case C
				  of nil then skip
				  [] _#Con then
				     {Show tellConstraint(Con)}
				     {SymbolicSolver tellConstraint(Con)}
				     %{System.gcDo}				     
				  else skip end
			       end}
			      {SymbolicSolver normalizeCons}
			      
			      {SymbolicSolver setEigenVars(@eigenVariables)} %% propagate the EigenVariable conditions...
			      {SymbolicSolver distribute(Sol)}               %% ...then search for a solution
			   end
	 Result
	 Solution
	 SearchTrace
	 
      in
	 %{Show all_stores(AllCStores)}
	 {Show renamed_stores(RenamedCStores)}
	 %{Show search_store(SearchStore)}
	                                              % initialize the explorer object
	 {Explorer.object delete(information all)}
	 
	 {Explorer.object add(information
			      proc {$ I S}
				 if {IsDet S} then {Browse {Value.type S}} else skip end
				 {Browse I#S}
			      end
			      label: 'Constraint Store'
			     )}
	 
	 	 
	 {@app.show 'Trying to find a solution for the problem variables!'}
	 
	 if Explore then
	    {Explorer.object script(SearchProcedure)}
	 else 
	    %skip
	    {SearchOne.depth SearchProcedure 1 _ Result} %kein KillP
	 end

	 if Result == nil then
	    {@app.show 'Sorry, could not find a solution :-( '}
	    'NIL'
	 else
	    [[Solution SearchTrace]] = Result
                                          % translate the formulae in the search trace to POST syntax
	    searchTrace <- {Translate.trace2Post SearchTrace Oz2PostDict}
	    
	    if {IsDet Solution} then      % really found a solution, i.e., a list of equations?
	                             % then re-rename the constants in the formulae
	       ReRenamedSolution = {ReReName Solution Renamings}
	                             % convert the solution into an assoctiation list that can be read by lisp
	       SolutionAList = {Map ReRenamedSolution fun{$ Equ}
							 case Equ
						      of '='(Var Term ...) then
							 Var#Term
						      else
							 a#b
						      end
						      end}
	    in
	       if SolutionAList == nil then
                  detVars <- nil
               else
                  detVars <- {VirtualString.toString "("#{Translate.aList2Post SolutionAList Oz2PostDict}#")"}
               end
	       {@app.show ReRenamedSolution}
	    %{@app.showInfo @detVars}
	  %  {@app.showInfo @searchTrace}
	       {Translate.list2PostConjunct ReRenamedSolution false Oz2PostDict}
	    else 'NIL' end
	 end
      end
      
   
      
     
      %%-----------------
      %% tells a constraint to all context nodes wrt. Hyps
      %% fails, iff all telling attempts fail
      %%-----------------
      /** tells a meta-constraint Con (e.g. a Eigenvariable-condition) to the constraint solver */
       %% @1=Con
      meth tellMetaConstraint(Con $)
	 case {Label Con} 
	 of nil then 'NIL'
	 [] 'not-occurs' then
	    metaConstraints <- Con|@metaConstraints
	    ConstName#const = Con.1       %% the Eigenvariable...
	    VarName#var = Con.2           %% the meta-variable...
	 in
	    if {Member ConstName#VarName @eigenVariables} then skip
	    else eigenVariables <- ConstName#VarName|@eigenVariables end
	    %{Show @eigenVariables}
	    Cosie, Trace("      {Cos tellMetaConstraint("#{Value.toVirtualString Con 20 100}#" _)}\n")
	    'T'
	 [] 'no-Eigenvars' then
	    {ForAll @eigenVariables proc{$ EVName#_}
				       {Show EVName}
				       Cosie, tellMetaConstraint('not-occurs'(EVName#const Con.1) _)
				    end}
	    Cosie,Trace("      {Cos tellMetaConstraint("#{Value.toVirtualString Con 20 100}#" _)}\n")
	    'T'
	 else
	    'T'
	 end
      end

      %%-----------------
      %% tries to tell a constraint to all context nodes wrt. Hyps
      %% fails, iff all telling attempts fail
      %%-----------------
      /** tests whether a constraint from a (proof planning) assumption with name NodeName is consistent with all constraints with an subcontext of Hyps. Creates a new computation space with a new symbolic constraint solver and starts the constraint inference in the new space. Returns 'T', if no inconsistency could be detected, 'NIL' otherwise */
	%% @1=NodeName @2=Hyps @3=Constraint			      
      meth testAssumption(NodeName Hyps Con $)
	 if {IsConstraint Con @eigenVariables} then
	    if {Member NodeName#Con @assumptions} orelse
	       {Member NodeName#Con @constraints} then
	       'NIL'
	    else 
	       if @contextTree == nil then
		  {@app.show 'The constraint-solver has not been initialized yet.'}
		  'NIL'
	       else
		  HypConstraints = {Map {GetConstraintsIn Hyps @eigenVariables} Cdr}  %% the constraints in the Hyps
		  OldAssumptions = {@contextTree getAssumptionsFor(Hyps $)} %the assumptions that belong to Hyps
		  OldConstraints = {@contextTree getConstraintsFor(Hyps $)} %the constraints that belong to Hyps
		  TellProcedure = proc{$ Sol}
				     SymbolicSolver = {New DomainSolver init}
				  in
				     {ForAll {Append HypConstraints OldAssumptions}
				      proc{$ A}                       %% first propagate all present constraints...
					 {SymbolicSolver tellAssumption(A)}
				      end}
				     
				     {ForAll OldConstraints
				      proc{$ C}                       %% first propagate all present constraints...
					 {SymbolicSolver tellConstraint(C)}
				      end}
				     
				     {SymbolicSolver tellAssumption(Con)}   %% ... then the new constraint Con
				     {SymbolicSolver normalizeCons}         %% simplify the constraint store
				  end
		  
		  NewCompSpace = {Space.new TellProcedure}
	       in
		  if {Space.ask NewCompSpace} == failed then
		     %{@app.show 'Test for constraint assumption '#Con#' failed!!!'}
		     'NIL'
		  else
		     %{@app.show 'Test for constraint assumption '#Con#' succeeded!!!'}
		     'T'
		  end %e
	       end % if contextTree
	    end % if {Member}
	 else
	    {Show Con#'is not a valid constraint assumption.'}
	    'NIL'
	 end %i
      end %m

      
      %%-----------------
      %% tells a constraint that is part of a proof assumption to all context nodes
      %% fails, iff all telling attempts fail
      %%-----------------
      /** similar to the method testAssumption. In the case of consistency, the Constraint is added as a new assumption constraint to the contextnode with context Hyps */
      %% @1=NodeName @2=Hyps @3=Constraint	
      meth tellAssumption(NodeName Hyps Con $)
	 if {IsConstraint Con @eigenVariables} then 
	    if {Member NodeName#Con @assumptions} orelse
	       {Member NodeName#Con @constraints} then
	       'NIL'
	    else
	       %{Show constraint_has_not_been_here_before}
	       if @contextTree == nil then
		  {@app.show 'The constraint-solver has not been initialized yet.'}
		  'NIL'
	       else
		  HypConstraints = {Map {GetConstraintsIn Hyps @eigenVariables} Cdr} %% the constraints in the Hyps
		  OldAssumptions = {@contextTree getAssumptionsFor(Hyps $)} %the assumptions that belong to Hyps
		  OldConstraints = {@contextTree getConstraintsFor(Hyps $)} %the constraints that belong to Hyps
		  TellProcedure = proc{$ Sol}
				     SymbolicSolver = {New DomainSolver init}
				  in
				     {ForAll {Append HypConstraints OldAssumptions}
				      proc{$ A}                       %% first propagate all present constraints...
					 {SymbolicSolver tellAssumption(A)}
				      end}
				     
				     {ForAll OldConstraints
				      proc{$ C}                       %% first propagate all present constraints...
					 {SymbolicSolver tellConstraint(C)}
				      end}
				     
				     {SymbolicSolver tellAssumption(Con)}  %% ... then the new constraint Con
				     {SymbolicSolver normalizeCons}        %% simplify the constraint store
				     
				     Sol = SymbolicSolver
				  end
		  NewCompSpace = {Space.new TellProcedure}
	       in
		  %{Show old#OldConstraints}
		  %{Show new#Con}
		  if {Space.ask NewCompSpace} == failed then 
		     {@app.show 'Telling assumption '#Con#' failed!!!'}
		     'NIL'
		  else 
		     {@app.show 'Telling assumption '#Con#' succeeded!!!'}
		     %{System.gcDo}            %% do a garbage collection to avoid the oz bug
		     
		     {@contextTree insertAssumption(NodeName Hyps Con _)}
		     assumptions <- (NodeName#Con)|@assumptions
		     
		     Cosie, UpdateNormalizedCons(Hyps NewCompSpace)
		     Cosie, Trace("      {Cos tellAssumption('"#NodeName#"' "
				  #{Value.toVirtualString Hyps 20 100}#" "
				  #{Value.toVirtualString Con 20 100}#" _)}\n")
		     %{@contextTree show}
		     'T'
		  end %e
	       end % if contextTree
	    end % if {Member}
	 else
%{Show Con#'is not a valid constraint.'}
	    'NIL'
	 end %i
      end %m


      %%-----------------
      %% tries to tell a constraint to all context nodes wrt. Hyps
      %% fails, iff all telling attempts fail
      %%-----------------
      /** tests whether a constraint from a planning goal with name NodeName is consistent with all constraints with an subcontext of Hyps. Creates a new computation space with a new symbolic constraint solver and starts the constraint inference in the new space. Returns true, if no inconsistency could be detected */
      %% @1=NodeName @2=Hyps @3=Constraint
      meth testConstraint(NodeName Hyps Con $)
	 if {IsConstraint Con @eigenVariables}
	    andthen {Not {IsGround Con}}
	 then
	    if {Member NodeName#Con @constraints} then
	       'NIL'
	    else 
	       if @contextTree == nil then
		  {@app.show 'The constraint-solver has not been initialized yet.'}
		  'NIL'
	       else
		  HypConstraints = {Map {GetConstraintsIn Hyps @eigenVariables} Cdr}  %% the constraints in the Hyps
		  OldAssumptions = {@contextTree getAssumptionsFor(Hyps $)} %the assumptions that belong to Hyps
		  OldConstraints = {@contextTree getConstraintsFor(Hyps $)} %the constraints that belong to Hyps
		  TellProcedure = proc{$ Sol}
				     SymbolicSolver = {New DomainSolver init}
				  in
				     {ForAll {Append HypConstraints OldAssumptions}
				      proc{$ A}                       %% first propagate all present constraints...
					 {SymbolicSolver tellAssumption(A)}
				      end}
				     
				     {ForAll OldConstraints
				      proc{$ C}                       %% first propagate all present constraints...
					 {SymbolicSolver tellConstraint(C)}
				      end}
				     
				     {SymbolicSolver tellConstraint(Con)}   %% ... then the new constraint Con
				     {SymbolicSolver normalizeCons}        %% simplify the constraint store
				  end
		  
		  NewCompSpace = {Space.new TellProcedure}
	       in
		  if {Space.ask NewCompSpace} == failed then
		     %Show 'Test for constraint '#Con#' failed!!!'}
		     'NIL'
		  else
		     %Show 'Test for constraint '#Con#' succeeded!!!'}
		     'T'
		  end %e
	       end % if contextTree
	    end % if {Member}
	 else
	    %{Show Con#'is not a valid constraint.'}
	    'NIL'
	 end %i
      end %m

      %%-----------------
      %% tells a constraint to all context nodes wrt. Hyps
      %% fails, iff all telling attempts fail
      %%-----------------
      /** similar to the method testConstraint. In the case of consistency, the Constraint is added as a new goal constraint to the contextnode with context Hyps */
      %% @1=NodeName @2=Hyps @3=Constraint	

      meth tellConstraint(NodeName Hyps Con $)
	 if {IsConstraint Con @eigenVariables}
	    andthen {Not {IsGround Con}} then
	    if {Member NodeName#Con @constraints} then
	       'NIL'
	    else
	       %{Show constraint_has_not_been_here_before}
	       if @contextTree == nil then
		  {@app.show 'The constraint-solver has not been initialized yet.'}
		  'NIL'
	       else
		  HypConstraints = {Map {GetConstraintsIn Hyps @eigenVariables} Cdr} %% the constraints in the Hyps
		  OldAssumptions = {@contextTree getAssumptionsFor(Hyps $)} %the assumptions that belong to Hyps
		  OldConstraints = {@contextTree getConstraintsFor(Hyps $)} %the constraints that belong to Hyps
		  TellProcedure = proc{$ Sol}
				     SymbolicSolver = {New DomainSolver init}
				  in
				     {ForAll {Append HypConstraints OldAssumptions}
				      proc{$ A}                       %% first propagate all present constraints...
					 {SymbolicSolver tellAssumption(A)}
				      end}
				     %{System.gcDo}            %% do a garbage collection to avoid the oz bug				     
				     {ForAll OldConstraints
				      proc{$ C}                       %% first propagate all present constraints...
					 {SymbolicSolver tellConstraint(C)}
				      end}
				     %{System.gcDo}            %% do a garbage collection to avoid the oz bug				     
				     {SymbolicSolver tellConstraint(Con)}   %% ... then the new constraint Con
				     {SymbolicSolver normalizeCons}        %% simplify the constraint store
				     {SymbolicSolver showProps}
				     Sol = SymbolicSolver
				  end
		  
		  NewCompSpace = {Space.new TellProcedure}
	       in
		  %{Show old#OldConstraints}
		  %{Show new#Con}
		  if {Space.ask NewCompSpace} == failed then 
		     {@app.show 'Telling constraint '#Con#' failed!!!'}
		     'NIL'
		  else 
		     {@app.show 'Telling constraint '#Con#' succeeded!!!'}
		     %{System.gcDo}            %% do a garbage collection to avoid the oz bug
		     
		     {@contextTree insertConstraint(NodeName Hyps Con _)}
		     
		     constraints <- (NodeName#Con)|@constraints
		     
		     Cosie, UpdateNormalizedCons(Hyps NewCompSpace)
		     Cosie, Trace("      {Cos tellConstraint('"#NodeName#"' "
				  #{Value.toVirtualString Hyps 20 100}#" "
				 #{Value.toVirtualString Con 20 100}#" _)}\n")
		     %{@contextTree show}
		     'T'
		  end %e
	       end % if contextTree
	    end % if {Member}
	 else
	    {Show Con#'is not a valid constraint.'}
	    'NIL'
	 end %i
      end %m

      
      %%-----------------
      %% asks whether the constraint Con is entailed in at least
      %% one context node node wrt. Hyps
      %%-----------------
      /** asks whether the given Constraint follows from all constraints with context Hyps. Returns 'T', if the Constraint follows, 'NIL' otherwise.*/
      %% @1=NodeName @2=Hyps @3=Constraint	
      meth askConstraint(NodeName Hyps Con $)
	 if {IsPureConstraint Con @eigenVariables} then 
	    if @contextTree == nil then
	       {@app.show 'The constraint-solver has not been initialized yet.'}
	       'NIL'
	    else
	       HypConstraints = {Map {GetConstraintsIn Hyps @eigenVariables} Cdr}  %% the constraints in the Hyps
	       Assumptions = {@contextTree getAssumptionsFor(Hyps $)} %the constraints that belong to Hyps
	       TellProcedure = proc{$ Sol}
				  SymbolicSolver = {New DomainSolver init}
			       in
				  {ForAll {Append HypConstraints Assumptions}
				   proc{$ C}                       %% first propagate all assumption constraints...
				      %{Show tellAssumption(C)}
				      {SymbolicSolver tellConstraint(C)}
				   end}
				  
				  {SymbolicSolver askConstraint(Con)}   %% ... then ask for entailment of Con
			       end
	       
	       NewCompSpace = {Space.new TellProcedure}
	    in
	       %{Show assumptions(Assumptions)}
	       %{Show ask(Con)}
	       if {Space.ask NewCompSpace} == failed then
		  {@app.show 'The constraint '#Con#' is NOT entailed!!!'}
		  'NIL'
	       else
		  if {Member NodeName#Con @entailed} then skip
		  else 
		     entailed <- (NodeName#Con)|@entailed
		     Cosie, Trace("      {Cos askConstraint('"#NodeName#"' "
				  #{Value.toVirtualString Hyps 20 100}#" "
				  #{Value.toVirtualString Con 20 100}#" _)}\n")
		  end
		  {@app.show 'The constraint '#Con#' is entailed!'}
		  'T'
	       end %e
	    end
	 else
	    %{Show Con#'is not a pure constraint, so it cannot be entailed!'}
	    'NIL'
	 end %i
      end %m
      

      /** private method: removes the constraint from the proof planning node with name NodeName from the contexttree. */
      meth withdrawConstraint(NodeName $)
	 if @contextTree == nil then
	    {@app.show 'The constraint-solver has not been initialized yet.'}
	    'NIL'
	 else
	    {@contextTree withdrawConstraint(NodeName)}

	    assumptions <- {RemoveConstraint NodeName @assumptions}
	    constraints <- {RemoveConstraint NodeName @constraints}
	    entailed <- {RemoveConstraint NodeName @entailed}
	    'T'
	 end
      end


      /** private method: extracts the normalized constraint set from a symbolic solver in the given computation space CompSpace. */
													    
      meth UpdateNormalizedCons(Hyps CompSpace)
	 Solver = {Space.merge CompSpace}
	 DetVars = {Solver getDetVars($)}
      in
	 if DetVars == nil then
            detVars <- nil
         else
            detVars <- {VirtualString.toString "("#{Translate.aList2Post DetVars Oz2PostDict}#")"}
         end
	 thread
	    {Thread.setThisPriority low}
	    lock
	       HypForms = {Map {GetConstraintsIn Hyps @eigenVariables} Cdr}
	       Ass = {Solver getAssumptions($)}
	       Cons = {Solver getConstraints($)}
	    in
	       normalizedCons <- {InsertCons HypForms Ass Cons @normalizedCons}
	       
	       Cosie, PrintConstraints
	    end
	 end
      end

	    
      %% gives three lists of strings  [Contexts] [Assumptions formulas] [Goal Formulas]
      %%-----------------
      /** delivers the string representation of the constraint stores for the term printer of LOUI */
      meth getConstraints($)
	 Result = {Map @normalizedCons
		   fun{$ CStore}
		      cstore(context: {Map CStore.context fun{$ Formula}
							     {VirtualString.toString
							      {Translate.oz2Post Formula true Oz2PostDict}}
							  end}
			     assumptions: {Map CStore.assumptions fun{$ Con}
								     {VirtualString.toString
								      {Translate.oz2Post Con true Oz2PostDict}}
								  end}
			     goals: {Map CStore.constraints fun{$ Con}
							       {VirtualString.toString
								{Translate.oz2Post Con true Oz2PostDict}}
							    end}
			    )
		   end}
% 	 Compiled = {Compiler.virtualStringToValue {Value.toVirtualString Result 1000 1000}}

      in
	 %{Browse vstrint({Value.toVirtualString	 Result 1000 1000})}
	 %{Browse compiled(Compiled)}
	 {Value.toVirtualString	Result 1000 1000}
      end
	     
	 %% prints all constraint stores to a text window
      %%-----------------
      /** private method: prints out all constraint stores of the context tree to the constraint printer window*/
      meth PrintConstraints
	 if @printer \= nil then
	    %{Show stores(@normalizedCons)}
	    {@printer clear}
	    {ForAll @normalizedCons
	     proc{$ CStore}
		Context = {Map CStore.context fun{$ Formula}
						 {VirtualString.toString
						  {Translate.oz2Post Formula true Oz2PostDict}}
					      end}
		AssStrings = {Map CStore.assumptions fun{$ Con}
							{VirtualString.toString
							 {Translate.oz2Post Con true Oz2PostDict}}
						     end}
		ConStrings = {Map CStore.constraints fun{$ Con}
							{VirtualString.toString
							 {Translate.oz2Post Con true Oz2PostDict}}
						     end}
	     in
		{@printer nextLine(_)}
		{@printer print("Context [")}
		{ForAll Context proc{$ C}
				   {@printer printTerm(C)}
				   {@printer print(", ")}
				end}
		{@printer print("]")}
		{@printer nextLine(_)}
		{@printer print("==============")}
		{@printer nextLine(_)}
		if AssStrings \= nil then
		   {@printer print("Assumptions:")}		
		   {@printer nextLine(_)}
		   {ForAll AssStrings
		    proc{$ S}
		       {@printer print("  ")}
		    {@printer printTerm(S)}
		       {@printer nextLine(_)}
		    end}
		else skip end
		if ConStrings \= nil then
		   {@printer print("Constraints:")}
		   {@printer nextLine(_)}
		   {ForAll ConStrings
		    proc{$ S}
		       {@printer print("  ")}
		       {@printer printTerm(S)}
		       {@printer nextLine(_)}
		    end}
		else skip end
	     end}  % proc
	 end
      end
      
      %%-----------------
      %% This method is called by the servicemanager if
      %% the user tries to call an unknown method.
      %%-----------------
      meth otherwise(M $)
	 {@app.show 'Sorry, unknown method for Cosie'#M}
	 'NIL'
      end
      
   end %c
end %define

/*
proc {ExploreSpaces Session Problem Engine WeightFn}
   
   proc {SearchAux S}
      GenSym Consts Vars Formulas Logic
   in
      {ReadProblem Problem GenSym Consts Vars Logic Formulas}
      S = {New Engine create(GenSym Consts Vars Formulas Logic)}
      {S prove}
   end

   Logic
in
   %% get logic again from problem
   if {HasFeature Problem logic}
   then Logic = Problem.logic
   else Logic = classical
   end

   {Explorer.object delete(information all)}

   
   {Explorer.object add(information
			proc {$ _ S}
			   if {HasFeature S kernel}
			   then   {Browse S.kernel}
			   else
			      {Browse 'no kernel computed'}
			   end
			end
			label: 'Kernel'
		       )}

   {Explorer.object add(information
			proc {$ _ S}
			   {PrettyPrintModel S Problem
			    fun {$ _}
			       true
			    end}
			end
			label: 'Herbrand Base'
		       )}

   {Explorer.object add(information
			proc {$ _ S}
			   {PrettyPrintModel S Problem
			    fun {$ _#v(var: Val price: _)}
			       {Not {IsDet Val}} orelse Val > 0
			    end}
			end
			label: 'Positive Part'
		       )}

   {Explorer.object script(SearchAux WeightFn)}
end

proc{ShowStream S}
   case S
   of N#V|Ss then
      if {IsAtom N} then R in
	 if {IsDet V} then
	    {System.ShowInfo " "#N#": "#V}
	 else B in
	    {System.ShowInfo " "#N#": ["#{RI.getLowerBound V}#" ; "#{RI.getUpperBound V}#"]"}
	 end
%	    {System.ShowInfo {AtomToString {RI.isConstant V}}}
      else skip
      end
      {Show Stream Ss}
   [] nil then
      skip
   end
end

*/
