%%
%% DaVinci interface (graph visualization)
%%

%% usage: _ = {New DaVinciGraph init(G)}
%%
%% where  G = graph(nodes:[a b c d]
%%                  edges:[a#d b#d c#d])
%%
%% or     G = children(a:[d] b:[d] c:[d] d:nil)
%%

functor
   
import
   DaVinci(daVinciClass: DaVinciClass) at 'x-oz://contrib/DaVinci.ozf'
   
export
   'class': DaVinciGraph
   
prepare
   
   local
      proc {DictAdd D Key X}
         L = {Dictionary.get D Key} in
         {Dictionary.put D Key X|L}
      end
   in
      proc {MakeChildrenDict Nodes Edges ?Children}
         Children = {NewDictionary}
         {ForAll Nodes proc {$ N} {Dictionary.put Children N nil} end}
         {ForAll Edges proc {$ From#To} {DictAdd Children From To} end}
      end
      fun {MakeChildrenRec Nodes Edges}
         {Dictionary.toRecord children {MakeChildrenDict Nodes Edges}}
      end
   end
   local
      fun {MakeEdge Target}
         e('' nil r(Target))
      end
   in
      fun {MakeNode ID Text Children}
         l(ID n('' [a('OBJECT' Text)] {Map Children MakeEdge}))   
      end
   end
   fun {MakeGraph ChildrenRec}
      {Map {Record.toListInd ChildrenRec}
       fun {$ Node#Children}
          {MakeNode Node Node Children}
       end}
   end

define
   
   class DaVinciManager from DaVinciClass
                           
      /*
      attr
         DaVinci: false
         lastdata: 'none'  
         lastnode: 'none'
         lastedge: 'none'
         
      /** create a pipe to the API of DaVinci. */
      /** start a new DaVinci window.          */
      meth open
         case @DaVinci of false then
            DAVINCIHOME = {OS.getEnv 'DAVINCIHOME'} in
            DaVinci <- {New Open.pipe init(cmd:  DAVINCIHOME#'/daVinci'
                                           args: ['-pipe'])}
            %%{Browse DAVINCIHOME#'/daVinci'}
         else skip end
      end
      
      /** close a DaVinci object created by init */
      meth close
         case @DaVinci of false then skip
         else
            DaVinciManager, command(menu "file(close)")
            {@DaVinci close}
            DaVinci <- false
         end
      end
      */
      
      /** send a command to DaVinci                             */
      /** Input: Kind  = what kind of command (menu,graph etc.) */
      /**         Cmd  = the command itself as a string         */
      meth sendCommand(Kind Cmd)
         DaVinciClass, sendVS(Kind#'('#Cmd#')')
      end
      
      /** add a node "Name" to the current graph */
      meth addNode(Name)
         Cmd = graph(update([new_node(Name "Node" [a("OBJECT" Name)])] nil)) in
         DaVinciManager, send(Cmd)
      end
      
      /** add an edge "Name" from "Node1" to "Node2" to the current graph */
      meth addEdge(Name Node1 Node2)
         Cmd = graph(update(nil [new_edge(Name "Edge" nil Node1 Node2)])) in 
         DaVinciManager, send(Cmd)
      end
      
      /*
      %%
      meth node_selected(?Node)
         Node=@lastnode
      end
      
      /** read the output of the DaVinci API in a loop */
      meth readloop
         lastdata <- {ReadDV @DaVinci}
         case @lastdata of "quit\n" then
            lastnode <- 'none'
            DaVinci  <- false
            %%{Browse "DaVinci wurde beendet."}
         [] "node_double_click\n" then
            %%{Browse "Doppelklick auf Knoten "#@LastNode#"!"}
            skip
         [] "edge_double_click\n" then
            %%{Browse "Doppelklick auf Kante "#@LastEdge#"!"}
            DaVinciManager,readloop
         [] &n|&o|&d|&e|&_|&s|&e|&l|&e|&c|&t|&i|&o|&n|&s|&_|&l|&a|&b|&e|&l|&s|&(|R then
            case R of "[])\n" then
               %%{Browse "Klick auf leeres Feld !"}
               DaVinciManager,readloop
            [] &[|&"|R2 then
               lastnode <- {Reverse {List.drop {Reverse R2} 4}}
               %%{Browse  "Der Knoten "#@LastNode#" wurde ausgewhlt!"}
               DaVinciManager,readloop
            else %%{Browse "Fehler!"}
               DaVinciManager,readloop
            end
         [] &e|&d|&g|&e|&_|&s|&e|&l|&e|&c|&t|&i|&o|&n|&_|&l|&a|&b|&e|&l|&(|&"|R then
            lastedge <- {Reverse {List.drop {Reverse R} 3}} 
            %%{Browse "Die Kante "#@LastEdge#" wurde ausgewhlt!"}
            DaVinciManager,readloop
         else % {Browse @LastData}
            DaVinciManager,readloop
         end
      end
      */   
   end
   
   class DaVinciGraph from DaVinciManager
                         
      meth init(Listener)
	 S in
	 DaVinciClass, init(S)
	 thread
	    {ForAll S proc {$ Msg}
			 {Listener Msg}
			 if Msg==quit then
			    {Thread.terminate {Thread.this}}
			 end
		      end}
	 end
      end

      meth show(GraphSpec)
         case GraphSpec
         of graph(nodes:Nodes edges:Edges) then
            G = {MakeGraph {MakeChildrenRec Nodes Edges}} in
            DaVinciManager, send(graph(new(G)))
         [] children(...) then
            G = {MakeGraph GraphSpec} in
	    DaVinciManager, send(graph(new(G)))
         end
      end
   end
   
end





