%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SolverClass.oz ; This file is part of the OMEGA system %
%   major updates: 11.02.2000,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA project                                                        
%   AG Siekmann/FB Informatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)
% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   %% The class of symbolic constraint solvers
   %%      jzimmer 18.1.99 

import
   Browser(browse: Browse)
   System(show:     Show
	  showInfo: ShowInfo
	  eq
%	  gcDo:     GC
	 )
   Property
   
   Reflect at 'x-oz://contrib/Reflect.ozf'
   FD
   RI      at 'x-oz://contrib/RI.ozf'
   FCP_RI  at 'x-oz://contrib/modRI.fcp.ozf'
   
   RISimpRules(constrSimpRules: SimpRules
	       termRewriteRules: RewriteRules)
   at 'x-ozlib://mathweb/cs/cosie/RISimpRules.ozf'
   
%Maple(simplify: MapleSimplify lineq2Maple:Lineq2Maple)
   
   RIDomain(funcSymbols:         CosieFuncs
	    relSymbols:          CosieRels
	    getVar:              GetVar              %i
	    getConst:            GetConst            %i
	    getConstants:        GetConstants
	    %getVariables:        GetVariables
	    getEigenVars:        GetEigenVars
	    getLowerBounds:      GetLowerBounds
	    getUpperBounds:      GetUpperBounds
	    getEquations:        GetEquations
	    getVarString:        GetVarString
	    isGroundTerm:        IsGroundTerm
	 ) 
   at 'x-ozlib://mathweb/cs/cosie/RIDomain.ozf'

   Share(
%         car:                 Car
	 myMember:            MyMember
	 propMember:          PropMember
	 setDifference:       SetDifference
	 getSolution:         GetSolution
	 constructTerm:       ConstructTerm
	 completeCons:        CompleteCons
	 extractConstraints:  ExtractConstraints

	 unify:               Unify
	 matchList:           MatchList
	 applyMatching:       ApplyMatching
	 removeProps:         RemoveProps         % independent
%	 removeVar:           RemoveVar           % inde
	 findName:            FindName
	 findVar:             FindVar
	 killProps:           KillProps
	 applySubst:          ApplySubst
	 debug:               Debug	)
   at 'x-ozlib://mathweb/cs/share/Share.ozf'
   
   Sets at 'x-ozlib://mathweb/cs/share/Sets.ozf'
   
export
   'class': SymbolicSolver   %The class of all symbolic constraint-solvers
   

define
   fun{EqnListToSubst Eqns Subst}
      case Eqns
      of nil then Subst
      [] E|Er then
	 case E of '='(VarName#var Term) then
	    {Dictionary.put Subst VarName Term}
	    {EqnListToSubst Er Subst}
	 else Subst end
      else Subst end
   end
   
   fun{VarValue Bounds Val}    %% heuristic to determine the best undetermined Variable
      case Bounds
      of nil then Val
      [] '<='(term: T ...)|Br then
	 if {IsGroundTerm T nil} then
	    {VarValue Br (30 + Val)}
	 else
	    {VarValue Br (10 + Val)}
	 end
      [] '>='(term: T ...)|Br then
	 if {IsGroundTerm T nil} then
	    {VarValue Br (30 + Val)}
	 else
	    {VarValue Br (10 + Val)}
	 end
      [] '<'(term: T ...)|Br then
	 if {IsGroundTerm T nil} then
	    {VarValue Br (30 + Val)}
	 else
	    {VarValue Br (5 + Val)}
	 end
      [] '>'(term: T ...)|Br then
	 if {IsGroundTerm T nil} then
	    {VarValue Br (30 + Val)}
	 else
	    {VarValue Br (5 + Val)}
	 end
      else 0 end
   end

   {Wait Reflect.propReflect}  %% ... for propagator_reference_id :-(
   
   %%===============================================
   %%===============================================
   %%  The SolverClass
   %%===============================================
   /** The class of all symbolic constraint solvers. The Symbolic Solver combines symbolic and numeric (propagation based) constraint inference. It should be created in a subordinate computation space. */
   class SymbolicSolver
      from BaseObject
      prop
	 locking

      feat
	 App
	 
      attr
	 probVars           %% problem variables given by the proof planner
	 undetVars          %% names of the still symbolically undetermined variables
	 nestedVars         %% nested variables introduced during propagation
	 alienVars          %% new variables representing alien terms (SOLEX...)
	 probConsts         %% problem constants given by the proof planner

	 eigenVars          %% the Eigenvariable condtitions for the search procedure
	 alienReplacements  %% a dictionary with the alien term replacements
	 %solution          %% 
	 
	 assProps           %% list with the assumption constraint propagators
	 conProps           %% list with the goal constraint propagators
	 termProps          %% list with the term propagators

	 assumptions        %% assumption constraints.... terms
	 constraints        %% constraints.... terms

	 appliedRewriteRules %% the term rewrite rules that have been applied
	 appliedSimpRules    %% the constraint simplification rules that have been applied
	 
	 searchTrace
	 newAssumptions
	 
	 nameCounter        %% a counter for new variable names
	 

      %%====================================================
      %%                 Public Methods 
      %%====================================================
	 
      %%------------------
      %% Initialize the constraint solver.
      %%------------------
      meth init
	 probVars <- nil
	 undetVars <- nil
	 nestedVars <- nil
	 alienVars <- nil
	 probConsts <- nil
	 eigenVars <- nil
	 
	 alienReplacements <- {Dictionary.new}

	 assProps <- nil
	 conProps <- nil
	 termProps <- nil

	 appliedRewriteRules <- nil
	 appliedSimpRules <- nil

	 searchTrace <- nil
	 newAssumptions <- nil
	 
	 nameCounter <- 0
	 
	 constraints <- con('<': nil '<=': nil
			    '=': nil           
			    '>=': nil '>': nil
			    'IN': nil)
	 
	 assumptions <- ass('<': nil '<=': nil
			    '=': nil           
			    '>=': nil '>': nil
			    'IN': nil)
%{ShowInfo 'Hello, I`m a symbolic solver'}
      end
      
      
      %%------------------
      %% Propagate the constraint Con and simplify the constraint store.
      %%------------------
      /** starts propagation for a new constraint Con from a planning goal, adds Con to the symbolic constraint store and simplifies the store. */
      meth tellConstraint(Con)
	 if {MyMember Con @constraints.{Label Con}} then
	    skip
	 else
	    NewCon = SymbolicSolver, Propagate(Con $)  %% get Con annotated with propagator references
	 in
	    SymbolicSolver, AddConstraint(NewCon)
	    SymbolicSolver, SimplifyCons
	    %SymbolicSolver, normalizeCons
	    %SymbolicSolver, showProps
	 end
      end
      
      %%------------------
      %% Propagate the constraint Con
      %%------------------
      /** starts propagation for a new constraint Con from a proof assumption and adds Con to the symbolic constraint store */
      meth tellAssumption(Con)
	 if {Member Con @assumptions.{Label Con}} then
	    skip
	 else
	    NewCon = SymbolicSolver, PropagateAss(Con $)
	 in
	    SymbolicSolver, AddAssumption(NewCon)
	 end
      end
      
      %%------------------
      %% Check for entailment of the constraint Con.
      %%------------------
      /** tests whether the constraint Con follows from the symbolic constraint store. */
      meth askConstraint(Con)
	 SymbolicSolver, Ask(Con)
      end
      
      %%------------------
      %% Set the Eigenvariable conditions
      %%------------------
      /** sets the sets of Eigenvariables to the given set EVars */
      meth setEigenVars(EVars)
	 eigenVars <- EVars
      end
      
      %%------------------
      %% Get the state of the local computation space, the solver is sited in.
      %%------------------
      /** delivers the state of the symbolic constraint store. Returns 'succeeded', iff all problem variables are determined. */
      meth ask($)
	 if @undetVars == nil then
	    succeeded
	 else
	    undetermined
	 end
      end

      %%------------------
      %% Try to compute a solution for the problem variables.
      %%------------------
      /** extracts as solution from a succeeded symbolic constraint store during search. If all problem variables are determined, the method returns a list of equations, i.e. a list [ '='(V1 T1).... '='(Vn Tn)], for the problem variables V1,...,Vn, otherwise the method failes. Failure is used by the standard oz-engine to determine failed branches in the search */
      meth getSolution($)
	 if @undetVars == nil then
	    SymbolicSolver, showProps
	    VarNames = {Map @probVars fun{$ V}            %% determine the names of the problem variables
					 Name#_ = V
				      in
					 Name#var
				      end}
	    Props = {Map {Append {Append @assProps @conProps} @termProps}
		     fun{$ P}                             %% choose all propagators
			{Record.subtract P ref}
		     end}
	    Solution =  {VirtualString.toString           %% compute the list of equations
			 {GetSolution
			  VarNames
			  Props
			  Props
			  @alienReplacements
			  CosieFuncs}}
	    Subst = {EqnListToSubst Solution {NewDictionary}}  %% transfrom it into a substitution for the instantiation of the search trace
	 in
	    if {All Solution fun{$ Sol}                  %% are all instances ground terms??
				case Sol
				of '='(Name#var Term) then
				   EVars = {GetEigenVars Name @eigenVars}
				in
				   {IsGroundTerm Term EVars}
				end
			     end}
	    then
	       SymbolicSolver, InstantiateTrace(Subst)   %% instantiate the search trace
	       {Show Solution}
	       [Solution @searchTrace]
	    else
	       {Show 'No ground solution found...'}
	       SymbolicSolver, InstantiateTrace(Subst)   %% instantiate the search trace
	       [Solution @searchTrace]
	    end
	 else fail end
      end
      
      
      %%------------------
      %% Simplify the list of propagators by applying constrain rules, if possible.
      %%------------------
      /** normalizes the symbolic constraint solver by applying term rewrite- and constraint simplification rules */
      meth normalizeCons
	 SymbolicSolver, RewriteTerms
	 SymbolicSolver, SimplifyCons
      end
      
      %%------------------
      %% Gives an association list with the instantiations for the
      %% determined variables. The list has the form ((Var1 . Term1) ... (Varn . Termn))
      %%------------------
      /** gives an association list with the instantiations for the
         determined variables. The list has the form [(Var1 . Term1) ... (Varn . Termn)] */
      meth getDetVars($)
	 DetVars = {SetDifference {Map @probVars fun{$ Name#_} Name end} @undetVars}
	 AllProps = {Append {Append @assProps @conProps} @termProps}
      in
	 
	 {Map DetVars fun{$ Name}
			 Equations = {GetEquations Name#var {Append @assProps @conProps}}
		      in
			 case Equations
			 of nil then nil
			 [] '='(term: T)|_ then
			    Res = {ConstructTerm T AllProps AllProps 
				   @alienReplacements CosieFuncs}
			 in
			    Name#Res
			 else nil end
		      end}
      end

      %%------------------
      %% Show all problem variables.
      %%------------------
      /** shows all problem variables */
      meth showVars
	 {Show 'Variables'}
	 {ForAll @probVars proc{$ Name#RIVar}
			      VS = {GetVarString Name#RIVar}
			   in
			      {ShowInfo VS}
			   end}
      end
      
      %%------------------
      %% Show all propagators.
      %%------------------
      /** shows all propagators */
      meth showProps
	 if {Property.condGet 'debug.props' false} then 
	    {Show 'Assumption Propagators'}
	    {ForAll @assProps proc{$ P}
				 %SymbProp = SymbolicSolver, SymbolicArgs(P 1 $)
			      %in
				 {Show '  prop:'#P}
%			      {Show '  prop:'#{Record.subtract SymbProp ref}}
			      end}
	    {Show 'Constraint Propagators'}
	    {ForAll @conProps proc{$ P}
				% SymbProp = SymbolicSolver, SymbolicArgs(P 1 $)
			      %in
				 {Show '  prop:'#P}
%			      {Show '  prop:'#{Record.subtract SymbProp ref}}
			      end}
	    {Show 'Term Propagators'}
	    {ForAll @termProps proc{$ P}
				%  SymbProp = SymbolicSolver, SymbolicArgs(P 1 $)
			       %in
				  {Show '  prop:'#P}
%			      {Show '  prop:'#{Record.subtract SymbProp ref}}
			       end}
	    {Show 'Trace'}
	    {ForAll @searchTrace Show}
	    %{ForAll @newAssumptions Show}
	    {Show 'appliedRewrites'#@appliedRewriteRules}	 
	    {Show 'appliedSimpRules '#@appliedSimpRules}
	 else skip end
      end

      %%------------------
      %% Reflect the current constraint store.
      %%------------------
      /** reflects the current constraint store */
      meth reflect($)
	 SymbolicSolver, showProps
	 'true'
      end
      
      
      %%------------------
      %% Gives a list of the constraints assumptions accumulated so far.
      %%------------------
      /** returns an ordered list of constraint assumptions */
      meth getAssumptions($)
	 Ordering = fun{$ C1 C2}
		       ({Label C2} == '=')
		       orelse
		       ({Label C1} == '<')
		       orelse
		       ({Label C1} == '>')
		    end
      in
	 {Sort {ExtractConstraints @assProps @termProps @alienReplacements CosieFuncs}
	  Ordering}
      end

      %%------------------
      %% Gives a list of the constraints accumulated so far.
      %%------------------
      /** returns an ordered list of goal constraints */
      meth getConstraints($)
	 Ordering = fun{$ C1 C2}
		       ({Label C2} == '=')
		       orelse
		       ({Label C1} == '<')
		       orelse
		       ({Label C1} == '>')
		    end
      in
	 {Sort {ExtractConstraints @conProps @termProps @alienReplacements CosieFuncs}
	  Ordering}
      end
      

      %%------------------
      %% The distribution strategy for RI-constraints
      %%------------------
      /** this huge method implements the distribution strategy used during the search to explore the search tree. It distinguishes 9 different cases. Have a  look at the code for further details */
      meth distribute(Solution exclude:ExVars <= nil)
	 SymbolicSolver, showProps
	 {Show  '================================='}
	 {Show  distribution(@undetVars)}
	 {Show  '---------------------------------'}
	 choice
	    Vars = {SetDifference @undetVars ExVars}
	 in
	    %{Show vars(Vars)}
	    %{Show {Sets.intersection ['a' 'b' 'M'] ['D' 'M' 'd']}}
	    if Vars == nil then                                %% no variables left to determine, so extract a solution
	       if ExVars == nil then
		  SymbolicSolver, getSolution(Solution)
	       else
		  SymbolicSolver, distribute(Solution)
	       end
	    else
	       SortedVars = SymbolicSolver, SortVars(Vars $)   %% sort the undetermined problem variables with a special heuristic.
	       Var = case SortedVars of V|_ then V else nil end  %% choose the first variable Var == Name(lbounds: ... ubounds: ...)
	    in
	       if Var == nil then
		  SymbolicSolver, getSolution(Solution)
	       else 
		  LBounds = Var.lbounds                       %% get the lower and upper bounds of Var ...
		  UBounds = Var.ubounds
		  VarTerm = {Label Var}#var
		  EigenVars = {GetEigenVars {Label Var} @eigenVars}  %% ... and the Eigenvariables of Var
	       in
		  {Show chosen_var(Var)}
		  {Show eigen_vars(EigenVars)}
		  case LBounds
		  %-----------------------------------------------------
		  % case I:  there are NO LOWER bounds
		  %-----------------------------------------------------
		  of nil then
		     {Show case1}
		     case UBounds                        
		     of nil then                                 %% case I.1: there aren't any bounds to the variable
			SymbolicSolver, tellConstraint('='(VarTerm 1.0))   %% so determine it to 1.0
			SymbolicSolver, Substitute(VarTerm by:1.0)
			SymbolicSolver, distribute(Solution)
			
		     [] '<='(term: U ref: URef)|_ then           %% case I.2: there is one closed upper bound, then try
			{Show case1_1}
			if {Sets.intersection {GetConstants U} EigenVars} \= nil then
			   {Show eigenVar}
			   fail
			else
			   choice                                %%   Var <= U)  ==>   Var = U or Var < U
			      conProps <- {RemoveProps @conProps [URef]}
			      SymbolicSolver, tellConstraint( '='(VarTerm U) )
			      SymbolicSolver, Substitute(VarTerm)
% 			      SymbolicSolver, UpdateTrace(['eq_is_leq'(lhs: ['='(VarTerm U)]
% 								       rhs: ['<='(VarTerm U)])
% 							  ])
			      SymbolicSolver, distribute(Solution)
			   []
			      conProps <- {RemoveProps @conProps [URef]}
			      SymbolicSolver, tellConstraint('<'(VarTerm U) )
			      SymbolicSolver, normalizeCons
			      SymbolicSolver, UpdateTrace(['less-is-leq'(lhs: ['<'(VarTerm U)]
									 rhs: ['<='(VarTerm U)]
									 args: [VarTerm U])
							  ])
			      SymbolicSolver, distribute(Solution)
			   end
			end
			
		     [] '<'(term: U ref: URef)|_ then            %% case I.3: there are ONLY _OPEN_ upper bounds
			{Show case1_2}
			if {Sets.intersection {GetConstants U} EigenVars} \= nil then
			   {Show eigenVar}
			   fail
			else
			   case U
			   of _#alien then 
			      choice
				 conProps <- {RemoveProps @conProps [URef]}
				 SymbolicSolver, tellConstraint('='(VarTerm '-'(U 1.0)))
				 SymbolicSolver, Substitute(VarTerm)
				 SymbolicSolver, UpdateTrace(['minus-one-is-less'(
								 rhs: ['<'('-'(U 1.0) U)]
								 args: ['-'(U 1.0) U])
							     ])
				 SymbolicSolver, distribute(Solution)
			      []
				 conProps <- {RemoveProps @conProps [URef]}
				 SymbolicSolver, tellConstraint('<'(VarTerm '-'(U 1.0)))
				 SymbolicSolver, normalizeCons
				 SymbolicSolver, UpdateTrace(['less-transitivity'(
								 lhs: ['<'(VarTerm '-'(U 1.0)) '<'('-'(U 1.0) U)
								      ]
								 rhs: ['<'(VarTerm U)]
								 args: [VarTerm '-'(U 1.0) U])
							     ])
				 SymbolicSolver, distribute(Solution)
			      end
			   else 
			      choice
				 conProps <- {RemoveProps @conProps [URef]}
				 SymbolicSolver, Ask('>'(U 0.0))
				 SymbolicSolver, tellConstraint('='(VarTerm '/'(U 2.0)))
				 SymbolicSolver, Substitute(VarTerm)
				 SymbolicSolver, distribute(Solution)
			      [] 
				 conProps <- {RemoveProps @conProps [URef]}
				 SymbolicSolver, Ask('<'(U 0.0))
				 SymbolicSolver, tellConstraint('='(VarTerm '*'(2.0 U)))
				 SymbolicSolver, Substitute(VarTerm)
				 SymbolicSolver, distribute(Solution)
			      []
				 conProps <- {RemoveProps @conProps [URef]}
				 SymbolicSolver, Ask('='(U 0.0))
				 SymbolicSolver, tellConstraint('='(VarTerm ~1.0))
				 SymbolicSolver, Substitute(VarTerm by: ~1.0)
				 SymbolicSolver, distribute(Solution)
			      []
				 conProps <- {RemoveProps @conProps [URef]}
				 SymbolicSolver, tellConstraint('='(VarTerm '/'(U 2.0)))
				 SymbolicSolver, Substitute(VarTerm)
				 SymbolicSolver, distribute(Solution)
				 
			      end
			   end
			end
			
		     else
			{Show 'The List of upper bounds is neither empty, nor anything else :-('}
			fail
		     end
		     
		  %-----------------------------------------------------
		  %--  case II: THERE ARE LOWER BOUNDS....            
		  %-----------------------------------------------------
		  
		  [] '>='(term: L ref: Ref)|_ then      %%  first try a closed lower bound, if it exits
		     {Show case2}
		     if {Sets.intersection {GetConstants L} EigenVars} \= nil then
			{Show eigenVar}
			fail
		     else
			choice                              %%  L <= Var <= U)  ==>   L = Var or L < Var
			   conProps <- {RemoveProps @conProps [Ref]}
			   SymbolicSolver, tellConstraint( '='(VarTerm L) )
			   SymbolicSolver, Substitute(VarTerm)
			   SymbolicSolver, distribute(Solution)
			[]
			   conProps <- {RemoveProps @conProps [Ref]}
			   SymbolicSolver, tellConstraint( '>'(VarTerm L) )
			   SymbolicSolver, normalizeCons
			   SymbolicSolver, distribute(Solution)
			end
		     end
		  
		  [] '>'(term: L ref: LRef)|_ then      %% case II.1: there are only open lower bounds left...
		     {Show case3}
		     case UBounds                        
		     of nil then                          %% ... and no upper bounds
			{Show case3_1}
			if {Sets.intersection {GetConstants L} EigenVars} \= nil then
			   {Show eigenVar}
			   fail
			else
			   case L
			   of _#alien then 
			      choice
				 conProps <- {RemoveProps @conProps [LRef]}  % set Var to lower bound + 1
				 SymbolicSolver, tellConstraint('='(VarTerm '+'(L 1.0)))
				 SymbolicSolver, Substitute(VarTerm)
				 SymbolicSolver, UpdateTrace(['plus-one-is-greater'(
								 rhs: ['<'(L '+'(L 1.0))]
								 args: [L])
							     ])
				 SymbolicSolver, distribute(Solution)
			      []
				 conProps <- {RemoveProps @conProps [LRef]}  % Var is greater than lower bound + 1
				 SymbolicSolver, tellConstraint('>'(VarTerm '+'(L 1.0)))
				 SymbolicSolver, normalizeCons
				 SymbolicSolver, UpdateTrace(['less-transitivity'(
								 lhs: ['<'(L '+'(L 1.0)) '<'('+'(L 1.0) VarTerm)
								      ]
								 rhs: ['<'(L VarTerm)]
								 args: [L '+'(L 1.0) VarTerm])
							     ])
				 SymbolicSolver, distribute(Solution)
			      end
			   else 
			      {Show here}
			      choice
				 conProps <- {RemoveProps @conProps [LRef]}
				 SymbolicSolver, tellConstraint('='(VarTerm '+'(L 1.0)))
				 SymbolicSolver, Substitute(VarTerm)
				 {Show here2}
				 SymbolicSolver, distribute(Solution)
			      []
				 {Show here3}
				 conProps <- {RemoveProps @conProps [LRef]}
				 SymbolicSolver, tellConstraint('>'(VarTerm '+'(L 1.0)))
				 SymbolicSolver, Substitute(VarTerm)
				 SymbolicSolver, distribute(Solution)
			  
			      end
			   end
			end
			
		     [] '<='(term: U ref:URef)|_ then     %% case II.2: there is a closed upper bound, then try
			{Show case3_2}
			if {Sets.intersection {GetConstants U} EigenVars} \= nil then
			   {Show inter({Sets.intersection {GetConstants U} EigenVars})} 
			   {Show eigenVar}
			   fail
			else
			   choice                                %%  L < Var <= U)  ==>   Var = U or Var < U
			      conProps <- {RemoveProps @conProps [URef]}
			      SymbolicSolver, DiscardProps([p(ref: URef)])
			      SymbolicSolver, tellConstraint('='(VarTerm U))
			      SymbolicSolver, Substitute(VarTerm)
% 			      SymbolicSolver, UpdateTrace(['eq_is_leq'(lhs: ['='(VarTerm U)]
% 								       rhs: ['<='(VarTerm U)])
% 							  ])
			      SymbolicSolver, distribute(Solution)
			   []
			      conProps <- {RemoveProps @conProps [URef]}
			      SymbolicSolver, tellConstraint('<'(VarTerm U))
			      SymbolicSolver, normalizeCons
			      SymbolicSolver, UpdateTrace(['less-is-leq'(lhs: ['<'(VarTerm U)]
									 rhs: ['<='(VarTerm U)]
									 args: [VarTerm U])
							  ])
			      SymbolicSolver, distribute(Solution)
			   end
			end
			
		     [] '<'(term: U ref: URef)|_ then    %% case II.3:  there are ONLY _open_ lower and upper bounds
			MiddleTerm = '/'('+'(L U) 2.0)   %%  then compute the middle of the interval and 
			                                  %  create a disjunction 
%M = {MapleSimplify MiddleTerm} %% NO I/O from local space!!!
		     in
			{Show case3_3}
			{Show middleTerm#VarTerm#MiddleTerm}
			if {Sets.intersection {GetConstants MiddleTerm} EigenVars} \= nil then
			   {Show eigenVar}
			   fail
			else
			   choice
			      conProps <- {RemoveProps @conProps [LRef URef]}   %% remove both old bounds
			      {Show telling#VarTerm#'='#MiddleTerm}
			      SymbolicSolver, tellConstraint('='(VarTerm MiddleTerm))
			      SymbolicSolver, Substitute(VarTerm)
			      SymbolicSolver, UpdateTrace(['middle-is-less'(
							      lhs: ['<'(L U)]
							      rhs: ['<'(MiddleTerm U)]
							      args: [L U])
							   'middle-is-greater'(
							      lhs: ['<'(L U)]
							      rhs: ['<'(L MiddleTerm)]
							      args: [L U])
							  ])
			      SymbolicSolver, distribute(Solution)
			   []
			      conProps <- {RemoveProps @conProps [URef]}        %% keep the lower bound, delete the upper
			      {Show telling#VarTerm#'<'#MiddleTerm}
			      SymbolicSolver, tellConstraint('<'(VarTerm MiddleTerm))
			      SymbolicSolver, normalizeCons
			      SymbolicSolver, UpdateTrace(['less-transitivity'(
							      lhs: ['<'(VarTerm MiddleTerm) '<'(MiddleTerm U) ]
							      rhs: ['<'(VarTerm U)]
							      args: [VarTerm MiddleTerm U])
							  ])
			      SymbolicSolver, distribute(Solution)
			   []
			      conProps <- {RemoveProps @conProps [LRef]}        %% keep the upper bound, delete the lower
			      {Show telling#VarTerm#'>'#MiddleTerm}
			      SymbolicSolver, tellConstraint('>'(VarTerm MiddleTerm))
			      SymbolicSolver, normalizeCons
			      SymbolicSolver, UpdateTrace(['less-transitivity'(
							      lhs: ['<'(L MiddleTerm) '<'(MiddleTerm VarTerm) ]
							      rhs: ['<'(L VarTerm)])
							  ])
			      SymbolicSolver, distribute(Solution)
			   end
			end
		     %M = {MapleSimplify MiddleTerm} %% Problem: NO I/O from local space!!!
		     else
			{Show 'The List of upper bounds is neither empty, nor anything else :-('}
			fail
		     end
		  else  {Show 'The List of lower bounds is neither empty, nor anything else :-('}
		     SymbolicSolver, distribute(Solution exclude: Var|ExVars)
		  end
	       end
	    end
	 end
      end
      

      %%====================================================
      %%====================================================
      %%                 Private Methods 
      %%====================================================
      %%====================================================
      /** private method: */
      meth SortLowerBounds(Bounds $)
	 Ordering = fun{$ B1 B2}       %% heuristic ordering on problem variables
		       case B1
		       of '>='(...) then
			  true
		       else
			  case B2
			  of '>='(...) then
			     false
			  else
			     SymbolicSolver, AskLocal('>='(B1 B2) $)
			  end
		       end
		    end
      in
	 %{Show sort_low_bounds(Bounds)}
	 {Sort Bounds Ordering}
      end
      
      /** private method: */
      meth SortUpperBounds(Bounds $)
	 Ordering = fun{$ B1 B2}       %% heuristic ordering on problem variables
		       case B1
		       of '<='(...) then
			  true
		       else
			  case B2
			  of '<='(...) then
			     false
			  else
			     SymbolicSolver, AskLocal('<='(B1 B2) $)
			  end
		       end
		    end
      in
	 {Sort Bounds Ordering}
      end
      
      /** private method: */
      meth SortVars(Vars $)
	 LBoundDict = {NewDictionary}
	 UBoundDict = {NewDictionary}
	 Ordering = fun{$ Var1 Var2}       %% heuristic ordering on problem variables
		       LBounds1 = {Dictionary.get LBoundDict Var1}
		       LBounds2 = {Dictionary.get LBoundDict Var2}
		       UBounds1 = {Dictionary.get UBoundDict Var1}
		       UBounds2 = {Dictionary.get UBoundDict Var2}
		       AllBounds1 = {Append LBounds1 UBounds1}
		       AllBounds2 = {Append LBounds2 UBounds2}
		       Val1 = if AllBounds1 \= nil then
				 {Int.'div' {VarValue AllBounds1 0} {Length AllBounds1}}
			      else 0 end
		       Val2 =  if AllBounds2 \= nil then
				  {Int.'div' {VarValue AllBounds2 0} {Length AllBounds2}}
			       else 0 end
		    in
		       %{Show Var1(Val1)#Var2(Val2)}
		       Val1 >= Val2
		    end
	 
      in
	 {ForAll Vars proc{$ V}
			 LBounds = SymbolicSolver, getLowerBounds(V#var $)
			 UBounds = SymbolicSolver, getUpperBounds(V#var $)
		      in
			 {Dictionary.put LBoundDict V LBounds}
			 {Dictionary.put UBoundDict V UBounds}
		      end}
	 
	 {Map {Sort Vars Ordering}
	  fun{$ V}
	     V(lbounds: SymbolicSolver, SortLowerBounds({Dictionary.get LBoundDict V} $)
	       ubounds: SymbolicSolver, SortUpperBounds({Dictionary.get UBoundDict V} $)
	      )
	  end}
      end
      
      %%------------------
      %% Add a new constraint to the list of @constraints
      %%------------------
      /** private method: */
      meth AddConstraint(Con)
	 Rel = {Label Con}
	 NewCons = Con|@constraints.Rel
      in
	 %{Show rel#Rel#NewCons}
	 case @constraints
	 of con('<':Less '<=':Leq '=':Eq '>=':Geq '>':Grt 'IN':In) then
	    case Rel
	    of '<'  then constraints <- con('<':NewCons '<=':Leq     '=':Eq
					    '>=':Geq     '>':Grt   'IN':In)
	    [] '<=' then constraints <- con('<':Less    '<=':NewCons '=':Eq
					    '>=':Geq     '>':Grt   'IN':In)
	    [] '='  then constraints <- con('<':Less    '<=':Leq     '=':NewCons
					    '>=':Geq     '>':Grt    'IN':In)
	    [] '>=' then constraints <- con('<':Less    '<=':Leq     '=':Eq
					    '>=':NewCons '>':Grt    'IN':In)
	    [] '>'  then constraints <- con('<':Less    '<=':Leq     '=':Eq
					    '>=':Geq     '>':NewCons 'IN':In)
	    [] 'IN' then constraints <- con('<':Less    '<=':Leq     '=':Eq
					    '>=':Geq     '>':Grt    'IN':NewCons)
	    else
	       {Show 'Unknown relation'#Rel}
	       fail
	    end
	 else
	    {Show 'Something went wrong with the @constraints feature'}
	 end
      end

      %%------------------
      %% Add a new assumption constraint to the list of @assumptions
      %%------------------
      /** private method: */
      meth AddAssumption(Con)
	 Rel = {Label Con}
	 NewCons = Con|@assumptions.Rel
      in
	% {Show assumptions(@assumptions)}
	 %{Show addAssumption(Rel NewCons)}
	 case @assumptions
	 of ass('<':Less '<=':Leq '=':Eq '>=':Geq '>':Grt 'IN':In) then
	    case Rel
	    of '<'  then assumptions <- ass('<':NewCons '<=':Leq     '=':Eq      '>=':Geq     '>':Grt   'IN':In)
	    [] '<=' then assumptions <- ass('<':Less    '<=':NewCons '=':Eq      '>=':Geq     '>':Grt   'IN':In)
	    [] '='  then assumptions <- ass('<':Less    '<=':Leq     '=':NewCons '>=':Geq     '>':Grt    'IN':In)
	    [] '>=' then assumptions <- ass('<':Less    '<=':Leq     '=':Eq      '>=':NewCons '>':Grt    'IN':In)
	    [] '>'  then assumptions <- ass('<':Less    '<=':Leq     '=':Eq      '>=':Geq     '>':NewCons 'IN':In)
	    [] 'IN' then assumptions <- ass('<':Less    '<=':Leq     '=':Eq      '>=':Geq     '>':Grt    'IN':NewCons)
	    else
	       {Show 'Unknown relation'#Rel}
	       fail
	    end
	 else
	    {Show 'Something went wrong with the @assumptions feature'}
	 end
      end
      
      %%------------------
      %% Insert a new entry to the list of assumption propagators.
      %%------------------
      /** private method: */
      meth AddAssPropagator(Con)
	 SymbProp = SymbolicSolver, SymbolicArgs(Con 1 $)
      in	
	 %{Show addAssPropagator(Con)}
	 %{Show symbPropagator(SymbProp)}
	 assProps <-  SymbProp|@assProps
      end

      %%------------------
      %% Insert a new entry to the list of propagators.
      %%------------------
      /** private method: */
      meth AddConPropagator(Con)
	 SymbProp = SymbolicSolver, SymbolicArgs(Con 1 $)
      in	
	 %{Show addConPropagator(Con)}
	 %{Show symbPropagator(SymbProp)}
	 conProps <-  SymbProp|@conProps
      end

      %%------------------
      %% Insert a new entry to the list of propagators.
      %%------------------
      /** private method: */
      meth AddTermPropagator(Term)
	 SymbProp = SymbolicSolver, SymbolicArgs(Term 2 $)
      in	
	 %{Show addTermPropagator(Term)}
	 %{Show symbPropagator(SymbProp)}
	 termProps <-  SymbProp|@termProps
      end
      

      
      /** private method: */
      meth Substitute(Var by:Term <= _)
	 skip
% 	 Name = {Car Var}
% 	 Subst = {NewDictionary}
%       in
% 	 if {IsDet Term} then 
% 	    {Dictionary.put Subst Name Term}
% 	    %%constraints <- {ApplySubst Subst @constraints}
% 	    assProps <- {ApplySubst Subst @assProps}
% 	    conProps <- {ApplySubst Subst @conProps}
% 	    termProps <- {ApplySubst Subst @termProps}
% 	    %SymbolicSolver, showProps
% 	 else 
% 	    Equations = {GetEquations Var {Append @assProps @conProps}}
% 	 in
% 	    if Equations == nil then skip
% 	    else '='(term: Term)|_ = Equations
% 	    in
% 	       {Dictionary.put Subst Name Term}
% 	       %%constraints <- {ApplySubst Subst @constraints}
% 	       assProps <- {ApplySubst Subst @assProps}
% 	       conProps <- {ApplySubst Subst @conProps}
% 	       termProps <- {ApplySubst Subst @termProps}
% 	       %SymbolicSolver, showProps
% 	    end
% 	 end
      end
      
      
      %%------------------
      %% Get the lower bounds for a variable.
      %%------------------
      /** private method: */
      meth getLowerBounds(Var $)
	 if Var== nil then nil
	 else
	    {GetLowerBounds Var {Append @assProps @conProps}}
	 end
      end
      
      %%------------------
      %% Get the upper bounds for a variable.
      %%------------------
      /** private method: */
      meth getUpperBounds(Var $)
	 if Var== nil then nil
	 else
	    {GetUpperBounds Var {Append @assProps @conProps}}
	 end
      end
      

      %%-------------------
      %%   Discards Propagators that belong to a given list of references-------------------
      %%-------------------
      /** private method: */
      meth DiscardProps(PRefs)
	 {Debug discardProps(PRefs)}
	 {KillProps PRefs}
      end

      %%------------------
      %% Update the search trace
      %%------------------
      /** private method: */
      meth UpdateTrace(Message)
	 %searchTrace <- Message|@searchTrace
	 searchTrace <- {Append Message @searchTrace}
      end

      
      %%------------------
      %% Update the assumptions newly introduced by alien replacement
      %%------------------
      /** private method: */
      meth UpdateNewAssumptions(Message)
	 newAssumptions <- Message|@newAssumptions
      end
      	 
      %%------------------
      %% Instantiate the search trace
      %%------------------
      /** private method: */
      meth InstantiateTrace(Subst)
	 AllProps = @termProps
      in
%	 {Show subst({Dictionary.entries Subst})}
	 {Show vorher}
	 {ForAll @searchTrace Show}

	 searchTrace <- {ApplySubst Subst @searchTrace}
	 searchTrace <- {Map @searchTrace
			 fun{$ Entry}
			    if {IsRecord Entry} then
			       {Record.mapInd Entry
				fun{$ Feat Val}
				   case Feat
				   of args then
				      {Map Val
				       fun{$ Term}
					  NewTerm = {ConstructTerm Term AllProps AllProps 
						     @alienReplacements CosieFuncs}
				       in
					  NewTerm
				       end}
				   else
				      {Map Val fun{$ Term}
						  {CompleteCons Term AllProps 
						   @alienReplacements '||'|CosieFuncs}
					       end}
				   end
				end}
			    else Entry end
			 end}
	 {Show 'Instantiated Search Trace:'}
	 {ForAll @searchTrace Show}
      end			       
      
      
      
      %%------------------
      %% Try to match a constraint rule with the current set of propagators.
      %%------------------
      /** private method: */
      meth Match(Rule $)
	 ConMatching = if {HasFeature Rule cons} then
			  {MatchList Rule.cons @conProps @termProps
			   nil CosieRels {NewDictionary}}   %% first try to match the constraints
		       else nil end
      in
	 if (ConMatching \= nil) then
	    Matching= if {HasFeature Rule ass} then
			  {MatchList Rule.ass @assProps @termProps
			   nil CosieRels ConMatching}
		       else ConMatching end
	 in
	    if (Matching \=nil) then
	       Guard = if {HasFeature Rule guard} then
			  %{Show guard({ApplyMatching Rule.guard Matching})}
			  {All {ApplyMatching Rule.guard Matching}
			   fun{$ G}
			      SymbolicSolver, AskLocal(G $)
			   end}
		       else true end
	    in
	       if Guard then 
		%  {Show ' found matching'#{Label Rule}} %#{Dictionary.entries Matching}}
		  DelProps = if {HasFeature Rule del} then
				{ApplyMatching Rule.del Matching}
			     else nil end 
		  
		  AddProps = if {HasFeature Rule add} then
				if Rule.add == failure then
				   fail
				else
				   {ApplyMatching Rule.add Matching}
				end
			     else nil end
		  
		  AddAssProps = if {HasFeature Rule addAss} then
				   {ApplyMatching Rule.addAss Matching}
				else nil end
		  SymbAddAss = {Map AddAssProps
				fun{$ PS}
				   SymbolicSolver, SymbolicArgs(PS 1 $)
				end}
	       in
		  if {Some SymbAddAss fun{$ Prop}
					 {PropMember Prop @assProps}
				      end}
		  then
		     {Debug 'Assumption already exists.'}
		     false
		  else 
		     if DelProps \= nil then 
			SymbolicSolver, DiscardProps(DelProps)
			conProps <- {RemoveProps @conProps DelProps}
			termProps <- {RemoveProps @termProps DelProps}
		     else skip end
		     
		     {ForAll AddProps
		      proc{$ PS}
			 SymbProp = SymbolicSolver, SymbolicArgs(PS 1 $)
		      in
			 if {Member {Label SymbProp} CosieFuncs} then
			    SymbolicSolver, PropagateTerm(SymbProp _ _)
			 elseif {Member {Label SymbProp} CosieRels} then 
			    NewCon = SymbolicSolver, Propagate(SymbProp $)
			 in
			    SymbolicSolver, AddConstraint(NewCon)
			 end
		      end}
		     
		     {ForAll SymbAddAss
		      proc{$ PS}
			 NewCon = SymbolicSolver, PropagateAss(PS $)
		      in
			 SymbolicSolver, AddAssumption(NewCon)
		      end}
		     
		     
		     if {HasFeature Rule trace} then
			SymbolicSolver, UpdateTrace({ApplyMatching Rule.trace Matching})
		     else skip end
		     %SymbolicSolver, showProps
		     true
		  end
	       else
		  {Debug 'Guard is not entailed.'}
		  false
	       end
	    else {Debug 'Assumptions could not be matched.'}
	       false
	    end
	 else
	    {Debug 'Constraints could not be matched.'}
	    false
	 end
      end
      
      %%------------------
      %% Apply the Termrewrite-rules.
      %%------------------
      /** private method: */
      meth RewriteTerms
	 AlreadyApplied = {Length @appliedRewriteRules}
      in
	 {Wait RewriteRules}
	 
	 {ForAll RewriteRules
	  proc{$ Rule}
	     if SymbolicSolver, Match(Rule $) then
		appliedRewriteRules <- {Label Rule}|@appliedRewriteRules
	     else skip end
	  end}
	 
	 %{self showProps}
	 
	 if {Length @appliedRewriteRules} > AlreadyApplied then
	    {Show repeat}
	    SymbolicSolver, RewriteTerms
	 else
	    skip %SymbolicSolver, showProps
	 end
      end
      
      %%------------------
      %% Simplify the list of propagators by applying constrain rules, if possible.
      %%------------------
      /** private method: */
      meth SimplifyCons
	 AlreadyApplied = {Length @appliedSimpRules}
      in
	 {Wait SimpRules}
	 
	 {ForAll SimpRules
	  proc{$ Rule}
	     if (SymbolicSolver, Match(Rule $)) then
		appliedSimpRules <- {Label Rule}|@appliedSimpRules
	     else skip end
	  end}
	 
	 if {Length @appliedSimpRules} > AlreadyApplied then
	    SymbolicSolver, SimplifyCons
	 else
	    skip %SymbolicSolver, showProps
	 end
      end
      
      %%------------------
      %% Propagates a lineq-constraint, adds new variables and constants and
      %% creates an entry in @conProps for each propagator
      %%------------------
      /** private method: */
      meth PropagateAss(Con $)
	 %{Show con#Con}
	 case Con
	    %% The relation propagators
	 of 'AND'(C1 C2) then                %% &
	    NewC1 = SymbolicSolver, PropagateAss(C1 $)
	    NewC2 = SymbolicSolver, PropagateAss(C2 $)
	 in
	    'AND'(NewC1 NewC2 nil)
	    
	 [] '<'(L R) then                    %% <
	    '>'(NewR NewL Ref) = SymbolicSolver, PropagateAss('>'(R L) $)
	 in
	    '<'(NewL NewR Ref)
	    
	 [] '<='(L R) then                   %% <=
	    X Y 
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	    Ref = {FCP_RI.'RI.lessEq' X Y}
	 in
	    %{Show x#y#X#Y}
	    SymbolicSolver, AddAssPropagator('<='(X Y ref:Ref))
	    '<='(NewL NewR Ref)
	    
	    %% equations are treated differently ...
	 [] '='(L R) then
	    X Y
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	    NestedX = {FindName X @nestedVars}
	    NestedY = {FindName Y @nestedVars}
	 in
	    X = Y                % really propagate the equation!
	    case NewL
	    of LName#var then    % equation 'determinates' a variable!
	       undetVars <- {SetDifference @undetVars [LName]} %.. so remove it from the List of undetermined Vars
	       if NestedY \= nil then
		  SymbolicSolver, AddAssPropagator('='(NewL NestedY#nested ref: nil))
	       else
		  SymbolicSolver, AddAssPropagator('='(NewL NewR ref: nil))
	       end
	    else
	       case NewR
	       of RName#var then
		  undetVars <- {SetDifference @undetVars [RName]}
		  if NestedX \= nil then
		     SymbolicSolver, AddAssPropagator('='(NewR NestedX#nested ref: nil))
		  else
		     SymbolicSolver, AddAssPropagator('='(NewR NewL ref: nil))
		  end
	       else
		  SymbolicSolver, AddAssPropagator('='(if NestedX \= nil then
						       NestedX#nested
						    else NewL end
						    if NestedY \= nil then
						       NestedY#nested
						    else NewR end
						    ref: nil))
	       end
	    end
	    '='(NewL NewR ref: nil)
	    
	 [] '>='(L R) then                   %% >=
	    NewR NewL
	    '<='(NewR NewL Ref) = SymbolicSolver, PropagateAss('<='(R L) $)
	 in
	    '>='(NewL NewR Ref)
	    
	 [] '>'(L R) then                    %% >
	    X Y
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	    Ref = {FCP_RI.'RI.greater' X Y}
	 in
	    %{Show ref#Ref}
	    %{Show rivars#X#Y}
	    SymbolicSolver, AddAssPropagator('>'(X Y ref:Ref))
	    '>'(NewL NewR Ref)

	 [] 'IN'(L Sort) then    %% At the moment _natural numbers_ must only be > 0 !!!
	    if Sort == 'NAT' then
	       X
	       D= {FD.decl}
	       NewL = SymbolicSolver, PropagateTerm(L X $)
	       Ref = {FCP_RI.'RI.intBounds' X D}
	    in
	       'IN'(NewL Sort Ref)
	    else fail end
	 else
	    {Show 'Cannot propagate constraint'#Con}
	    fail
	 end
      end %method
      
      %%------------------
      %% starts propagation for a constraint, adds new variables and constants and
      %% creates an entry in @conProps for each propagator
      %%------------------
      /** private method: */
      meth Propagate(Con $)
	 %{Show con#Con}
	 case Con
	    %% The relation propagators
	 of 'AND'(C1 C2) then                %% &
	    NewC1 = SymbolicSolver, Propagate(C1 $)
	    NewC2 = SymbolicSolver, Propagate(C2 $)
	 in
	    'AND'(NewC1 NewC2 nil)
	    
	 [] '<'(L R) then                    %% <
	    '>'(NewR NewL Ref) = SymbolicSolver, Propagate('>'(R L) $)
	 in
	    '<'(NewL NewR Ref)
	    
	 [] '<='(L R) then                   %% <=
	    X Y 
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	    Ref = {FCP_RI.'RI.lessEq' X Y}
	 in
	    %{Show x#y#X#Y}
	    SymbolicSolver, AddConPropagator('<='(X Y ref:Ref))
	    '<='(NewL NewR Ref)
	    
	    %% equations are treated differently ...
	 [] '='(...) then
	    SymbolicSolver, PropagateEquation(Con $)
	    
	 [] '>='(L R) then                   %% >=
	    NewR NewL
	    '<='(NewR NewL Ref) = SymbolicSolver, Propagate('<='(R L) $)
	 in
	    '>='(NewL NewR Ref)
	    
	 [] '>'(L R) then                    %% >
	    X Y
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	    Ref = {FCP_RI.'RI.greater' X Y}
	 in
	    %{Show ref#Ref}
	    %{Show rivars#X#Y}
	    SymbolicSolver, AddConPropagator('>'(X Y ref:Ref))
	    '>'(NewL NewR Ref)

	 [] 'IN'(L Sort) then    %% At the moment _natural numbers_ must only be > 0 !!!
	    if Sort == 'NAT' then
	       X
	       D = {FD.decl}
	       NewL = SymbolicSolver, PropagateTerm(L X $)
	       Ref = {FCP_RI.'RI.intBounds' X D}
	    in
	       'IN'(NewL Sort Ref)
	    else fail end
	 else
	    {Show 'Cannot propagate constraint'#Con}
	    fail
	 end
      end %method
      
      %%------------------
      %% Propagates an equation, adds new variables and constants and
      %% creates an entry in @conProps for each propagator
      %%------------------
      /** private method: */
      meth PropagateEquation(Con $)
	 '='(L R) = Con
	 fun{LocalProp Con}
	    X Y
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	    NestedY = {FindName Y @nestedVars}
	    AlienY = {FindName Y @alienVars}
	 in
	    %{Show equality(NewL NewR NestedY)}
	    %if {IsDet X} orelse {IsDet Y} then skip else X = Y end
	    X = Y                 % really propagate the equation!
	    case NewL
	    of LName#var then    % equation 'determinates' a variable!
	       Equations = {GetEquations LName#var {Append @assProps @conProps}}
	    in
	       case Equations 
	       of nil then                  %% there are no equations for LName#var so far ...
		  undetVars <- {SetDifference @undetVars [LName]}
                  %%.. so remove it from the List of undetermined Vars
		  if (NestedY \= nil) then
		     if {IsDet Y} then
                        SymbolicSolver, AddConPropagator('='(NewL Y ref: nil))
                     else
                        SymbolicSolver, AddConPropagator('='(NewL NestedY#nested ref: nil))
                     end
		  elseif (AlienY \= nil) then
		     SymbolicSolver, AddConPropagator('='(NewL AlienY#alien ref: nil))
		  else
		     SymbolicSolver, AddConPropagator('='(NewL NewR ref: nil))
		  end
	       [] '='(term: T)|_ then       %% there is another equation for LName#var ...
		  AllProps = {Append {Append @assProps @conProps} @termProps}
		  NewT = {ConstructTerm T AllProps AllProps @alienReplacements CosieFuncs}
		  MGU = {Unify NewT R}      %% ... so try to unify the terms on the right hand side
	       in
		  if MGU == nil  then
		     fail                   %% terms could not be unified
		  else 
% 		     {Show term1(NewT)}
% 		     {Show term2(R)}
% 		     {Show unify({Dictionary.entries MGU})}
		     if {Dictionary.entries MGU} == nil then
			skip
		     else
			MGUTerm = {ApplySubst MGU NewT}
			NewR = SymbolicSolver, PropagateTerm(MGUTerm _ $)
		     in
			undetVars <- {SetDifference @undetVars [LName]} %.. so remove it from the List of undetermined Vars
			{ForAll {Dictionary.entries MGU}
			 proc{$ VarName#Term}
			    SymbolicSolver, PropagateEquation('='(VarName#var Term) _)
			    SymbolicSolver, Substitute(VarName#var)
			 end}
			
			conProps <- {RemoveProps @conProps ['='(NewL T)]}
			SymbolicSolver, AddConPropagator('='(NewL NewR ref: nil))
		     end
		  end
	       else skip end
	    else skip end
	    '='(NewL NewR ref: nil)
	 end
      in
	 %{Show equation#Con}
	 case R 
	 of _#var then
	    case L
	    of _#var then
	       {LocalProp Con}
	    else
	       SymbolicSolver, PropagateEquation('='(R L) $)
	    end
	 else
	    {LocalProp Con}
	 end
      end
      
      
      %%------------------
      %% Propagates a term and adds a propagator reference to
      %% each subterm, i.e., returns terms of the form '+'(A B  Ref)
      %% The method creates an entry in @termProps for each propagator
      %% The parameter RIVar must be unbound. It is bound to the RI-variable
      %% representing the term Term.
      %%------------------
      /** private method: */
      meth PropagateTerm(Term RIVar $)
	 %{Show term#Term}
	 case Term 

	    %% ARITHMETIC PROPAGATORS
	 of 'MIN'(L R) then
	    %Ref
	    NewL = SymbolicSolver, PropagateTerm(L _ $)
	    NewR = SymbolicSolver, PropagateTerm(R _ $)
	 in
	    RIVar = SymbolicSolver, NewNestedVar($)
	    %Ref = {RI.makeMin X Y Z))}
	    SymbolicSolver, AddTermPropagator('MIN'(NewL NewR RIVar ref:nil) ) 
	    'MIN'(NewL NewR nil)

	 [] 'MAX'(L R) then
	    %Ref
	    NewL = SymbolicSolver, PropagateTerm(L _ $)
	    NewR = SymbolicSolver, PropagateTerm(R _ $)
	 in
	    RIVar = SymbolicSolver, NewNestedVar($)
	    %Ref = {RI.makeMin X Y Z))}
	    SymbolicSolver, AddTermPropagator('MAX'(NewL NewR RIVar ref:nil) ) 
	    'MAX'(NewL NewR nil)

	 [] '+'(L R) then
	    X Y Ref
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	 in
	    RIVar = SymbolicSolver, NewNestedVar($)
	    Ref = {FCP_RI.'RI.plus' X Y RIVar}
	    SymbolicSolver, AddTermPropagator('+'(X Y RIVar ref:Ref))
	    '+'(NewL NewR Ref)
	    
	 [] '-'(L R) then
	    X Y Ref
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	 in 
	    RIVar = SymbolicSolver, NewNestedVar($)
	    Ref = {FCP_RI.'RI.plus' RIVar Y X}
	    SymbolicSolver, AddTermPropagator('-'(X Y RIVar ref:Ref) )
	    '-'(NewL NewR Ref)
	    
	 [] '*'(L R) then
	    X Y Ref
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	 in
	    SymbolicSolver, NewNestedVar(RIVar)	    
	    %{Show times(X Y RIVar)}
	   % Ref = {FCP_RI.'`Times`' X Y RIVar}
            {RI.times X Y RIVar}
            Ref = nil
	    %{Show times(X Y RIVar)}
	    SymbolicSolver, AddTermPropagator('*'(X Y RIVar ref:Ref) )
	    '*'(NewL NewR Ref)
	    
	 [] '/'(L R) then
	    X Y Ref
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	 in
	    RIVar = SymbolicSolver, NewNestedVar($)
	    Ref = {FCP_RI.'RI.times' Y RIVar X}
	    %{Show divi#X#Y#RIVar}
	    SymbolicSolver, AddTermPropagator('/'(X Y RIVar ref:Ref) )
	    '/'(NewL NewR Ref)
	    
	 [] '/'(L R N) then
	    X Y Z
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	    SymbolicSolver, PropagateTerm(N Z _)
	    Ref = {FCP_RI.'RI.times' Z Y X}
	 in
	    %{Show propadiv#X#Y#Z}
	    SymbolicSolver, AddTermPropagator('/'(X Y Z ref:Ref))
	    '/'(NewL NewR Ref)
	    
	 []'||'(T) then
	    Name 
	 in
	    Name#RIVar = SymbolicSolver, NewAlienVar($)
	    %{Show 'Replacing the alien term '#Term#' by '#Name}
	    {Dictionary.put @alienReplacements Name Term}
	    SymbolicSolver, tellAssumption('>='(Name#alien 0))
	    %SymbolicSolver, UpdateNewAssumptions(absval_geq_zero(rhs: ['>='('||'(T) 0)]))
	    SymbolicSolver, UpdateTrace(['absval-geq-zero'(rhs: ['<='(0 Name#alien)]
							   args: [T]
							  )
					])
	    Name#alien
	    
	 [] '^'(N E) then
	    if {IsNumber N} andthen {IsNumber E} then
	       V = {Pow N E}
	       F = if {IsInt V} then {IntToFloat V} else V end
	    in
	       RIVar = {RI.var.new F F}
	       '^'(N E)
	    else fail unit end
	 [] 'SQRT'(N) then
	    if {IsNumber N} then
	       R = if {IsInt N} then {Sqrt {IntToFloat N}}
		   else {Sqrt N} end
	    in
	       RIVar = {RI.var.new R R}
	       'SQRT'(N)
	    else fail unit end
	    
	    %% VARIABLES, SYMBOLIC CONSTANTS and NUMBERS 
	 [] N#Type then Found in
	    if {IsAtom N} then
	       case Type
	       of var then 
		  %{Show varcase1#N}
		  probVars <- {GetVar N RIVar Found @probVars}
		  if {IsDet Found} then
		     skip
		  else
		     %{Show varcase2#N}
		     undetVars <- N|@undetVars
		  end
		  N#Type

	       [] nested then
		  %{Show nestcase#N}
		  nestedVars <- {GetVar N RIVar Found @nestedVars}
		  N#Type

	       [] alien then
		  %{Show aliencase#N}
		  alienVars <- {GetVar N RIVar Found @alienVars}
		  N#Type
		  
	       [] const then
		  %{Show constcase#N}
		  probConsts <- {GetConst N RIVar @probConsts}
		  N#Type
	       else
		  %{Show nestcase#N}
		  nestedVars <- {GetVar N RIVar Found @nestedVars}
		  N#Type
	       end
	    else
	       RIVar = {RI.var.new 1.0 1.0}
	       Term
	    end
	 else
	    if {IsAtom Term} then
	       nestedVars <- {GetVar Term RIVar _ @nestedVars}
	       Term
	    elseif {IsNumber Term} then
	       if {IsInt Term} then
		  F = {IntToFloat Term}
	       in
		  RIVar = F
		  F
	       else
		  RIVar = Term
		  Term
	       end
	    else
	       %%Term
	       Name
	    in
	       Name#RIVar = SymbolicSolver, NewAlienVar($)
	       %{Show 'Replacing the alien term '#Term#' by '#Name}
	       {Dictionary.put @alienReplacements Name Term}
	       Name#alien
	    end
	 end
      end % method
      
      %%------------------
      %% Test whether the constraint Con is entailed by the constraint store.
      %%------------------
      /** private method: */
      meth Ask(Con)
	 {Show 'askcon'#Con}
	 case Con
	    %% The relation propagators
	 of 'AND'(C1 C2) then
	    SymbolicSolver, Ask(C1 _)
	    SymbolicSolver, Ask(C2 _)
	    
	 [] '<'(L R) then 
	    SymbolicSolver, Ask('>'(R L))
	    
	 [] '<='(L R) then
	    X = SymbolicSolver, PropagateTerm(L $ _)
	    Y = SymbolicSolver, PropagateTerm(R $ _)
	    XU = if {IsDet X} then X else {RI.getUpperBound X} end
	    YL = if {IsDet Y} then Y else {RI.getLowerBound Y} end
	 in
	    %%{Show xu#yl#XU#YL#(XU =< YL)#(XU>YL)}
	    if (XU =< YL) then skip
	    else fail end
	    
	    %% equality
	 [] '='(L R) then
	    X = SymbolicSolver, PropagateTerm(L $ _)
	    Y = SymbolicSolver, PropagateTerm(R $ _)
	 in
            if {System.eq X Y} then
               skip
            else
               fail
            end
%             if {IsDet X} andthen {IsNumber X} then
% 	       if {IsDet Y} then
% 		  if {IsNumber Y} andthen (X == Y) then
% 		     skip
% 		  else fail end
% 	       else fail end
% 	    elseif {IsGroundTerm L nil} andthen {IsGroundTerm R nil} then 
% 	       MGU = {Unify L R}      %% ... so try to unify the terms of the equality
% 	    in
% 	       {Browse MGU}
% 	       if MGU == nil then
% 		  fail
% 	       else
% 		  if ({RI.getUpperBound X} == {RI.getUpperBound Y})
% 		     andthen ({RI.getLowerBound X} == {RI.getLowerBound Y}) then
% 		     skip
% 		  else fail end
% 	       end
% 	    else
% 	       if {System.eq X Y} then skip else fail end
% 	    end
	 [] '>='(L R) then
	    SymbolicSolver, Ask('<='(R L))
	    
	 [] '>'(L R) then
	    X = SymbolicSolver, PropagateTerm(L $ _)
	    Y = SymbolicSolver, PropagateTerm(R $ _)	    
	    YU = if {IsDet Y} then Y else {RI.getUpperBound Y} end
	    XL = if {IsDet X} then X else {RI.getLowerBound X} end
	 in
	    {Show X#Y#XL#YU}
	    if XL > YU then skip
	    else fail end
	 else fail end
      end
      
      /** private method: */
      meth InstantiateVar(Term RIVar $)
	 %{Show InstantiateVar#Term}
	 %% variables, symbolic constants and numbers 
	 case Term
	 of N#Type then 
	    if {IsAtom N} then Found in
	       case Type
	       of var then 
		  %{Show varcase1#N}
		  probVars <- {GetVar N RIVar Found @probVars}
		  Found
		  
	       [] nested then
		  %{Show nestcase#N}
		  nestedVars <- {GetVar N RIVar Found @nestedVars}
		  Found

	       [] alien then
		  %{Show aliencase#N}
		  alienVars <- {GetVar N RIVar Found @alienVars}
		  Found
		  
	       [] const then
		  %{Show constcase#N}
		  probConsts <- {GetConst N RIVar @probConsts}
		  Found
	       else
		  %{Show nestcase#N}
		  nestedVars <- {GetVar N RIVar Found @nestedVars}
		  Found
	       end
	    else false end
	 else
	    if {IsNumber Term} then
	       if {IsInt Term} then
		  F = {IntToFloat Term}
	       in
		  RIVar = F
		  true
	       else
		  RIVar = Term
		  true
	       end
	    else false end
	 end
      end % method

      %%------------------
      %% Test whether the constraint Con is entailed by the constraint store.
      %%------------------
      /** private method: */
      meth AskLocal(Con $)
	 %{Show 'ask_local_con'#Con}
	 case Con
	    %% The relation propagators
	    
	 of '<'(L R) then 
	    SymbolicSolver, AskLocal('>'(R L) $)
	    
	 [] '<='(L R) then
	    X Y
	 in
	    if (SymbolicSolver, InstantiateVar(L X $)) andthen
	       (SymbolicSolver, InstantiateVar(R Y $)) then 
	       XU = if {IsDet X} then X else {RI.getUpperBound X} end
	       YL = if {IsDet Y} then Y else {RI.getLowerBound Y} end
	    in
	       %%{Show xu#yl#XU#YL#(XU =< YL)#(XU>YL)}
	       (XU =< YL) 
	    else false end
	    
	    %% equality
	 [] '='(L R) then
	    X Y
	 in
	    if (SymbolicSolver, InstantiateVar(L X $)) andthen 
	       (SymbolicSolver, InstantiateVar(R Y $)) then 
	       if {IsDet X} then
		  if {IsDet Y} then
		     X == Y 
		  else false end
	       else
		  if {IsDet Y} then
		     false
		  else 
		     ({RI.getUpperBound X} == {RI.getUpperBound Y})
		     andthen ({RI.getLowerBound X} == {RI.getLowerBound Y}) 
		  end
	       end
	    end
	 [] '>='(L R) then
	    SymbolicSolver, AskLocal('<='(R L) $)
	    
	 [] '>'(L R) then
	    X Y
	 in
	    if (SymbolicSolver, InstantiateVar(L X $)) andthen
	       (SymbolicSolver, InstantiateVar(R Y $)) then 
	       YU = if {IsDet Y} then Y else {RI.getUpperBound Y} end
	       XL = if {IsDet X} then X else {RI.getLowerBound X} end
	    in
	       %{Show ask_local(X Y XL YU)}
	       (XL > YU)
	    else false end
	 else false end
      end
      
      
      %%------------------
      %% Get the names for a list of variables.
      %%------------------
      /** private method: */
      meth SymbolicArgs(Prop Index $)
	 %{Show symbolic#Prop}
	 if {IsDet Prop} then
	    if {IsRecord Prop} then
	       case Prop
	       of '='(Var Term ref: Ref) then 
		  '='(SymbolicSolver, GetRIVarName1(Var $)
		      SymbolicSolver, GetRIVarName2(Term $)
		      ref: Ref)
	       else 
		  {Record.map Prop fun{$ A}
				      SymbolicSolver, SymbolicArgs(A 1 $)
				   end}
	       end
	    else 
	       Prop
	    end
	 else
	    if Index == 1 then
	       SymbolicSolver, GetRIVarName1(Prop $)
	    else
	       SymbolicSolver, GetRIVarName2(Prop $)
	    end
	 end
      end

      %%------------------
      %% Create a new name for a nested variable.
      %%------------------
      /** private method: */
      meth NewName(Prefix $)
	 nameCounter <- @nameCounter + 1
	 {String.toAtom {VirtualString.toString Prefix#@nameCounter}}
      end
      
      %%------------------
      %% Create a new nested RI variable.
      %%------------------
      /** private method: */
      meth NewNestedVar($)
	 V = {RI.var.decl}
      in
	 nestedVars <- {Append [SymbolicSolver, NewName("Nested" $)#V]
			@nestedVars}
	 V    %% return the RI-var
      end
      
      %%------------------
      %% Create a new RI variable as a placeholder for an alien term.
      %%------------------
      /** private method: */
      meth NewAlienVar($)
	 Name = SymbolicSolver, NewName("Alien" $)
	 Var = {RI.var.decl}
      in
	 alienVars <- {Append [Name#Var]
		       @alienVars}
	 Name#Var  %% return the NewName _and_ the RI-var
      end
      
      %%------------------
      %% Get the name of a RI variable.
      %%------------------
      /** private method: */
      meth GetRIVarName1(Var $)
	 if {IsDet Var} then
	    Var
	 else
	    VarResult = {FindName Var @probVars} in
	    if VarResult == nil then 
	       ConstResult = {FindName Var @probConsts} in
	       if ConstResult == nil then
		  NestedResult = {FindName Var @nestedVars} in
		  if NestedResult == nil then 
		     AlienResult = {FindName Var @alienVars} in
		     if AlienResult == nil then
			Var
		     else AlienResult#alien end
		  else NestedResult#nested end
	       else ConstResult#const end
	    else VarResult#var    end
	 end
      end
      
      %%------------------
      %% Get the name of a RI variable.
      %%------------------
      /** private method: */
      meth GetRIVarName2(Var $)
	 if {IsDet Var} then
	    case Var
	    of Name#alien then 
	       AlienTerm = {Dictionary.get @alienReplacements Name}
	    in
	       AlienTerm
	    else
	       Var
	    end
	 else
	    ConstResult = {FindName Var @probConsts} in
	    if ConstResult == nil then
	       VarResult = {FindName Var @probVars} in
	       if VarResult == nil then 
		  NestedResult = {FindName Var @nestedVars} in
		  if NestedResult == nil then 
		     AlienResult = {FindName Var @alienVars} in
		     if AlienResult == nil then
			var
		     end
		  else NestedResult#nested end
	       else VarResult#var    end
	    else ConstResult#const end
	 end
      end

      %%------------------
      %% Get RI variable for a given name or alien term
      %%------------------
      /** private method: gives the RI-variable with a given name*/
      meth GetRIVar(Name $)
	 NestedResult = {FindVar Name @nestedVars} in
	 if NestedResult == nil then 
	    VarResult = {FindVar Name @probVars} in
	    if VarResult == nil then 
	       ConstResult = {FindVar Name @probConsts} in
	       if ConstResult == nil then
		  AlienResult = {FindVar Name @alienVars} in
		  if AlienResult == nil then
		     Name
		  else AlienResult end
	       else ConstResult end
	    else VarResult    end
	 else NestedResult end
      end
      

   end %class
end %define
