%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FSCosClass.oz ; This file is part of the OMEGA system
%
%   major updates: 26.10.1999,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA Project                                                        
%   AG Siekmann/FB Informatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)
% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   %% The class of FSCos constraint solvers, Version 2.0
   %%      jzimmer 19.1.99 
import
   System(show:Show showInfo: ShowInfo printInfo gcDo)
   Browser(browse: Browse)
   Space
   Open
   
   Explorer
   DaVinci
   Search(one: SearchOne)
   
   Printer('class': PrinterClass) at 'x-ozlib://mathweb/cs/share/Printer.ozf'
   ContextTreeClass('class': ContextTree) at 'x-ozlib://mathweb/cs/share/ContextTreeClass.ozf'
   FSDomain(isConstraint:     IsConstraint
	    isPureConstraint: IsPureConstraint
	    isGround:         IsGround
	    solver:           DomainSolver
	    oz2PostDict:      Oz2PostDict
	   ) 
   at 'x-ozlib://mathweb/cs/fscos/FSDomain.ozf'
   Share(car: Car
	 cdr: Cdr)
   at 'x-ozlib://mathweb/cs/share/Share.ozf'
   Translate at 'x-ozlib://mathweb/cs/share/Translate.ozf'
   
export
   'class': FSCos

define
   Version = 1.0
   
   fun {RenameSymbolicConstrs CStores}
      CStores
   end
   
   fun{MergeConstraintStores CStores}
      case CStores
      of nil then nil
      [] C|Cr then
	 {Append C.constraints {MergeConstraintStores Cr}}
      else nil end 
   end
   

   %%===============================================
   %%===============================================
   %%  The class of all constraint solvers.
   %%===============================================
   class FSCos             
      from BaseObject
	 
      prop
	 locking

      attr
	 contextTree: nil     %% the context tree
	 messages: nil        %% all important messages received
	 metaConstraints: nil %% the meta constraints (e.g. not-in(x D) )
	 eigenVariables: nil  %% the Eigenvariable conditions

	 problemName: 'UNKNOWN' %% the name of the planning problem (for the term printer)
	 vinci: nil             %% the DaVinci object showing the context tree
	 app: nil               %% the XApplication (for output)
	 traceFile: nil         %% the traceFile if open
	 printer: nil           %% the TermPrinter


	 %%-----------------
	 %% initialize the constraint solver with
	 %% an application manager and a trader URL.
	 %%-----------------
      meth init(App TraderURL)
	 {Show hallo_init}
	 app <- App
	 {Show 'FSCos-service version '#Version#'requested from '#TraderURL}
      end
      
      %%-----------------
      %% reset the Solver with hyps Hyps
      %%-----------------
      /** reset FSCos for planning a new problem */
      %% @1=ProblemName
      meth reset(ProblemName $)
	 contextTree <- {New ContextTree init}  %% open a new context tree
	 
	 messages <- nil              %% no messages
	 metaConstraints <- nil       %% no meta-constraints
	 eigenVariables <- nil        %% no Eigenvariables 
	 
	 
	 FSCos, stopTrace(_)           %% close the trace file if open
	 
	 problemName <- ProblemName   %% store the problem name 
	 
	 if @printer \= nil then      %% reset the TermPrinter if it's initialized.
	    Title = 'FSCos constraint store(s) for '#ProblemName
	 in
	    {@printer newTitle(Title)} 
	    {@printer clear}
	 else skip end
	 
	 {ShowInfo 'FSCos has been reseted for planning '#ProblemName}
	 'T'
      end
      
      %%-----------------
      %% browse the object
      %%-----------------
      meth browse($)
	 {@contextTree browse}
	 'T'
      end
      
      %%-----------------
      %% Show the object
      %%-----------------
      meth show($)
	 {Show 'Messages:'#@messages}
	 {@contextTree show}
	 'T'
      end
      
      %%-----------------
      %% Open a constraint printer window
      %%-----------------
      /** open a constraint viewer */
      meth showConstraints($)
	 Title = 'FSCos constraint store(s) for '#@problemName
      in
 	 if @printer \= nil then
	    {@printer close}
	 else skip end
	 
	 printer <- {New PrinterClass init(Title)}
	 
	 if @contextTree \= nil then
	    FSCos, PrintConstraints
	 else skip end
	 
	 'T'
      end

      %%-----------------
      %% close the constraint printer
      %%-----------------
      /** close the constraint viewer */
      meth unshowConstraints($)
	 if @printer \= nil then
	    {@printer close}
	    printer <- nil
	 else skip end
	 'T'
      end
      
      %%-----------------
      %% opens the DaVinci browser.
      %%-----------------
      /** open a daVinci browser for the ContextTree */
      meth openDaVinci($)
	 if @vinci == nil then %% Stream braucht man im Moment nicht
	    vinci <- {New DaVinci.daVinciClass init(_)}
	    {@vinci send(nil)}
	    FSCos, daVinci
	    'T'
	 else 'NIL' end
      end
      
      %%-----------------
      %% close the DaVinci browser.
      %%-----------------
      /** close a daVinci browser for the ContextTree */
      meth closeDaVinci($)
	 if @vinci \= nil then
	    {@vinci close}
	    vinci <- nil
	    'T'
	 else 'NIL' end
      end
	 
      %%-----------------
      %% sends a graph representation of the context tree to
      %% daVinci
      %%-----------------
      /** explicitly show the context tree in the daVinci browser */
      meth daVinci
	 if @contextTree \= nil then
	    VS = {@contextTree daVinci($)}
	 in
%	    {Browse vs#VS}
	    if @vinci \= nil then
	       {@vinci graph(VS)}
	    else skip end
	 else skip end
      end

      %%-----------------
      %% start tracing the messsages in an oz file with the given name.
      %%-----------------
      /** start tracing certain messages in a file */
      meth startTrace(FileName $)
	 if @traceFile == nil then
	    {Show 'startTrace'#FileName}
	    traceFile <- {New Open.file init(name:FileName flags:[write create])}
	    {@traceFile write(vs:"functor\nimport\n")}
	    {@traceFile write(vs:"   FSCosClass(FSCos: FSCos)\n   Application(exit)\n   XApplication\n   System(showInfo: ShowInfo)\n\n")}
	    {@traceFile write(vs:"define\n   local Cos={New FSCos init({New XApplication.manager init} 'Lim.oz')} Result Sol\n   in\n")}
	    {@traceFile write(vs:"      {Cos reset('"#@problemName#"' _)}\n\n")}
	    {@traceFile write(vs:"      %{Cos showConstraints(_)}\n\n")}
	    'T'
	 else
	    'NIL'
	 end
      end
      
      %%-----------------
      %% stop tracing messages.
      %%-----------------
      /** stop tracing messages */
      meth stopTrace($)
	 if @traceFile == nil then
	    'T'
	 else 
	    {@traceFile write(vs:"      %{Cos unshowConstraints(_)}\n")}
	    {@traceFile write(vs:"   end\n   %{Application.exit 0}\nend")}
	    {@traceFile close}
	    traceFile <- nil
	    'T'
	 end
      end
      
      /** private */
      meth Trace(VS)
	 if @traceFile \= nil then
	    {@traceFile write(vs:VS)}
	 else skip end
      end
      
      %%-----------------
      %% computes the reflection of the constraint store, i.e. the
      %% constraints collected from all context nodes.
      %%-----------------
      /** give a reflection of the current constraint stores */
      meth reflect($)
	 CStores = {@contextTree getAllCStores($)} % note: a constraint store has the
	                                                    % form  cstore(context: [...] constraints:[..])
	 Implications = {Map CStores fun{$ CStore}
					Context =  CStore.context
				     in
					if Context == nil then
					   {Map CStore.constraints Cdr}
					else
					   '->'({Map CStore.context Cdr}
						{Map CStore.constraints Cdr})
					end
				     end}
      in
	 {Show Implications}
%{@contextTree show}
	 {Translate.list2PostConjunct Implications false Oz2PostDict}
      end
      
      %%-----------------
      %% gives a POST representation of the list of Eigenvariables accumulated so far.
      %%-----------------
      /** give a list of the current Eigenvariables */
      meth eigenVars($)
	 B = {Translate.list2Post {Map @eigenVariables Car} false Oz2PostDict}
      in
	 {ShowInfo B}
	 B
	 %'TRUE'
      end
      
      %%-----------------
      %% gives a string representation of an association list with for the determined
      %% variables. The list has the form: ((Var1 . Term1) ... (Varn . Termn))
      %%-----------------
      /** give a list of the variables currently determined */
      meth getDetVars($)
	 if @contextTree == nil then
	    {Show 'The constraint-solver has not been initialized yet.'}
	    'NIL'
	 else
	    Leaf = {@contextTree getLeaf($)}
	    Clone = {Leaf clone($)}
	    Solver = {Space.merge {Clone space($)}}
	    DetVars = {Solver getDetVars($)}
	 in
	    {Show DetVars}
	    if DetVars == nil then
	       'NIL'
	    else
	       Res = {VirtualString.toString "("#{Translate.aList2Post DetVars Oz2PostDict}#")"}
	    in
	       {ShowInfo res#Res}
	       Res
	    end
	 end
      end
      
      %%-----------------
      %% Searches for a solution (instantiations) for the problem variables.
      %% Returns a post representation of the solution, if found,
      %% 'NIL' otherwise.
      %%-----------------
      /** try to find a solution for the accumulated constraints */
      meth findSolution($)
	 %% First, collect all constraint stores from the leaf nodes of context tree
	 %%
	 LeaveCStores = {@contextTree getAllLeafCStores($)} % note: a constraint store has the
	                                                     % form  cstore(context: [...] constraints:[..])
	            % then rename the symbolic constants occuring in different contexts.
	 RenamedCStores = {RenameSymbolicConstrs LeaveCStores}
	             % finally, merge all constraint stores to one list of constraints.
	 Constraints = {MergeConstraintStores RenamedCStores}	 
	 
             % compute the unary search procedure for the root computation space
	 SearchProcedure = proc{$ Sol}
			      SymbolicSolver = {New DomainSolver init}
			   in
			      {ForAll Constraints
			       proc{$ C}                       %% first propagate all constraints...
				  case C
				  of nil then skip
				  [] _#Con then
				     {Show tell(Con)}
				     {SymbolicSolver tellConstraint(Con)}
				     {System.gcDo}				     
				  else skip end
			       end}
			      
			      {SymbolicSolver setEigenVars(@eigenVariables)} %% propagate the EigenVariable conditions...
			      {SymbolicSolver distribute(Sol)}               %% ...then search for a solution
			   end
	 Result
	 Solution
      in
	 {Show RenamedCStores}
	 {Show Constraints}
	 {Explorer.object delete(information all)}
	 
	 {Explorer.object add(information
			      proc {$ I S}
				 if {IsDet S} then {Browse {Value.type S}} else skip end
				 {Browse I#S}
			      end
			      label: 'Constraint Store'
			     )}
	 
	    %{Explorer.object script(SearchProcedure)}	
	 
	 {Show 'Trying to find a solution for the problem variables!'}
	 
	 {SearchOne.depth SearchProcedure 1 _ Result} %kein KillP
	 
	 [Solution] = Result
	 {ShowInfo 'Solution'#{Translate.list2PostConjunct Solution false Oz2PostDict}}
	 
	 if {IsDet Solution} then      %% really found a solution, i.e., a list of equations?
	    {Translate.list2PostConjunct Solution false Oz2PostDict}
	 else 'NIL' end
      end
      
   
      
      %%-----------------
      %% prints all constraint stores to a text window
      %%-----------------
      %%-----------------
      %% prints all constraint stores to a text window
      %%-----------------
      /** private */
      meth PrintConstraints
	 if @printer \= nil then
	    CStores = {@contextTree getAllCStores($)} % note: a constraint store has the
	                                                    % form  cstore(context: [...] constraints:[..])
	 in

	    thread       %%  do it parallel due to efficiency of tell/ask
	       lock       %% lock the TermPrinter to serialize the output
		  {@printer clear}
		  {List.forAllInd CStores
		   proc{$ Index CStore}
		      Context = {Map CStore.context fun{$ _#Formula}
						       {VirtualString.toString
							{Translate.oz2Post Formula true Oz2PostDict}}
						    end}
		      ConStrings = {Map CStore.constraints fun{$ _#Con}
							      {VirtualString.toString
							       {Translate.oz2Post Con true Oz2PostDict}}
							   end}
		   in
		      %{Browse ConStrings}
		      {@printer nextLine(_)}
		      {@printer print("Constraints with context [")}
		      {ForAll Context proc{$ C}
					 {@printer printTerm(C)}
					 {@printer print(", ")}
				      end}
		      {@printer print("]")}
		      {@printer nextLine(_)}
		      {@printer print("=========================")}
		      {@printer nextLine(_)}
		      
		      {ForAll ConStrings
		       proc{$ S}
			  {@printer print("  ")}
			  {@printer printTerm(S)}
			  {@printer nextLine(_)}
		       end}
		   end}  % proc
	       end  % lock
	    end  % thread
	 end
      end
      
     
      %%-----------------
      %% tells a constraint to all context nodes wrt. Hyps
      %% fails, iff all telling attempts fail
      %%-----------------
      meth tellMetaConstraint(Con $)
	 case {Label Con} 
	 of nil then 'NIL'
	 [] 'not-occurs' then
	    metaConstraints <- Con|@metaConstraints
	    ConstName#const = Con.1       %% the Eigenvariable...
	    VarName#var = Con.2         %% the meta-variable...
	 in
	    if {Member ConstName#VarName @eigenVariables} then skip
	    else eigenVariables <- ConstName#VarName|@eigenVariables end
	    %{Show @eigenVariables}
	    FSCos, Trace("      {Cos tellMetaConstraint("#{Value.toVirtualString Con 20 100}#" _)}\n")
	    'T'
	 [] 'no-Eigenvars' then
	    {ForAll @eigenVariables proc{$ EVName#_}
				       FSCos, tellMetaConstraint('not-occurs'(EVName#const Con.1) _)
				    end}
	    %%FSCos,Trace("      {Cos tellMetaConstraint("#{Value.toVirtualString Con 20 100}#" _)}\n")
	    'T'
	 else
	    'T'
	 end
      end

      %%-----------------
      %% tries to tell a constraint to all context nodes wrt. Hyps
      %% fails, iff all telling attempts fail
      %%-----------------
      meth testConstraint(NodeName Hyps Con $)
	 if {IsConstraint Con @eigenVariables}
	    andthen {Not {IsGround Con}} then
	    if {Member tell#NodeName#Con @messages} then
	       'NIL'
	    else 
	       if @contextTree == nil then
		  {Show 'The constraint-solver has not been initialized yet.'}
		  'NIL'
	       else
		  OldConstraints = {@contextTree getConstraintsFor(Hyps $)} %the constraints that belong to Hyps
		  TellProcedure = proc{$ Sol}
				     SymbolicSolver = {New DomainSolver init}
				  in
				     {ForAll OldConstraints
				      proc{$ C}                       %% first propagate all present constraints...
					 {SymbolicSolver tellConstraint(C)}
				      end}
				     
				     {SymbolicSolver tellConstraint(Con)}   %% ... then the new constraint Con
				     {SymbolicSolver simplify}        %% simplify the constraint store
				  end
		  
		  NewCompSpace = {Space.new TellProcedure}
	       in
		  if {Space.ask NewCompSpace} == failed then
		     %Show 'Test for constraint '#Con#' failed!!!'}
		     'NIL'
		  else
		     %Show 'Test for constraint '#Con#' succeeded!!!'}
		     'T'
		  end %e
	       end % if contextTree
	    end % if {Member}
	 else
	    %{Show Con#'is not a valid constraint.'}
	    'NIL'
	 end %i
      end %m

      %%-----------------
      %% tries to tell a constraint to all context nodes wrt. Hyps
      %% fails, iff all telling attempts fail
      %%-----------------
      meth testAssumption(NodeName Hyps Con $)
	 if {IsConstraint Con @eigenVariables} then
	    if {Member tell#NodeName#Con @messages} then
	       'NIL'
	    else 
	       if @contextTree == nil then
		  {Show 'The constraint-solver has not been initialized yet.'}
		  'NIL'
	       else
		  OldConstraints = {@contextTree getConstraintsFor(Hyps $)} %the constraints that belong to Hyps
		  TellProcedure = proc{$ Sol}
				     SymbolicSolver = {New DomainSolver init}
				  in
				     {ForAll OldConstraints
				      proc{$ C}                       %% first propagate all present constraints...
					 {SymbolicSolver tellConstraint(C)}
				      end}
				     
				     {SymbolicSolver tellConstraint(Con)}   %% ... then the new constraint Con
				     {SymbolicSolver simplify}        %% simplify the constraint store
				  end
		  
		  NewCompSpace = {Space.new TellProcedure}
	       in
		  if {Space.ask NewCompSpace} == failed then
		     %{Show 'Test for constraint assumption '#Con#' failed!!!'}
		     'NIL'
		  else
		     %{Show 'Test for constraint assumption '#Con#' succeeded!!!'}
		     'T'
		  end %e
	       end % if contextTree
	    end % if {Member}
	 else
	    {Show Con#'is not a valid constraint assumption.'}
	    'NIL'
	 end %i
      end %m

      %%-----------------
      %% tells a constraint to all context nodes wrt. Hyps
      %% fails, iff all telling attempts fail
      %%-----------------
      meth tellConstraint(NodeName Hyps Con $)
	 if {IsConstraint Con @eigenVariables}
	    andthen {Not {IsGround Con}} then
	    if {Member tell#NodeName#Con @messages} then
	       'NIL'
	    else
	       %{Show constraint_has_not_been_here_before}
	       if @contextTree == nil then
		  {Show 'The constraint-solver has not been initialized yet.'}
		  'NIL'
	       else
		  OldConstraints = {@contextTree getConstraintsFor(Hyps $)} %the constraints that belong to Hyps
		  TellProcedure = proc{$ Sol}
				     SymbolicSolver = {New DomainSolver init}
				  in
				     {ForAll OldConstraints
				      proc{$ C}                       %% first propagate all present constraints...
					 %{Show tell#C}
					 {SymbolicSolver tellConstraint(C)}
				      end}
				     {Show telling#Con}
				     {SymbolicSolver tellConstraint(Con)}   %% ... then the new constraint Con
				     {Show simplify}
				     {SymbolicSolver simplify}        %% simplify the constraint store
				  end

		  NewCompSpace = {Space.new TellProcedure}
	       in
		  %{Show old#OldConstraints}
		  %{Show new#Con}
		  if {Space.ask NewCompSpace} == failed then 
		     {Show 'Telling constraint '#Con#' failed!!!'}
		     'NIL'
		  else 
		     {Show 'Telling constraint '#Con#' succeeded!!!'}
		     {System.gcDo}            %% do a garbage collection to avoid the oz bug
		     {@contextTree insertConstraint(NodeName Hyps Con _)}
		     messages <- (tell#NodeName#Con)|@messages
		     FSCos, PrintConstraints
		     FSCos, Trace("      {Cos tellConstraint('"#NodeName#"' "
				 #{Value.toVirtualString Hyps 20 100}#" "
				 #{Value.toVirtualString Con 20 100}#" _)}\n")
		     %{@contextTree show}
		     'T'
		  end %e
	       end % if contextTree
	    end % if {Member}
	 else
	    %{Show Con#'is not a valid constraint.'}
	    'NIL'
	 end %i
      end %m
      
      %%-----------------
      %% tells a constraint that is part of a proof assumption to all context nodes
      %% fails, iff all telling attempts fail
      %%-----------------
      meth tellAssumption(NodeName Hyps Con $)
	 if {IsConstraint Con @eigenVariables} then 
	    if {Member tell#NodeName#Con @messages} then
	       'NIL'
	    else
	       %{Show constraint_has_not_been_here_before}
	       if @contextTree == nil then
		  {Show 'The constraint-solver has not been initialized yet.'}
		  'NIL'
	       else
		  OldConstraints = {@contextTree getConstraintsFor(Hyps $)} %the constraints that belong to Hyps
		  TellProcedure = proc{$ Sol}
				     SymbolicSolver = {New DomainSolver init}
				  in
				     {ForAll OldConstraints
				      proc{$ C}                       %% first propagate all present constraints...
					 {Show tell#C}
					 {SymbolicSolver tellConstraint(C)}
				      end}
				     {Show telling#Con}
				     {SymbolicSolver tellConstraint(Con)}   %% ... then the new constraint Con
				     {Show simplify}
				     {SymbolicSolver simplify}        %% simplify the constraint store
				  end
		  
		  NewCompSpace = {Space.new TellProcedure}
	       in
		  %{Show old#OldConstraints}
		  %{Show new#Con}
		  if {Space.ask NewCompSpace} == failed then 
		     {Show 'Telling assumption '#Con#' failed!!!'}
		     'NIL'
		  else 
		     {Show 'Telling assumption '#Con#' succeeded!!!'}
		     {System.gcDo}            %% do a garbage collection to avoid the oz bug
		     {@contextTree insertConstraint(NodeName Hyps Con _)}
		     messages <- (tell#NodeName#Con)|@messages
		     FSCos, PrintConstraints
		     FSCos, Trace("      {Cos tellAssumption('"#NodeName#"' "
				 #{Value.toVirtualString Hyps 20 100}#" "
				 #{Value.toVirtualString Con 20 100}#" _)}\n")
		     %{@contextTree show}
		     'T'
		  end %e
	       end % if contextTree
	    end % if {Member}
	 else
	    %{Show Con#'is not a valid constraint.'}
	    'NIL'
	 end %i
      end %m
      
      %%-----------------
      %% asks whether the constraint Con is entailed in at least
      %% one context node node wrt. Hyps
      %%-----------------
      meth askConstraint(NodeName Hyps Con $)
	 if {IsPureConstraint Con @eigenVariables} then 
	    if @contextTree == nil then
	       {Show 'The constraint-solver has not been initialized yet.'}
	       'NIL'
	    else
	       OldConstraints = {@contextTree getConstraintsFor(Hyps $)} %the constraints that belong to Hyps
	       TellProcedure = proc{$ Sol}
				  SymbolicSolver = {New DomainSolver init}
			       in
				  {ForAll OldConstraints
				   proc{$ C}                       %% first propagate all present constraints...
				      %{Show tell#C}
				      {SymbolicSolver tellConstraint(C)}
				   end}
				  
				  {SymbolicSolver askConstraint(Con)}   %% ... then the new constraint Con
			       end
	       
	       NewCompSpace = {Space.new TellProcedure}
	    in
	       %{Show old#OldConstraints}
	       %{Show ask#Con}
	       if {Space.ask NewCompSpace} == failed then
		  %%FSCos,Trace("      {Cos askConstraint('"#NodeName#"' "
		  %%      #{Value.toVirtualString Hyps 20 100}#" "
		  %%    #{Value.toVirtualString Con 20 100}#" _)}\n")
		  {Show 'The constraint '#Con#' is NOT entailed!!!'}
		  'NIL'
	       else
		  if {Member 'ask'#Con @messages} then skip
		  else 
		     messages <- {Append @messages ['ask'#Con]}
		     FSCos, Trace("      {Cos askConstraint('"#NodeName#"' "
				 #{Value.toVirtualString Hyps 20 100}#" "
				 #{Value.toVirtualString Con 20 100}#" _)}\n")
		  end
		  {Show 'The constraint '#Con#' is entailed!'}
		  'T'
	       end %e
	    end
	 else
	    %{Show Con#'is not a pure constraint, so it cant be entailed!'}
	    'NIL'
	 end %i
      end %m
      
      %%-----------------
      %% This method is called by the servicemanager if
      %% the user tries to call an unknown method.
      %%-----------------
      meth otherwise(M $)
	 {Show 'Sorry, unknown method for FSCos'#M}
	 'NIL'
      end

   end %c
end %define

/*
proc {ExploreSpaces Session Problem Engine WeightFn}
   
   proc {SearchAux S}
      GenSym Consts Vars Formulas Logic
   in
      {ReadProblem Problem GenSym Consts Vars Logic Formulas}
      S = {New Engine create(GenSym Consts Vars Formulas Logic)}
      {S prove}
   end

   Logic
in
   %% get logic again from problem
   if {HasFeature Problem logic}
   then Logic = Problem.logic
   else Logic = classical
   end

   {Explorer.object delete(information all)}

   
   {Explorer.object add(information
			proc {$ _ S}
			   if {HasFeature S kernel}
			   then   {Browse S.kernel}
			   else
			      {Browse 'no kernel computed'}
			   end
			end
			label: 'Kernel'
		       )}

   {Explorer.object add(information
			proc {$ _ S}
			   {PrettyPrintModel S Problem
			    fun {$ _}
			       true
			    end}
			end
			label: 'Herbrand Base'
		       )}

   {Explorer.object add(information
			proc {$ _ S}
			   {PrettyPrintModel S Problem
			    fun {$ _#v(var: Val price: _)}
			       {Not {IsDet Val}} orelse Val > 0
			    end}
			end
			label: 'Positive Part'
		       )}

   {Explorer.object script(SearchAux WeightFn)}
end

proc{ShowStream S}
   case S
   of N#V|Ss then
      if {IsAtom N} then R in
	 if {IsDet V} then
	    {System.ShowInfo " "#N#": "#V}
	 else B in
	    {System.ShowInfo " "#N#": ["#{RI.getLowerBound V}#" ; "#{RI.getUpperBound V}#"]"}
	 end
%	    {System.ShowInfo {AtomToString {RI.isConstant V}}}
      else skip
      end
      {Show Stream Ss}
   [] nil then
      skip
   end
end

*/
