%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% FSDomain.oz ; This file is part of the OMEGA system
%
%   major updates: 3.12.1999,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA Project                                                        
%   AG Siekmann/FB Informatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)
% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
functor     
   %%  the file with all domain specific data an functions for FSCoS
   %%      jzimmer 3.12.99 

import
   System
   FS

   FSSolver('class': Solver)
   at 'x-ozlib://mathweb/cs/fscos/FSSolver.ozf'
   Share(car: Car
	 myEq: MyEq)
   at 'x-ozlib://mathweb/cs/share/Share.ozf'
   
export
   IsConstraint
   IsPureConstraint
   IsGround
   
   FuncSymbols
   RelSymbols
   Oz2PostDict
   
   GetVar
   GetConst
   
   GetLowerBounds
   GetUpperBounds
   GetEquations
   
   GetVariables
   GetVarString

   solver: Solver
   
define
   
   FuncSymbols = [intersect union set]
   RelSymbols  = ['=' '<=' '>=' '||' 'IN'] % something special!!!
   
   Oz2PostDict = {Record.toDictionary
		  functions(
			     '=':  '='
                            '<=': 'SUBSET'
                            '>=': 'SUPERSET'
                            '||': 'DISJOINT'
%'->': 'IMPLIES'  
			   )}
   
   fun{IsConstraint Con EVars}
      %{Show isconstr#Con#EVars}
      {IsRecord Con}                    %% a constraint is a feature tree...
      andthen (Rel = {Label Con}
	       Vars = {GetVariables Con}                  %% with certain variables ...
	       %Help = {Map EVars Car}
	       EigenVars = {List.flatten {Map Vars        %%... that have Eigenvariable conditions
					  fun{$ V}  
					     {GetEigenVars V EVars}
					  end}}  
	      in
		 case Rel
		 of 'AND' then        %% conjunct of constraints
		       {IsConstraint Con.1 EVars} andthen {IsConstraint Con.2 EVars}
		    
		 else
		    {Member Rel RelSymbols}   %% is it a valid relation?
		    andthen {Record.all Con fun{$ A}
					       {IsPureTerm A EigenVars}  %% the terms must be pure ...
					       orelse
					       {IsGroundTerm A EigenVars} %%... or ground alien terms ...
					    end}
		 end)
   end
   
   fun{IsPureConstraint Con EVars}
      {IsRecord Con} 
      andthen (Rel = {Label Con}
	      in
		 if Rel == 'AND' then
		    {IsPureConstraint Con.1 EVars} andthen {IsPureConstraint Con.2 EVars}
		 else
		    {Member {Label Con} RelSymbols}
		    andthen {Record.all Con fun{$ T} {IsPureTerm T EVars}
					    end}
		 end)
   end
   
   fun{IsGround Con}
      %{Show isconstr#Con#EVars}
      {IsRecord Con}                    %% a constraint is a feature tree...
      andthen {Record.all Con fun{$ A}
				 {IsGroundTerm A nil}  %% there must be at least one variable
			      end}
   end
   
   fun{IsConstrVar Con}
      case Con of
	 _#var then true
      else false end
   end
   
   fun{IsConstrConst Con EVars}
      %%{Show isconst#Con#EVars}
      case Con
      of Name#I then
	 (I=='const') andthen {Not {Member Name EVars}}
      else {IsNumber Con}
      end
   end

   fun{NoAlienTerm Term EVars}
      {IsConstrConst Term EVars} 
      orelse {IsRecord Term}
      andthen {Member {Label Term} ['+' '-' '*' '/']}
      andthen {Record.all Term fun{$ A} {NoAlienTerm A EVars} end}
   end
   
   fun{IsNumericTerm Term}
      ({IsNumber Term} orelse {IsConstrVar Term})
      orelse {IsRecord Term}
      andthen {Member {Label Term} FuncSymbols}
      andthen {Record.all Term fun{$ A} {IsNumericTerm A} end}
   end

   fun{IsTerm Term EVars}
      {IsConstrConst Term EVars}
      orelse {IsConstrVar Term}
      orelse {IsRecord Term}
      andthen case {Label Term}  
	      of '||' then {Record.all Term fun{$ A} {NoAlienTerm A EVars} end}
	      [] '^' then {Record.all Term fun{$ A} {IsNumber A} end}  %% special cases for  Ameier!!!!!!!!
	      [] 'SQRT' then {Record.all Term fun{$ A} {IsNumber A} end}  %% special cases for  Ameier!!!!!!!!
	      else {Record.all Term fun{$ A} {IsTerm A EVars} end}
	      end
   end        
   
   fun{IsGroundTerm Term EVars}
      %Show isground#Term#EVars}
      {IsConstrConst Term EVars}
      orelse {IsRecord Term}
      andthen ({Label Term} \= '#')
      andthen {Record.all Term fun{$ A}
				  {IsGroundTerm A EVars}
			       end}
   end
   
   fun{IsPureTerm Term EVars}
      %%{Show ispure#Term#EVars}
      {IsConstrConst Term EVars}
      orelse {IsConstrVar Term}
      orelse {IsRecord Term}
      andthen {Member {Label Term} FuncSymbols}
      andthen {Record.all Term fun{$ A}
				  {IsPureTerm A EVars}
				  orelse {IsList A}
			       end}
   end

   
   fun{GetEigenVars Var EVars}
      case EVars
      of nil then nil
      [] EV#V|EVr then
	 if V == Var then
	    EV|{GetEigenVars Var EVr}
	 else
	    {GetEigenVars Var EVr}
	 end
      else nil end
   end

   fun{GetLowerBounds Var Props}
      {Append
       {GetClosedLowerBounds Var Props}
       {GetOpenLowerBounds Var Props}}
   end
   
   fun{GetUpperBounds Var Props}
      {Append
       {GetClosedUpperBounds Var Props}
       {GetOpenUpperBounds Var Props}}
   end
   
   fun{GetClosedLowerBounds Var Props}
      case Props
      of nil then nil
      [] P|Pr then
	 if {Label P}== '<=' then 
	    L = P.1
	    R = P.2
	    Ref = P.ref
	 in
	    if {MyEq Var R} then
	       '>='#L#Ref|{GetClosedLowerBounds Var Pr}
	    else
	       {GetClosedLowerBounds Var Pr}
	    end
	 else {GetClosedLowerBounds Var Pr}
	 end
      else nil end
   end
   
   fun{GetOpenLowerBounds Var Props}
      case Props
      of nil then nil
      [] P|Pr then
	 case {Label P} 
	 of '>' then
	    L = P.1
	    R = P.2
	    Ref = P.ref
	 in
	    if {MyEq Var L} then
	       '>'#R#Ref|{GetOpenLowerBounds Var Pr} 
	    else
	       {GetOpenLowerBounds Var Pr}
	    end
	 else {GetOpenLowerBounds Var Pr}
	 end
      else nil end
   end
   
   fun{GetClosedUpperBounds Var Props}
      case Props
      of nil then nil
      [] P|Pr then
	 case {Label P}
	 of '<=' then
	    L = P.1
	    R = P.2
	    Ref = P.ref
	 in
	    if {MyEq Var L} then
	       '<='#R#Ref|{GetClosedUpperBounds Var Pr}
	    else
	       {GetClosedUpperBounds Var Pr}
	    end
	 else {GetClosedUpperBounds Var Pr}
	 end
      else nil end
   end

   fun{GetOpenUpperBounds Var Props}
      case Props
      of nil then nil
      [] P|Pr then
	 case {Label P}
	 of '>' then
	    L = P.1
	    R = P.2
	    Ref = P.ref
	 in
	    if {MyEq Var R} then
	       '<'#L#Ref|{GetOpenUpperBounds Var Pr} 
	    else
	       {GetOpenUpperBounds Var Pr}
	    end
	 else {GetOpenUpperBounds Var Pr}
	 end
      else nil end
   end
      

   fun{GetEquations Var Props}
      case Props
      of nil then nil
      [] P|Pr then
	 case {Label P} 
	 of '=' then
	    L = P.1
	    R = P.2
	 in
	    if {MyEq Var L} then
	       '='#R|{GetEquations Var Pr}
	    elseif {MyEq Var R} then
	       '='#L|{GetEquations Var Pr}
	    else
	       {GetEquations Var Pr}
	    end
	 else {GetEquations Var Pr} end
      else nil end
   end

   
   fun {GetVar Label Var Found Stream}
%      {Show getvar#Label#Stream} 
      case Stream
      of L#V|Xs then
	 if L == Label then
	    if {IsDet Found} then skip
	    else Found = true end
	    Var = V
	    Stream
	 else
	    L#V|{GetVar Label Var Found Xs}
	 end
      [] nil then
	 Var = {FS.var.decl}
	 [Label#Var]
      end
   end
   
   
   fun {GetConst Label Const Stream}
      case Stream
      of L#V|Xs then
	 if {System.eq L Label} then
	    Const = V
	    Stream
	 else
	    L#V|{GetConst Label Const Xs}
	 end
      [] nil then
	 Const = {FS.var.decl} 
	 %{RI.toConstant Const}
	 [Label#Const]
      end
   end
   
   fun{GetVariables Con}
      if {IsRecord Con} then
	 if {IsConstrVar Con} then
	    [{Car Con}]
	 else
	    ConList = {Record.toList Con}
	 in
	    {Flatten {Map ConList GetVariables}}
	 end
      else nil end	 
   end

   fun{GetVarString Var}
      N#RIVar = Var
   in
      if {IsDet RIVar} then
	 {VirtualString.toString N#" "#RIVar}
      else
	 {VirtualString.toString
	  N#" ["#{FS.getLowerBound RIVar}#" ; "#{FS.getUpperBound RIVar}#"]"}
      end
   end

   
end %define
