%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% SolverClass.oz ; This file is part of the OMEGA system
%
%   major updates: 26.10.1999,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA Project                                                        
%   AG Siekmann/FB Informatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)
% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   %% The class of symbolic constraint solvers
   %%      jzimmer 18.1.99 

import
   %Browser(browse: Browse)
   System(show: Show eq gcDo)

   Reflect at 'x-oz://contrib/Reflect.ozf'
   %RealInterval(rI: FS)
   %FD
   FS
   FCP_FS at 'x-oz://contrib/modFSP.fcp.ozf'

   FSSimpRules(simpRules: Rules)  at 'x-ozlib://mathweb/cs/fscos/FSSimpRules.ozf'
   %Maple(simplify: MapleSimplify lineq2Maple: Lineq2Maple)
   FSDomain(funcSymbols:         CosieFuncs
	    relSymbols:          CosieRels
	    getVar:              GetVar              %i
	    getConst:            GetConst            %i
	    getLowerBounds:      GetLowerBounds
	    getUpperBounds:      GetUpperBounds
	    getEquations:        GetEquations
	    %getVarString:        GetVarString
	   ) 
   at 'x-ozlib://mathweb/cs/fscos/FSDomain.ozf'
   Share(
	 setDifference:       SetDifference
	 getSolution:         GetSolution
	 constructTerm:       ConstructTerm
	 matchList:           MatchList
	 applyMatching:       ApplyMatching
	 removeProps:         RemoveProps         % independent
	 removeVar:           RemoveVar           % inde
	 findName:            FindName
	 findVar:             FindVar
	 killProps:           KillProps
	 applySubst:          ApplySubst
	)
   at 'x-ozlib://mathweb/cs/share/Share.ozf'
   
export
   'class': SymbolicSolver   %The class of all symbolic FS-constraint-solvers
   
   
define
   
   {Wait Reflect.propReflect}  %% ... for propagator_reference_id :-(
   
   %%===============================================
   %%===============================================
   %%  The SolverClass
   %%===============================================
   class SymbolicSolver
      from BaseObject
      prop
	 locking

      feat
	 App
	 
      attr
	 probVars           %% the problem variables given by the proof planner
	 undetVars          %% the still symbolically undetermined variables
	 nestedVars         %% the nested variables introduced during propagation
	 alienVars          %% the new variables representing alien terms (SOLEX...)
	 probConsts         %% the problem constants given by the proof planner
	 eigenVars          %% the Eigenvariable condtitions for the search procedure
	 alienReplacements  %% a dictionary with the alien term replacements
	 %solution           %% 
	 propList           %% the list with the current propagators
	 constraints        %% the constraints.... terms
	 
	 applied_rules      %% the simplification rules that have been applied
	 nameCounter        %% a counter for new variable names
	 
      %%------------------
      %% Initialize the constraint solver.
      %%------------------
      meth init()
	 probVars <- nil
	 undetVars <- nil
	 nestedVars <- nil
	 alienVars <- nil
	 probConsts <- nil
	 eigenVars <- nil
	 
	 alienReplacements <- {Dictionary.new}
	 propList <- nil
	 applied_rules <- nil
	 
	 constraints <- con('<=': nil
			    '=': nil           
			    '>=': nil
			    '||': nil
			    'IN': nil)
	 nameCounter <- 0
	 %{ShowInfo 'Hello, I`m a symbolic solver'}
      end
      
      
      %%------------------
      %% Propagate the constraint Con and simplify the constraint store.
      %%------------------
      meth tellConstraint(Con)
	 %{Show vars#@probVars}
	 %{Show @constraints.{Label Con}}
	 if {Member Con @constraints.{Label Con}}
	 then skip
	 else
	    %%SymbolicSolver, showProps()
	    
	    NewCon = SymbolicSolver, Propagate(Con $)  %% Con with propagator references
	 in
	    %{Show @constraints}
	    %{Show newconstraint#NewCon}
	    %SymbolicSolver, SimplifyProps()
	    %%SymbolicSolver, showProps()
	    SymbolicSolver,showVars
	    
	    SymbolicSolver, AddConstraint(NewCon)
	 end
      end
      
      %%------------------
      %% Propagate the constraint Con and simplify the constraint store.
      %%------------------
      meth tellAssumption(Con)
	 %{Show vars#@probVars}
	 if {Member Con @constraints} then skip
	 else
	    NewCon = SymbolicSolver, Propagate(Con $)
	 in
	    SymbolicSolver, AddConstraint(NewCon)
	 end
      end
      
      %%------------------
      %% Check for entailment of the constraint Con.
      %%------------------
      meth askConstraint(Con)
	 SymbolicSolver, Ask(Con)
      end
      
      %%------------------
      %% Set the Eigenvariable conditions
      %%------------------
      meth setEigenVars(EVars)
	 eigenVars <- EVars
      end
      
      %%------------------
      %% Add a new constraint to the list of @constraints
      %%------------------
      meth AddConstraint(Con)
	 Rel = {Label Con}
	 NewCons = Con|@constraints.Rel
      in
	 %{Show rel#Rel#NewCons}
	 case @constraints
	 of con('<=':Subset '=':Eq '>=':Superset '||':Disj 'IN':In) then
	    case Rel
	    of '<=' then constraints <- con('<=':NewCons '=':Eq      '>=':Superset '||':Disj    'IN':In)
	    [] '='  then constraints <- con('<=':Subset  '=':NewCons '>=':Superset '||':Disj    'IN':In)
	    [] '>=' then constraints <- con('<=':Subset  '=':Eq      '>=':NewCons  '||':Disj    'IN':In)
	    [] '||' then constraints <- con('<=':Subset  '=':Eq      '>=':Superset '||':NewCons 'IN':In)
	    [] 'IN' then constraints <- con('<=':Subset  '=':Eq      '>=':Superset '||':Disj    'IN':NewCons)
	    else
	       {Show 'Unknown relation'#Rel}
	       fail
	    end
	 else
	    {Show 'Something went wrong with the @constraints feature'}
	 end
      end
      
      %%------------------
      %% Simplify the constraint store
      %%------------------
      meth simplify
	 skip
      end

      /*
      %%------------------
      %% Get a list of string representations of all constraints in the constraint store
      %% wrt. the problem variables.
      %%------------------
      meth getConstraintStrings($)
	 {Map @probVars fun{$ V} {GetConstraintString V @propList @alienReplacements} end}
      end
      
      %%------------------
      %% Get a list of string representations of the domains of all problem variables
      %%------------------
      meth getVarStrings($)
	 {Map @probVars GetVarString}
      end
      */
      meth getConstraints($)
	 @constraints
      end

      %%------------------
      %% Gives an association list with the instantiations for the
      %% determined variables. The list has the form ((Var1 . Term1) ... (Varn . Termn))
      %%------------------
      meth getDetVars($)
	 DetVars = {SetDifference @probVars @undetVars}
      in
	 {Map DetVars fun{$ Name#_}
			 Equations = {GetEquations Name#var @propList}
		      in case Equations
			 of nil then "TRUE"
			 [] '='#T|_ then
			    Res = {ConstructTerm T @propList @propList @alienReplacements}
			 in
			    {Show res1#Res}
			    Name#Res
			 else nil end
		      end}
      end
			    
      %%------------------
      %% Try to compute a solution for the problem variables.
      %%------------------
      meth getSolution($)
	 if @undetVars == nil then
	    VarNames = {Map @probVars fun{$ V}
					 Name#_ = V
				      in
					 Name#var
				      end}
	    Props = {Map @propList fun{$ P}
				      {Record.subtract P ref}
				   end}
	    Solution =  {VirtualString.toString
			 {GetSolution
			  VarNames
			  Props
			  Props
			  @alienReplacements
			  CosieFuncs}}
	 in
	    Solution
	 else "failure" end
      end
      
      
      
      %%------------------
      %% Get the state of the computation space, the solver is sited in.
      %%------------------
      meth ask($)
	 if {All @probVars fun{$ Pair}
			      _#Var = Pair
			   in
			      SymbolicSolver, IsDetermined(Var $)
			   end}
	 then 
	    succeeded
	 else
	    undetermined
	 end
      end
      
      %%------------------
      %% The distribution strategy for symbolic constraint solving.
      %%------------------
      meth distribute(Solution)
	 {Show  distribution}
	 {System.gcDo}
	 choice
	    Var = SymbolicSolver, FirstUndetVar($)
	 in
	    SymbolicSolver, showVars
	    if Var == nil then
	       SymbolicSolver, getSolution(Solution)
	    else 
	       LBounds = SymbolicSolver, getLowerBounds(Var $)
	       UBounds = SymbolicSolver, getUpperBounds(Var $)
	    in
	       {Show var#Var#LBounds#UBounds}
	       case LBounds
	       of nil then
		  {Show case1}
		  case UBounds                        
		  of nil then                %% there aren't any bounds to the variable
		     SymbolicSolver, tellConstraint('='(Var 1.0))
		     SymbolicSolver, substitute(Var by:1.0)
		     SymbolicSolver, distribute(Solution)
		  [] ('<='#U#URef)|_ then                %% there is a closed upper bound, then try
		     {Show case1_1}
		     choice                                %%  L < Var <= U)  ==>   Var = U or Var < U
			propList <- {RemoveProps @propList [URef]}
			SymbolicSolver, tellConstraint( '='(Var U) )
			SymbolicSolver, substitute(Var)
			SymbolicSolver, distribute(Solution)
		     []
			propList <- {RemoveProps @propList [URef]}
			SymbolicSolver, tellConstraint('<'(Var U) )
			SymbolicSolver, SimplifyProps()
			SymbolicSolver, distribute(Solution)
		     end
		     
		  [] ('<'#U#URef)|_ then             %% there are only _open_ lower and upper bounds
		     MiddleTerm = if {IsNumber U} then %% calculate a term smaller than U
				     case U
				     of 0.0 then 1.0    %% still a BIG HACK!!!!!!!!!!!!
				     else               %%=============================
					if (U < 0.0) then '*'(U 2.0) 
					else '/'(U 2.0) end
				     end
				  else
				     '/'(U 2.0)
				  end
		  in
		     {Show case1_2}
		     {Show middleTerm#Var#MiddleTerm}
		     choice
			propList <- {RemoveProps @propList [URef]}
			SymbolicSolver, tellConstraint('='(Var MiddleTerm))
			SymbolicSolver, substitute(Var)
			SymbolicSolver, distribute(Solution)
		     []
			propList <- {RemoveProps @propList [URef]}
			SymbolicSolver, tellConstraint('<'(Var MiddleTerm))
			SymbolicSolver, SimplifyProps()
			SymbolicSolver, distribute(Solution)
		     []
			SymbolicSolver, tellConstraint('>'(Var MiddleTerm))
			SymbolicSolver, SimplifyProps()
			SymbolicSolver, distribute(Solution)
		     end
		  else
		     {Show 'The List of upper bounds is neither empty, nor anything else :-('}
		     fail
		  end
		  
		  
	       %%=== there are lower bounds....            
		  
	       [] '>='#L#Ref|_ then                 %%     first try a closed lower bound, if it exits
		  {Show case2}
		  choice                              %%  L <= Var <= U)  ==>   L = Var or L < Var
	     propList <- {RemoveProps @propList [Ref]}
		     SymbolicSolver, tellConstraint( '='(Var L) )
		     SymbolicSolver, substitute(Var)
		     SymbolicSolver, distribute(Solution)
		  []
		     propList <- {RemoveProps @propList [Ref]}
		     SymbolicSolver, tellConstraint( '>'(Var L) )
		     SymbolicSolver, SimplifyProps()
		     SymbolicSolver, distribute(Solution)
		  end
		  
	       [] ('>'#L#LRef)|_ then                %% there are only open lower bounds left...
		  {Show case3}
		  case UBounds                        
		  of nil then                          %% ... and no upper bounds
		     MiddleTerm = if {IsNumber L} then %% calculate a term greater than L
				     case L
				     of 0.0 then 1.0    %% still a BIG HACK!!!!!!!!!!!!
				     else               %%=============================
					if (L < 0.0) then '/'(L 2.0) 
					else '*'(L 2.0) end
				     end
				  else
				     '*'(L 2.0)
				  end
		  in
		     {Show case3_1}
		     choice
			propList <- {RemoveProps @propList [LRef]}
			{Show telling#Var#'='#MiddleTerm}
			SymbolicSolver, tellConstraint('='(Var MiddleTerm))
			SymbolicSolver, substitute(Var)
			SymbolicSolver, distribute(Solution)
		     []
			{Show telling#Var#'<'#MiddleTerm}
			SymbolicSolver, tellConstraint('<'(Var MiddleTerm))
			SymbolicSolver, SimplifyProps()
			SymbolicSolver, distribute(Solution)
		     []
			propList <- {RemoveProps @propList [LRef]}
			SymbolicSolver, tellConstraint('>'(Var MiddleTerm))
			SymbolicSolver, SimplifyProps()
			SymbolicSolver, distribute(Solution)
		     end
		     
		  [] ('<='#U#URef)|_ then                %% there is a closed upper bound, then try
		     {Show case3_2}
		     choice                                %%  L < Var <= U)  ==>   Var = U or Var < U
			propList <- {RemoveProps @propList [URef]}
			SymbolicSolver, tellConstraint( '='(Var U) )
			SymbolicSolver, substitute(Var)
			SymbolicSolver, distribute(Solution)
		     []
			propList <- {RemoveProps @propList [URef]}
			SymbolicSolver, tellConstraint( '<'(Var U) )
			SymbolicSolver, SimplifyProps()
			SymbolicSolver, distribute(Solution)
		     end
		     
		  [] ('<'#U#URef)|_ then             %% there are only _open_ lower and upper bounds
		     MiddleTerm = '/'('+'(U L) 2.0)   %%  then compute the middle of the interval and 
		     		                      %  create a disjunction 
%M = {MapleSimplify MiddleTerm} %% NO I/O from local space!!!
		  in
		     %{Browse mapleTerm#M} %{VirtualString.toString M}}
		     {Show case3_3}
		     {Show middleTerm#Var#MiddleTerm}
		     choice
			propList <- {RemoveProps @propList [LRef URef]}   %% remove both old bounds
			{Show telling#Var#'='#MiddleTerm}
			SymbolicSolver, tellConstraint('='(Var MiddleTerm))
			SymbolicSolver, substitute(Var)
			SymbolicSolver, distribute(Solution)
		     []
			propList <- {RemoveProps @propList [URef]}        %% keep the lower bound, delete the upper
			{Show telling#Var#'<'#MiddleTerm}
			SymbolicSolver, tellConstraint('<'(Var MiddleTerm))
			SymbolicSolver, SimplifyProps()
			SymbolicSolver, distribute(Solution)
		     []
			propList <- {RemoveProps @propList [LRef]}        %% keep the upper bound, delete the lower
			{Show telling#Var#'>'#MiddleTerm}
			SymbolicSolver, tellConstraint('>'(Var MiddleTerm))
			SymbolicSolver, SimplifyProps()
			SymbolicSolver, distribute(Solution)
		     end
		     %M = {MapleSimplify MiddleTerm} %% NO I/O from local space!!!
		  else
		     {Show 'The List of upper bounds is neither empty, nor anything else :-('}
		     fail
		  end
	       else  {Show 'The List of lower bounds is neither empty, nor anything else :-('}
		  fail
	       end
	    end
	 end
      end
      
      meth substitute(Var by:Term <= _)
	 Name#_ = Var
	 Subst = {NewDictionary}
      in
	 if {IsDet Term} then 
	    {Dictionary.put Subst Name Term}
	    %%constraints <- {ApplySubst Subst @constraints}
	    propList <- {ApplySubst Subst @propList}
	    SymbolicSolver, showProps
	 else 
	    Equs = {GetEquations Var @propList}
	 in
	    if Equs == nil then skip
	    else '='#Term|_ = Equs in
	       {Dictionary.put Subst Name Term}
	       %%constraints <- {ApplySubst Subst @constraints}
	       propList <- {ApplySubst Subst @propList}
	       SymbolicSolver, showProps
	    end
	 end
      end
	 
      
      %%------------------
      %% Get the lower bounds for a variable.
      %%------------------
      meth getLowerBounds(Var $)
	 if Var== nil then nil
	 else
	    {GetLowerBounds Var @propList}
	 end
      end
      
      %%------------------
      %% Get the upper bounds for a variable.
      %%------------------
      meth getUpperBounds(Var $)
	 if Var== nil then nil
	 else
	    {GetUpperBounds Var @propList}
	 end
      end
      
      %%------------------
      %% Show all problem variables.
      %%------------------
      meth showVars
	 {Show 'Variables'}
	 {ForAll @probVars
	  proc{$ FSVar}
	     {Show FSVar}
	  end}
      end
      
      %%------------------
      %% Show all propagators.
      %%------------------
      meth showProps
	 {Show 'Propagators'}
	 {ForAll @propList proc{$ P}
			      SymbProp = SymbolicSolver, SymbolicArgs(P $)
			   in
			      {Show '  prop:'#{Record.subtract SymbProp ref}}
			   end}
	 {Show 'applied_rules '#@applied_rules}	 
      end

      
      
      %%------------------
      %% Reflect the current constraint store.
      %%------------------
      meth reflect($)
	 SymbolicSolver, showProps()
	 'true'
      end
      
      
      %%%%================================================
      %%%% Internal methods:
      %%%%================================================
      
      %%------------------
      %% Check whether a variable Var is determined by an equality propagator.
      %%------------------
      meth IsDetermined(Var $)
	 {Some @propList fun{$ Prop}
			    {And ({Label Prop} == '=') 
			     {Or {System.eq Prop.1 Var}
			      {System.eq Prop.2 Var}}}
			 end}
      end

      %%------------------
      %% Get the first undetermined variable.
      %%------------------
      meth FirstUndetVar($)
	 {Show 'still undet:'#{Map @undetVars fun{$ N#_} N end}}
	 case @undetVars
	 of nil then nil
	 [] N#_|_ then
	    N#var
	 else nil end 
      end
      
      %%------------------
      %% Try to match a constraint rule with the current set of propagators.
      %%------------------
      meth Match(Rule)
	 Matching = {MatchList Rule.prec @propList nil {NewDictionary}}   %% first try to match the
	                                                                 %% precondition
      in
	 if Matching == nil then skip 
	 else 
	    %{Show ' found matching'#{Label Rule}} %#{Dictionary.entries Matching}}
	    DelProps = {ApplyMatching Rule.del Matching}
	 in
	    {Show discardProps(DelProps)}
	    {KillProps DelProps}
	    propList <- {RemoveProps @propList DelProps}
	    	    
	    if Rule.add == failure then
	       fail
	    else
	       {ForAll {ApplyMatching Rule.add Matching}
		proc{$ PS}
		   SymbProp = SymbolicSolver, SymbolicArgs(PS $)
		in
		   if {Member {Label SymbProp} CosieFuncs} then
		      SymbolicSolver, PropagateTerm(SymbProp _ _)
		   elseif {Member {Label SymbProp} CosieRels} then 
		      NewCon = SymbolicSolver, Propagate(SymbProp $)
		   in
		      SymbolicSolver, AddConstraint(NewCon)
		   end
		end}
	    end
	    applied_rules <- {Label Rule}|@applied_rules
	 end
      end
      
      %%------------------
      %% Simplify the list of propagators by applying constrain rules, if possible.
      %%------------------
      meth SimplifyProps
	 AlreadyApplied = {Length @applied_rules}
      in
	 {Wait Rules}
	 
	 {ForAll Rules
	  proc{$ Rule}
	     SymbolicSolver, Match(Rule)
	  end}
	 %{self showProps}
	 
	 if {Length @applied_rules} > AlreadyApplied then
	    SymbolicSolver, SimplifyProps()
	 else
	    skip %SymbolicSolver, showProps()
	 end
      end
      
      
      
      %%------------------
      %% Propagates a lineq-constraint, adds new variables and constants and
      %% creates an entry in @propList for each propagator
      %%------------------
      meth Propagate(Con $)
	 %{Show con#Con}
	 case Con
	    %% The relation propagators
	 of 'AND'(C1 C2) then                %% &
	    NewC1 = SymbolicSolver, Propagate(C1 $)
	    NewC2 = SymbolicSolver, Propagate(C2 $)
	 in
	    'AND'(NewC1 NewC2 nil)
	    
	 [] '<='(L R) then                    %% <
	    '>='(NewR NewL Ref) = SymbolicSolver, Propagate('>='(R L) $)
	 in
	    '<='(NewL NewR Ref)
	    
	 [] '>='(L R) then                   %% <=
	    X Y
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	    Ref = {FCP_FS.'`Subsume`' Y X}   
	 in
	    {Show '>='#X#Y}
	    SymbolicSolver, NewProp('>='(X Y ref:Ref))
	    '>='(NewL NewR Ref)
	    
	 [] '||'(L R) then                   %% <=
	    X Y
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	    Ref = {FCP_FS.'`Disjoint`' X Y}   
	 in
	    SymbolicSolver, NewProp('||'(X Y ref:Ref))
	    '||'(NewL NewR Ref)
	    
	    %% equations are treated differently ...
	 [] '='(L R) then
	    X Y
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	 in
	    %{Show 'equality2'#X#Y}
	    X = Y                 % really propagate the equation!
	    case L
	    of LName#_ then    % equation 'determinates' a variable!
	       undetVars <- {RemoveVar LName @undetVars} %.. so remove it from the List of undetermined Vars
	       case R
	       of RName#_ then
		  undetVars <- {RemoveVar RName @undetVars}
		  SymbolicSolver, NewProp('='(L R ref:nil) )  % no propagator reference for '='
		  '='(L R nil)
	       else
		  SymbolicSolver, NewProp('='(L Y ref:nil) )
		  '='(L NewR nil)
	       end
	    else
	       SymbolicSolver, NewProp('='(X Y ref:nil) )
	       '='(NewL NewR nil)
	    end
	    
	 [] 'IN'(L Set) then    %% At the moment _natural numbers_ must only be > 0 !!!
	    X Y
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    SymbolicSolver, PropagateTerm(Set Y _)
	    Ref = {FCP_FS.'`Include`' X Y}
	 in
	    'IN'(NewL Sort Ref)
	 else
	    {Show 'Cannot propagate constraint'#Con}
	    fail
	 end
      end %method
      
      
      %%------------------
      %% Propagates a term and adds a propagator reference to
      %% each subterm, i.e., returns terms of the form '+'(A B  Ref)
      %% The method creates an entry in @propList for each propagator
      %% The parameter FSVar must be unbound. It is bound to the RI-variable
      %% representing the term Term.
      %%------------------
      meth PropagateTerm(Term FSVar $)
	 %{Show term#Term}
	 case Term 
	    %% more arithmetic propagators

	 of union(L R) then
	    X Y Ref
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	 in
	    FSVar = SymbolicSolver, NewNestedVar($)
	    Ref = {FCP_FS.'`Union`' X Y FSVar}
	    SymbolicSolver, NewProp(union(X Y FSVar ref:Ref))
	    union(NewL NewR Ref)
	    
	 [] intersect(L R) then
	    X Y Ref
	    NewL = SymbolicSolver, PropagateTerm(L X $)
	    NewR = SymbolicSolver, PropagateTerm(R Y $)
	 in 
	    FSVar = SymbolicSolver, NewNestedVar($)
	    Ref = {FCP_FS.'`Intersect`' FSVar Y X}
	    SymbolicSolver, NewProp(intersect(X Y FSVar ref:Ref) )
	    intersect(NewL NewR Ref)
	    
	 [] set(Spec) then
	    FSVar = {FS.value.make Spec}
	    {Show set(FSVar)}
	    set(Spec)
	    
	    %% variables, symbolic constants and numbers 
	 [] N#Type then Dummy Found in
	    if {IsAtom N} then
	       case Type
	       of var then 
		  %{Show varcase1#N}
		  probVars <- {GetVar N FSVar Found @probVars}
		  %{Show new(@probVars Found FSVar)}
		  if {IsDet Found} then
		     N#Type
		  else
		     %{Show varcase2#N}
		     undetVars <- {GetVar N Dummy Found @undetVars}
		     N#Type
		  end
	       [] nested then
		  %{Show nestcase#N}
		  nestedVars <- {GetVar N FSVar Found @nestedVars}
		  N#Type

	       [] alien then
		  alienVars <- {GetVar N FSVar Found @alienVars}
		  N#Type
		  
	       [] const then
		  %{Show constcase#N}
		  probConsts <- {GetConst N FSVar @probConsts}
		  N#Type
	       else
		  %{Show nestcase#N}
		  nestedVars <- {GetVar N FSVar Found @nestedVars}
		  N#Type
	       end
	    else
	       FSVar = {FS.value.single 1}
	       Term
	    end
	 else
	    if {IsAtom Term} then
	       nestedVars <- {GetVar Term FSVar _ @nestedVars}
	       Term
	    elseif {IsInt Term} then
	       FSVar = Term
	    else
	       %%Term
	       Name
	    in
	       Name#FSVar = SymbolicSolver, NewAlienVar($)
	       %{Show 'Replacing the alien term '#Term#' by '#Name}
	       {Dictionary.put @alienReplacements Name Term}
	       Name#alien
	    end
	 end
      end % meod
      
      %%------------------
      %% Test whether the constraint Con ist entailed by the constraint store.
      %%------------------
      meth Ask(Con)
	 %{Show 'askcon'#Con}
	 case Con
	    %% The relation propagators
	 of 'AND'(C1 C2) then
	    SymbolicSolver, Ask(C1 _)
	    SymbolicSolver, Ask(C2 _)
	    
	 [] '<'(L R) then 
	    SymbolicSolver, Ask('>'(R L))
	    
	 [] '<='(L R) then
	    X = SymbolicSolver, PropagateTerm(L $ _)
	    Y = SymbolicSolver, PropagateTerm(R $ _)
	    XU = if {IsDet X} then X else {FS.getUpperBound X} end
	    YL = if {IsDet Y} then Y else {FS.getLowerBound Y} end
	 in
	    %%{Show xu#yl#XU#YL#(XU =< YL)#(XU>YL)}
	    if (XU =< YL) then skip
	    else fail end
	    
	    %% equality
	 [] '='(L R) then
	    X = SymbolicSolver, PropagateTerm(L $ _)
	    Y = SymbolicSolver, PropagateTerm(R $ _)
	 in
	    if {System.eq X Y} then skip
	    else fail end
	    
	 [] '>='(L R) then
	    SymbolicSolver, Ask('<='(R L))

	 [] '>'(L R) then
	    X = SymbolicSolver, PropagateTerm(L $ _)
	    Y = SymbolicSolver, PropagateTerm(R $ _)	    
	    YU = if {IsDet Y} then Y else {FS.getUpperBound Y} end
	    XL = if {IsDet X} then X else {FS.getLowerBound X} end
	 in
	    {Show X#Y#XL#YU}
	    if XL > YU then skip
	    else fail end
	    
	 else fail end
      end
      
	    
      %%------------------
      %% Get the names for a list of variables.
      %%------------------
      meth SymbolicArgs(Prop $)
	 %{Show symbolic#Prop}
	 if {IsDet Prop} then
	    if {IsRecord Prop} then
	       {Record.map Prop fun{$ A}
				   SymbolicSolver, SymbolicArgs(A $)
				end}
	    else 
	       Prop
	    end
	 else
	    SymbolicSolver, GetFSVarName(Prop $)
	 end
      end

      %%------------------
      %% Create a new name for a nested variable.
      %%------------------
      meth NewName(Prefix $)
	 nameCounter <- @nameCounter + 1
	 {String.toAtom {VirtualString.toString Prefix#@nameCounter}}
      end
      
      %%------------------
      %% Create a new nested RI variable.
      %%------------------
      meth NewNestedVar($)
	 V = {FS.var.decl}
      in
	 nestedVars <- {Append [SymbolicSolver, NewName("Nested" $)#V]
			@nestedVars}
	 V    %% return the RI-var
      end
      
      %%------------------
      %% Create a new RI variable as a placeholder for an alien term.
      %%------------------
      meth NewAlienVar($)
	 Name = SymbolicSolver, NewName("Alien" $)
	 Var = {FS.var.decl}
      in
	 alienVars <- {Append [Name#Var]
		       @alienVars}
	 Name#Var  %% return the NewName _and_ the RI-var
      end
      
      %%------------------
      %% Get the name of a RI variable.
      %%------------------
      meth GetFSVarName(Var $)
	 if {IsDet Var} then
	    Var
	 else
	    NestedResult = {FindName Var @nestedVars} in
	    if NestedResult == nil then 
	       VarResult = {FindName Var @probVars} in
	       if VarResult == nil then 
		  ConstResult = {FindName Var @probConsts} in
		  if ConstResult == nil then
		     AlienResult = {FindName Var @alienVars} in
		     if AlienResult == nil then
			Var
		     else AlienResult#alien end
		  else ConstResult#const end
	       else VarResult#var      end
	    else NestedResult#nested end
	 end
      end
      
      %%------------------
      %% Get RI variable for a given name or alien term
      %%------------------
      meth GetFSVar(Name $)
	 NestedResult = {FindVar Name @nestedVars} in
	 if NestedResult == nil then 
	    VarResult = {FindVar Name @probVars} in
	    if VarResult == nil then 
	       ConstResult = {FindVar Name @probConsts} in
	       if ConstResult == nil then
		  AlienResult = {FindVar Name @alienVars} in
		  if AlienResult == nil then
		     Name
		  else AlienResult end
	       else ConstResult end
	    else VarResult    end
	 else NestedResult end
      end
      

      %%------------------
      %% Insert a new entry to the list of propagators.
      %%------------------
      meth NewProp(Con)
	 %{Show newprop#Con}
	 propList <-  SymbolicSolver, SymbolicArgs(Con $)|@propList
      end
      
   end %class
end %define

/*

   fun {Replace X Y Ps}
      case Ps
      of nil then nil
      [] P|Pr then
	 if {System.eq P X} then
	    if Y == nil then
	       {Replace X Y Pr}
	    else 
	       Y|{Replace X Y Pr}
	    end
	 else P|{Replace X Y Pr}
	 end
      else nil end
   end
   

      meth TestCycle(PName Start)
	 Props Rest
      in
	 {List.partition @propList fun{$ P}
				      P.name == PName
				   end
	  Props Rest}
	 if {CheckCycle Start Start Props Props} then fail
	 else skip end
      end


      fun {CheckCycle Start Act Ps AllPs}
      %{Show 'Check'#Act}
      case Ps
      of nil then false
      [] P|Pr then
	 case P.params
	 of [X Y] then 
	    if {System.eq Y Start} then true
	    elseif {System.eq Act X} then 
	       {CheckCycle Start Y AllPs AllPs}
	    else
	       {CheckCycle Start Act Pr AllPs}
	    end
	 else
	    {CheckCycle Start Act Pr AllPs}
	 end
      else
	 false
      end
   end
   
*/
