%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% ContextTreeClass.oz ; This file is part of the OMEGA system
%
%   major updates: 26.10.1999,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA Project                                                        
%   AG Siekmann/FB Informatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)
% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   %% The class of context trees
   %%      jzimmer 22.9.99 
import
   System(show: Show )
   Browser
   
   NodeClass('class': ContextNode)
   at 'x-ozlib://mathweb/cs/share/NodeClass.ozf'
   Share(cdr: Cdr
	 setDifference: SetDifference
        )
   at 'x-ozlib://mathweb/cs/share/Share.ozf'
   
export
   'class': ContextTree
   
prepare
   /*
   fun {Flatten Ls}
      {FoldL Ls
       fun {$ L R}
	  if {IsDet R} then
	     case R of nil then L
	     [] T|Rs then {Append L {Flatten R}}
	     else [R] end 
	  else
	     R|L
	  end
       end nil}
   end
   */
   %%-----------------
   %% Recursive version of getLeaf.
   %%-----------------
   fun{GetLeaf Node}
      Cs = {Node children($)}
   in
      case Cs of nil then Node
      [] C|_ then 
	 {GetLeaf C}
      else Node end
   end
   
   fun{SubSet Xs Ys}
      {All Xs fun{$ X}
		 {Member X Ys}
	      end}
   end
   
   %%-----------------
   %% Recursive version of getAllLeaves
   %%-----------------
   fun{GetAllLeaves Node}
      Cs = {Node children($)}
   in
      case Cs of nil then [Node]
      [] _|_ then 
	 {Flatten {Map Cs GetAllLeaves}}
      else [Node] end
   end

   %%-----------------
   %% Recursive version of getAllLeafCStores
   %%-----------------
   fun{GetAllLeafCStores Node Cons Ass}
      Cs = {Node children($)}
      NewCons = {Append Cons {Node constraints($)}}
      NewAss  = {Append Ass {Node assumptions($)}}
   in
      case Cs of nil then
	 [cstore(context:{Node context($)} constraints:NewCons assumptions:NewAss)]
      [] _|_ then
	 {Flatten {Map Cs fun{$ C} {GetAllLeafCStores C NewCons NewAss} end}}
      else
	 [cstore(context:{Node context($)} constraints:NewCons assumptions:NewAss) ]
      end
   end
   
   %%-----------------
   %% Recursive version of getAllCStores
   %%-----------------
   fun{GetAllCStores Node}
      Cs = {Node children($)}
   in
      cstore(context: {Node context($)}
	     constraints: {Node constraints($)}
	     assumptions: {Node assumptions($)})
      |{Flatten {Map Cs GetAllCStores}}
   end
   
   %%-----------------
   %% Recursive version of getAllNodes
   %%-----------------
   fun{GetAllNodes Nodes}
      case Nodes
      of nil then nil
      [] N|Nr then
	 {Append {GetAllNodes N}
	  {GetAllNodes Nr}}
      [] N then
	 N|{GetAllNodes {N children($)}}
      else nil end
   end
   
define
   
   %%-----------------
   %% Get all constraints and assumptions wrt. a given set of hypotheses
   %% within a list of context nodes.
   %%-----------------
   fun{GetConstraints Hyps Nodes}
      case Nodes
      of nil then nil
      [] N|Nr then {Append
		    {GetConstraints Hyps N}
		    {GetConstraints Hyps Nr}}
      [] N then
	 NodeHyps = {N hyps($)}
	 Children = {N children($)}
      in
	 if {SubSet NodeHyps Hyps}
	 then
	    {Append
	     {Map {N constraints($)} Cdr}
	     {GetConstraints Hyps Children}}
	 else nil end
      end
   end

   %%-----------------
   %% Get all assumption constraints wrt. a given set of hypotheses
   %% within a list of context nodes.
   %%-----------------
   fun{GetAssumptions Hyps Nodes}
      case Nodes
      of nil then nil
      [] N|Nr then {Append
		    {GetAssumptions Hyps N}
		    {GetAssumptions Hyps Nr}}
      [] N then
	 NodeHyps = {N hyps($)}
	 Children = {N children($)}
      in
	 if {SubSet NodeHyps Hyps}
	 then
	    {Append
	     {Map {N assumptions($)} Cdr}
	     {GetAssumptions Hyps Children}}
	 else nil end
      end
   end

   fun{InsertConstraint NodeName Hyps Con Nodes}
      case Nodes
      of nil then false
      [] N then
	 NodeHyps = {N hyps($)}
	 Children = {N children($)}
      in
	 if {SubSet NodeHyps Hyps} then
	    if {SubSet Hyps NodeHyps} then     %% the hyp sets are identical, so insert Con in N
	       {N addConstraint(NodeName Con)}
	       true
	    else          %% the node hyps are a subset of Hyps
	       if {Some Children fun{$ Child}    %% can Con be inserted in a Child of N?
				    {InsertConstraint NodeName Hyps Con Child $}
				 end}
	       then
		  true
	       else                              %% if not, insert a new child node
		  NewNode = {New ContextNode init(Hyps
						  {SetDifference Hyps NodeHyps}
						  nil
						  [NodeName#Con]
						  nil
						  N)}
	       in
		  {N addChild(NewNode)}
		  true
	       end
	    end
	 else              %% the node hyps are not a subset of Hyps, so we went too far...
	    false
	 end
      else
	 false
      end
   end

   fun{InsertAssumption NodeName Hyps Con Nodes}
      case Nodes
      of nil then false
      [] N then
	 NodeHyps = {N hyps($)}
	 Children = {N children($)}
      in
	 if {SubSet NodeHyps Hyps} then
	    if {SubSet Hyps NodeHyps} then     %% the hyp sets are identical, so insert Con in N
	       {N addAssumption(NodeName Con)}
	       true
	    else          %% the node hyps are a subset of Hyps
	       if {Some Children fun{$ Child}    %% can Con be inserted in a Child of N?
				    {InsertAssumption NodeName Hyps Con Child $}
				 end}
	       then
		  true
	       else                              %% if not, insert a new child node
		  NewNode = {New ContextNode init(Hyps
						  {SetDifference Hyps NodeHyps}
						  [NodeName#Con]
						  nil
						  nil
						  N)}
	       in
		  {N addChild(NewNode)}
		  true
	       end
	    end
	 else              %% the node hyps are not a subset of Hyps, so we went too far...
	    false
	 end
      else
	 false
      end
   end

   proc{WithdrawConstraint NodeName Nodes}
      case Nodes
      of nil then skip
      [] N|Nr then
	 {WithdrawConstraint NodeName N }
	 {WithdrawConstraint NodeName Nr }
      [] N then
	 {N withdrawConstraint(NodeName)}
	 {WithdrawConstraint NodeName {N children($)}}
      else skip end
   end

   
   %%-------------- The class itself -------------------------------
   class ContextTree
      from BaseObject
	 
      prop
	 locking
	 
      attr
	 rootNode: nil   %% the root node of the tree
	 hypDict: nil    %% the dictionary that maps hyp names to hyp formulas
	 nodeDict: nil   %% the dictionary that maps node names to node formulas
	 
	 
	 %%-----------------
	 %% initialize the context tree (set the empty root node)
	 %%-----------------
      meth init
	 rootNode <- {New ContextNode init(nil   % no hyps
					   nil   % no new hyps
					   nil   % no assumption constraints
					   nil   % no constraints
					   nil   % no children
					   nil)} % not father
	 hypDict <- {Dictionary.new}
	 nodeDict <- {Dictionary.new}
      end
      
      
      %%-----------------
      %% Get all constraints for the hypotheses set 'Hyps'.
      %%-----------------
      meth getConstraintsFor(Hyps $)
	 Cons = {GetConstraints Hyps @rootNode}
      in
	 %{Show Cons}
	 Cons
      end
      
      %%-----------------
      %% Get all assumption constraints for the hypotheses set 'Hyps' including
      %%  Hyps.
      %%-----------------
      meth getAssumptionsFor(Hyps $)
	 Cons = {GetAssumptions Hyps @rootNode}
      in
	 %{Show cons(Cons)}
	 Cons
      end

      %%-----------------
      %% Insert a constraint Con in the subtree T_Hyps
      %% of the context tree.
      %%-----------------
      meth insertConstraint(NodeName Hyps Con $)
	 {InsertConstraint NodeName Hyps Con @rootNode}
      end
      
      %%-----------------
      %% Insert a assumption constraint Con in the subtree T_Hyps
      %% of the context tree.
      %%-----------------
      meth insertAssumption(NodeName Hyps Con $)
	 {InsertAssumption NodeName Hyps Con @rootNode}
      end
      
      %%-----------------
      %% Withdraws a constraint Con from the context tree.
      %%-----------------
      meth withdrawConstraint(NodeName)
	 {WithdrawConstraint NodeName @rootNode}
      end
      
      %%-----------------
      %% Recursively show the context nodes 'NODES' and their children.
      %%-----------------
      meth RecShow(Nodes)
	 case Nodes of
	    nil then skip
	 [] N|Nr then ContextTree,RecShow(N)
	    ContextTree,RecShow(Nr)
	 [] N then
	    {N show}
	    ContextTree,RecShow({N children($)})
	 end
      end
      
      %%-----------------
      %% Show the context tree.
      %%-----------------
      meth show
	 ContextTree,RecShow([@rootNode])
	 {Show ''}
      end
      
      %%-----------------
      %% Recursively browse the context nodes 'NODES' and their children.
      %%----------------- 
      meth RecBrowse(Nodes)
	 case Nodes of
	    nil then skip
	 [] N|Nr then ContextTree,RecBrowse(N)
	    ContextTree,RecBrowse(Nr)
	 [] N then
	    {N browse}
	    ContextTree,RecBrowse({N children($)})
	 end
      end

      %%-----------------
      %% Browse the context tree.
      %%-----------------
      meth browse
	 ContextTree,RecBrowse([@rootNode])
	 {Browser.browse ''}
      end

      %%-----------------
      %% Send a graph representation of the context tree to
      %% DaVinci.
      %%-----------------
      meth daVinci($)
	 VS= "["#{@rootNode daVinci($)}#"]"
      in
	 %{Browser.browse {String.toAtom {VirtualString.toString VS}}}
	 VS
      end
      
      
      
      %%-----------------
      %% Create an entry in the hypDict dictionary for each hypothesis 
      %%-----------------
      meth ProcessHyps(Hyps)
	 {ForAll Hyps proc{$ H}
			 case H
			 of N#F then 
			    if {Dictionary.member @hypDict N} then skip
			    else {Dictionary.put @hypDict N F} end
			 else skip end
		      end}
      end
      


      %%-----------------
      %% Get all leaf nodes
      %%-----------------
      meth getAllLeaves($)
	 {GetAllLeaves @rootNode}
      end
      
      %%-----------------
      %% Get all leaf constraint stores 
      %%-----------------
      meth getAllLeafCStores($)
	 {GetAllLeafCStores @rootNode nil nil}
      end
      
      %%-----------------
      %% Get all constraint stores 
      %%-----------------
      meth getAllCStores($)
	 {GetAllCStores @rootNode}
      end

      %%-----------------
      %% Get all nodes
      %%-----------------
      meth getAllNodes($)
	 {GetAllNodes @rootNode}
      end


      %%-----------------
      %% Get the leftmost child in the context tree.
      %%-----------------
      meth getLeaf($)
	 {GetLeaf @rootNode}
      end
      
      
      %%-----------------
      %% Connect two context nodes, i.e., set the children and father pointers.
      %%-----------------
      meth connect(Father Child)        % connect nodes and add entry to NodeDict
	 Name = {Child plannerNode($)}  %for the new leaf
      in
	 {Father addChild(Child)}
	 {Child setFather(Father)}
	 
	 if {Dictionary.member @nodeDict Name $} then
	    L in
	    {Dictionary.get @nodeDict Name L}
	    {Dictionary.put @nodeDict Name 
	     {Append L [Child]}}
	 else
	    {Dictionary.put @nodeDict Name [Child]}
	 end
      end %m

      
      meth otherwise(M $)
	 {@app.show 'Sorry, unknown method for a ContextTree '#M}
	 'NIL'
      end
	 

   end % c
end %define
 
