%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% NodeClass.oz ; This file is part of the OMEGA system
%
%   major updates: 26.10.1999,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA Project                                                        
%   AG Siekmann/FB Informatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)
% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   %% The class of exploration nodes
   %%      jzimmer 18.1.99 
import
   System(show: Show)
   Browser
   Space
   
   ConfigCS(nodeColour)
   at 'x-ozlib://mathweb/cs/share/ConfigCS.ozf'
   Share(cdr
	 removeConstraint: RemoveConstraint
	)
   at 'x-ozlib://mathweb/cs/share/Share.ozf'
   Translate
   at 'x-ozlib://mathweb/cs/share/Translate.ozf'
   
export
   'class': ContextNode

prepare
   
   %%-----------------
   %% Helpful method for the computation of the daVinci representation of a node.
   %%-----------------
   fun{ChildrenToDaVinci NodeName L Children}
      case Children of
	 nil then nil
      [] C|nil then
	 % if {Length {C hyps($)}} == L then nil
	 "l(\" "#NodeName#"\","
	 #"e(\"\",[a(\"_DIR\",\"normal\")],"
	 #{C daVinci($)}
	 #"))"
      [] C|Cr then
	% if {Length {C hyps($)}} == L then nil
	 "l(\" "#NodeName#"\","
	 #"e(\"\",[a(\"_DIR\",\"normal\")],"
	 #{C daVinci($)}
	 #")),"
	 #{ChildrenToDaVinci NodeName L Cr}
      else nil end
   end
   
   fun{ListToVS Xs}
      case Xs
      of nil then nil
      [] X|nil then
	 X
      [] N#_|nil then
	 N
      [] N#_|Xr then
	 N#" "#{ListToVS Xr}
      [] X|Xr then
	 X#" "#{ListToVS Xr}
      end
   end
   
   /*
   fun{ListEqual Xs Ys}
      {And {All Xs fun{$ X}
		      {Member X Ys}
		   end}
       {All Ys fun{$ Y}
		  {Member Y Xs}
	       end}}
   end
*/
   
define
   class ContextNode
      from BaseObject

      prop
	 locking

      attr
	 hyps: nil
	 newHyps: nil

	 assumptions: nil
	 constraints: nil

	 children: nil
	 father: nil

	 %%-----------------
	 %% Initialize a node with a set of hypotheses, a list of children nodes, and a father node..
	 %%-----------------
      meth init(Hyps NewHyps Assumptions Constraints Children Father)
	 hyps <- Hyps
	 newHyps <- NewHyps
	 assumptions <- Assumptions
	 constraints <- Constraints
	 children <- Children
	 father <- Father
      end

      %%-----------------
      %% Clone the exploration node.
      %%-----------------
      meth clone($)
	 {New ContextNode init({Space.clone @compSpace}
				   nil
				   nil
				   nil
				   nil
				   self)}
      end
      
      %%-----------------
      %% Browse the exploration node.
      %%-----------------
      meth browse()
	 {Browser.browse '  Hyps:'#@hyps}
	 {Browser.browse '  Constraints:'#@constraints}
	 {Browser.browse '  Children:'#@children}
	 {Browser.browse '------------------------'}
      end
	 
      %%-----------------
      %% Show the exploration node.
      %%-----------------
      meth show()
	 {Show '  hyps:'#@hyps}
	 {Show '  Assumptions:'#@assumptions}
	 {Show '  Constraints:'#@constraints}
	 {Show '  Children:'#@children}
	 {Show ''}
      end
      
      
      %%-----------------
      %% Get a daVinci representation of the exploration node.
      %%-----------------
      meth daVinci($)
	 Name = if @hyps == nil then
		   "{ }"
		else
		   {VirtualString.toString
		    {Translate.list2DaVinci {Map @hyps Share.cdr}}}
		end
      in
	 "l(\" "#Name#"\",n(\"\",[a(\"OBJECT\", \" "#Name#"\"),"  %%a(\"OBJECT\",\""#{ListToVS @hyps}#"\"),"
	 #"a(\"COLOR\",\""#ConfigCS.nodeColour#"\")],"
	 #"["
	 #{ChildrenToDaVinci Name {Length Name} @children}
	 #"]))"
      end
      
      %%-----------------
      %% Get the names of the planning nodes occuring in this exploration node.
      %%-----------------
      meth plannerNodes($)
	 {Map @constraints fun{$ N#_}
			      N
			   end}
      end
      
      %%-----------------
      %% Get all hyps of a node.
      %%-----------------
      meth hyps($)
	 @hyps
      end
      
      %%-----------------
      %% Get the context of a node.
      %%-----------------
      meth context($)
	 @hyps
      end
      
      %%-----------------
      %% Get the hypotheses new to a node wrt. to it's father's hyps.
      %%-----------------
      meth newHyps($)
	 @newHyps
      end
      
      %%-----------------
      %% Get all constraints of a node.
      %%-----------------
      meth constraints($)
	 @constraints
      end

      %%-----------------
      %% Get all assumption constraints of a node.
      %%-----------------
      meth assumptions($)
	 @assumptions
      end

      %%-----------------
      %% Get all children of a node.
      %%-----------------
      meth children($)
	 @children
      end
      
      %%-----------------
      %% Set the hypotheses annotation.
      %%-----------------
      meth setHyps(Hyps)
	 hyps <- Hyps
      end

      %%-----------------
      %% Set the hypotheses that are new wrt. the hyps of the father node.
      %%-----------------
      meth setNewHyps(Hyps)
	 newHyps <- Hyps
      end

      
      %%-----------------
      %% Add a new constraint to the node
      %%-----------------
      meth addConstraint(NodeName Con)
	 if {Member Con {Map @constraints
			 fun{$ _#C} C end}}
	 then
	    skip
	 else 
	    constraints <- {Append @constraints [NodeName#Con]}
	 end
      end

      %%-----------------
      %% Add a new assumption constraint to the node
      %%-----------------
      meth addAssumption(NodeName Con)
	 if {Member Con {Map @assumptions
			 fun{$ _#C} C end}}
	 then
	    skip
	 else 
	    assumptions <- {Append @assumptions [NodeName#Con]}
	 end
      end


      %%-----------------
      %% Withdraws a constraint Con from the context node.
      %%-----------------
      meth withdrawConstraint(NodeName)
	 assumptions <- {RemoveConstraint NodeName @assumptions}
	 constraints <- {RemoveConstraint NodeName @constraints}
      end
      
      %%-----------------
      %% Add a node to the list of children of the current node.
      %%-----------------
      meth addChild(Child)
	 children <- {Append @children [Child]}
      end
      
      %%-----------------
      %% Set the father pointer.
      %%-----------------
      meth setFather(Node)
	 father <- Node
      end

      meth otherwise(M $)
	 {@app.show 'Sorry, unknown method for a ContextNode '#M}
	 'NIL'
      end
	 
   end % class
end %define

/*
	       NewSpace = {New MySpace init(NewCompSpace nil)}
	    in
	       nextSpaces <- [NewSpace]
	       B = true
	       NewSpace
*/
