%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Share.oz ; This file is part of the OMEGA system
%
%   major updates: 3.12.1999,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA Project                                                        
%   AG Siekmann/FB Informatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)

% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   %% All the helpful functions for sets
   %%      jzimmer 23.3.00
   
import
   System(eq: Eq
	 % show:Show
	 )
   
export
   In                % fun {$ X S}        --> Bool
   Subset            % fun {$ S1 S2}      --> Bool
   Equal             % fun {$ S1 S2}      --> Bool
   
   Union             % fun {$ S1 S2}      --> S1 U S2
   Intersection      % fun {$ S1 S2}      --> S1 /\ S2
   Difference        % fun {$ S1 S2}      --> S1 - S2

   PairSets          % fun {$ [E1 ... En]} --> [ E1#E2 E1#E3 ... En-1#En]
%   PowerSet         % fun {$ [E1 ... En] i} --> [ nil [E1 ... Ei] [E2... Ei+1].....]
   MapUnion          % fun {$ [S1...Sn]}  --> S1 U S2 U ... U Sn
   MapIntersection   % fun {$ [S1...Sn]}  --> S1 /\ S2 /\ ... /\ Sn
   
   
prepare
   fun {TopFlatten Ls}
      case Ls
      of nil then nil
      [] L|Lr then
	 {Append L {TopFlatten Lr}}
      else Ls end
   end

define
   
   fun {In X L} % --> bool
      case L
      of nil then false
      [] V|LRest then {Eq X V} orelse {In X LRest}
      end
   end
   
   fun {Subset L1 L2}
      {All L1 fun{$ L} {Member L L2} end}
   end
   
   fun {Equal L1 L2}
      {Subset L1 L2} andthen {Subset L2 L1}
   end


   fun {Union L1 L2} % --> ResultList; merges two lists of variables 
      case L1
      of nil then L2
      [] V|L1Rest then
	 if {In V L2}
	 then {Union L1Rest L2}
	 else V|{Union L1Rest L2}
	 end
      end
   end
   
   
   fun {Difference A B} % --> A\B; A and B are Lists, but they 'mean' sets
      case A
      of nil then nil
      [] A1|ARest then
         if {In A1 B}
         then {Difference ARest B}
         else A1|{Difference ARest B}
         end
      end
   end
   
   fun {Intersection L1 L2} % --> ResultList
      %{Show inter(L1 L2)}
      case L2
      of nil then nil
      [] V|L2Rest then
	 if {In V L1}
	 then V|{Intersection L1 L2Rest}
	 else {Intersection L1 L2Rest}
	 end
      end
   end

   
   fun {PairSets Xs}
      {TopFlatten
       {List.mapInd Xs
	fun{$ I1 X}
	   {List.mapInd
	    {List.drop Xs I1}
	    fun{$ I2 Y}
	       [I1#X I1+I2#Y]
	   end}
	end}}
   end
   
   fun {MapUnion L} % --> ResultList; L is a List of Lists
      case L
      of nil then nil
      [] L1|LRest then {Union L1 {MapUnion LRest}}
      end
   end

   fun {MapIntersection Ls} % --> ResultList
      %{Show inters(Ls)}
      case Ls
      of nil then nil
      [] [L] then L
      [] L|Lr then
	 {Intersection L {MapIntersection Lr}}
      end
   end


end










