%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Share.oz ; This file is part of the OMEGA system
%
%   major updates: 3.12.1999,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA Project                                                        
%   AG Siekmann/FB Informatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)

% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   %% All the helpful functions for CoSIE
   %%      jzimmer 1.8.99 

import
   Property
   System(show: Show
	  showInfo: ShowInfo
	  eq
	 )
   Browser(browse: Browse)
   Prop at 'x-oz://contrib/Propagator.ozf'
   
export
   Car
   Cdr
   SetDifference
   MyEq
   MyMember
   PropMember
   
   GetSolution
   ExtractConstraints
   ConstructTerm
   CompleteCons
   
   FindName
   FindVar

   RemoveConstraint
   RemoveProps
   RemoveVar
   KillProps

   MatchList
   ApplyMatching
   ApplySubst
   ApplyRenaming

   Unify
   Debug
   
define

   %%----------------------------------------------------
   proc{Debug X}
      if {Property.condGet 'debug.out' false} then 
	 {Show X}
      else skip end
   end
   
   fun{Car Pair}
      case Pair
      of X#_ then X else Pair end
   end

   fun{Cdr Pair}
      case Pair
      of _#Y then Y else Pair end
   end

   fun{SetDifference Ys Xs}
      case Ys
      of nil then nil
      [] Y|Yr then
	 if {Member Y Xs} then 
	    {SetDifference Yr Xs}
	 else
	    Y|{SetDifference Yr Xs}
	 end
      else Ys
      end
   end

   fun{RemoveDuplicates Xs}
      case Xs
      of nil then nil
      [] X|Xr then
	 if {Member X Xr} then
	    {RemoveDuplicates Xr}
	 else
	    X|{RemoveDuplicates Xr}
	 end
      else Xs end
   end

   fun{RemoveConstraint NodeName Cons}
      case Cons
      of nil then nil
      [] Name#Con|Cr then
	 if Name==NodeName then Cr
	 else (Name#Con)|{RemoveConstraint NodeName Cr} end
      else Cons end
   end

  %%---------------- end      Is Constraint --------------------
   
   fun{IsConstrVar Con}
      case Con of
	 _#var then true
      else false end
   end

   fun{GetVariables Con}
      if {IsRecord Con} then
	 if {IsConstrVar Con} then
	    [{Car Con}]
	 else
	    ConList = {Record.toList Con}
	 in
	    {Flatten {Map ConList GetVariables}}
	 end
      else nil end	 
   end
   
      
   fun{MyEq X Y}
      %{Show myEq(X Y)}
      if {IsDet X} then
	 if {IsDet Y} then
	    if {IsAtom X} then
	       {And {IsAtom Y} X==Y}
	    elseif {IsNumber X} then
	       {And {IsNumber Y} X==Y}
	    else case X 
		 of N1#_ then
		    case Y
		    of N2#_ then
		       N1==N2
		    else false end
		 else
		    {IsRecord X}
		    andthen {IsRecord Y}
		    andthen ({Label X} == {Label Y})
		    andthen {Record.allInd X fun{$ I F} {MyEq F Y.I} end}
		 end
	    end
	 else false end
      else
	 if {IsDet Y} then
	    false
	 else {System.eq X Y} end
      end
   end
   
   fun{MyMember X Ys}
      case Ys
      of nil then false
      [] Y|Yr then
	 {MyEq X Y} orelse {MyMember X Yr}
      else false end
   end


   
   fun{PropEq P1 P2}
      %{Show propEq(P1 P2)}
      if {IsRecord P2} then
	 if {IsRecord P1} then
	    {Label P1} == {Label P2}
	    andthen
	    {Record.allInd {Record.subtract P1 ref}
	     fun{$ I A}
		{MyEq A {Record.subtract P2 ref}.I}
	     end}
	 else
	    {System.eq P1 P2.ref}
	 end
      else
	 if {IsRecord P1} then
	    {System.eq P1.ref P2}
	 else {System.eq P1 P2} end
      end
   end
   
   fun{PropMember X Ps}
      %{Show propMember(X Ps)}
      case Ps of
	 nil then false
      [] P|Pr then
	 if {PropEq X P} then true
	 else {PropMember X Pr} end
      end
   end

   fun{RemoveSchema X Schemas}
      case Schemas 
      of nil then nil
      [] S|Sr then
	 if {System.eq S X} then 
	    {RemoveSchema X Sr}
	 else
	    S|{RemoveSchema X Sr}
	 end
      end
   end

   fun {VarMember N Stream}
      case Stream
      of nil then false
      [] L#_|Xs then
	 if L == N then
	    true 
	 else
	    {VarMember N Xs}
	 end
      else false end
   end

   
   fun{RemoveProps Props DelProps}
      %{Show delpr#DelProps}
      case Props
      of nil then nil
      [] P|Pr then
	 if {PropMember P DelProps} then
	    {RemoveProps Pr DelProps}
	 else
	    P|{RemoveProps Pr DelProps}
	 end
      end
   end
   
   fun{GetSolution Vars Props AllProps AlienReps FuncSymbols}
      case Vars
      of nil then nil
      [] Name#var|Vr then
	 %{Show 'Var1'#Name}
	 case Props
	 of nil then nil
	 [] P|Pr then
	    if {Label P} == '=' then
	       case P.1
	       of VarName#Type then
		  if (Type == var orelse Type == nested) andthen (Name == VarName)
		  then
		     RestProps = {RemoveProps AllProps [P]}
		  in
		     '='(Name#var {ConstructTerm P.2 RestProps RestProps AlienReps FuncSymbols})
		     |{GetSolution Vr RestProps AllProps AlienReps FuncSymbols}
		  else
		     {GetSolution Vars Pr AllProps AlienReps FuncSymbols}
		  end
		  
	       elsecase P.2
	       of VarName#Type then
		  if (Type == var orelse Type == nested) andthen (Name == VarName)
		  then
		     RestProps = {RemoveProps AllProps [P]}
		  in
		     '='(Name#var {ConstructTerm P.1 RestProps RestProps AlienReps FuncSymbols})
		     |{GetSolution Vr RestProps AllProps AlienReps FuncSymbols}
		  else 
		     {GetSolution Vars Pr AllProps AlienReps FuncSymbols}
		  end
	       else
		  {GetSolution Vars Pr AllProps AlienReps FuncSymbols}
	       end
	    else
	       {GetSolution Vars Pr AllProps AlienReps FuncSymbols}
	    end
	 else
	    {GetSolution Vr AllProps AllProps AlienReps FuncSymbols}
	 end
      else 'T' end
   end
   
   fun{ExtractConstraints ConProps TermProps AlienReps FuncSymbols}
      case ConProps
      of nil then nil
      [] P|Pr then
	 {Record.map {Record.subtract P ref}
	  fun{$ I} {ConstructTerm I TermProps TermProps AlienReps FuncSymbols} end}
	 |{ExtractConstraints Pr TermProps AlienReps FuncSymbols}
      else nil end
   end
   
   
   fun{ConstructTerm T Props AllProps AlienReps FuncSymbols}
      %{Show construct#T}
      if {IsAtom T} orelse {IsNumber T} then
	 T
      else
	 case T
	 of nil then nil
	 [] H|Tr then         %% handle lists of terms
	    {ConstructTerm H Props AllProps AlienReps FuncSymbols} 
	    |{ConstructTerm Tr Props AllProps AlienReps FuncSymbols}
	 [] _#const then   %% constants are just printed
	    T
	 [] _#var then     %% problem vars are just printed.
	    T
	 [] _#nested then  %% nested vars have to be replaced by the right term.
	    case Props
	    of nil then {Show 'worst case 1'} T
	    [] P|Pr then Func = {Label P} in
	       %{Show matchpropa#T#P}
	       if {Member Func FuncSymbols} then
		  if P.3 == T then
		     Func({ConstructTerm P.1 AllProps AllProps AlienReps FuncSymbols}
			  {ConstructTerm P.2 AllProps AllProps AlienReps FuncSymbols}
			 )
		  else {ConstructTerm T Pr AllProps AlienReps FuncSymbols} end
	       else
		  {ConstructTerm T Pr AllProps AlienReps FuncSymbols}
	       end
	    end
	 [] Name#alien then  %% alien vars must be replaced by the alien term they stand for.
	    Replace = {Dictionary.get AlienReps Name}
	 in
	    %{Show replace(Replace)}
	    Replace
	            %{ConstructTerm Replace Props AllProps AlienReps}
	 else
	    if {IsRecord T} then
	       {Record.map T fun{$ E}
				{ConstructTerm E Props AllProps AlienReps FuncSymbols}
			     end}
	    else T end
	 end
      end
   end

   fun{CompleteCons T TermProps AlienReps FuncSymbols}
      %{Show complete_cons(T)}
      if {IsRecord T} then
	 {Record.map T fun{$ E}
			  {ConstructTerm E TermProps TermProps 
			   AlienReps FuncSymbols}
		       end}
      else T end
   end
   
   fun{RemoveVar Name Vars}
      case Vars
      of nil then nil
      [] V|Vr then
	 if Name == {Car V} then 
	    {RemoveVar Name Vr}
	 else
	    V|{RemoveVar Name Vr}
	 end
      end
   end

   fun{FindName V Ss}
      case Ss
      of nil then nil
      [] Name#Var|Sr then
	 if {System.eq Var V} then Name
	 else {FindName V Sr} end
      end
   end
   
   fun{FindVar N Ss}
      case Ss
      of nil then nil
      [] Name#Var|Sr then
	 %{Show vars#(Name == N)}
	 if Name == N then Var
	 else {FindVar N Sr} end
      end
   end
   
   proc{KillProps Ps}
      case Ps
      of nil then skip
      [] P|Pr then
%         {Browse propa(P.ref)}
         case {Prop.is P.ref}
         of true then 
	    {Prop.discard P.ref}
	 else
	    {Show no_reference}
	 end
	 {KillProps Pr}
      end
   end

   %%------------------------------------------------
   %% matching propagator schemas with real Propagators 
   %%------------------------------------------------
   fun{MatchArgs PArgs SArgs Matching}
%      {Debug match_args(SArgs PArgs {Dictionary.entries Matching})}
      case SArgs
      of nil then
	 if PArgs == nil then Matching else nil end
      [] S|Sr then
	 case PArgs
	 of nil then nil
	 [] P|Pr then
	    if {IsAtom S} then
	       if {Dictionary.member Matching S} then
		  case P
		  of PName#_ then
		     if {Dictionary.member Matching S} then
			Entry = {Dictionary.get Matching S}
		     in
			case Entry
			of EName#_ then
			   if {System.eq EName PName} then
			      {MatchArgs Pr Sr Matching}
			   else nil end
			else nil end
		     else 
			{Dictionary.put Matching S P}
			{MatchArgs Pr Sr Matching}
		     end
		  else
		     Entry = {Dictionary.get Matching S}
		  in
		     if {IsNumber Entry} then
			if {IsNumber P} then
			   if (P == Entry) then
			      {MatchArgs Pr Sr Matching}
			   else nil end
			else nil end
		     elseif {IsNumber P} then
			nil
		     elseif {System.eq Entry P} then
			{MatchArgs Pr Sr Matching}
		     else nil end 
		  end
	       else
		  if {IsList P} then nil else
		     {Dictionary.put Matching S P}
		     {MatchArgs Pr Sr Matching}
		  end
	       end
	    elseif {IsNumber S} then
%	       {Show 'number'#S#P#(S==P)}
	       if {And {IsNumber P} (S==P)} then
		  {MatchArgs Pr Sr Matching}
	       else nil end
	    else case S
		 of SName#num then
		    if {IsNumber P} then
		       if {Dictionary.member Matching SName} then
			  Entry = {Dictionary.get Matching SName}
		       in
			  if (P == Entry) then
			     {MatchArgs Pr Sr Matching}
			  else nil end
		       else
			  {Dictionary.put Matching SName P}
			  {MatchArgs Pr Sr Matching}
		       end
		    else nil end
		    
		 [] SName#SType then
		    if {Dictionary.member Matching SName} then
		       case P
		       of PName#_ then
			  Entry = {Dictionary.get Matching SName}
		       in
			  case Entry
			  of EName#_ then
			     if {System.eq EName PName} then
				{MatchArgs Pr Sr Matching}
			     else nil end
			  else nil end
		       else nil end
		    else 
		       case P
		       of _#PType then
			  if (SType == PType) then
			     {Dictionary.put Matching SName P}
			     {MatchArgs Pr Sr Matching}
			  else nil end
		       else
			  if SType == const andthen {IsNumber P} then
			     {Dictionary.put Matching SName P}
			     {MatchArgs Pr Sr Matching}
			  else nil end 
		       end
		    end
		 else
		    {Show 'unknown schema'#S}
		    nil
		 end 
	    end 
	 end
      end
   end
   

   
   fun{MatchSchemaProp S P Matching}
%      {Debug match_schema_prop(S P)}
%      {Debug subst({Dictionary.entries Matching})}
      case S
      of '>='(LHS RHS) then
	 {MatchSchemaProp '<='(RHS LHS) P Matching}
      [] PName#'>='(LHS RHS) then
	 {MatchSchemaProp PName#'<='(RHS LHS) P Matching}
      [] '<'(LHS RHS) then
	 {MatchSchemaProp '>'(RHS LHS) P Matching}
      [] PName#'<'(LHS RHS) then
	 {MatchSchemaProp PName#'>'(RHS LHS) P Matching}

      [] PropName#Rec then
	 Fun = {Label Rec}
      in
	 if {Label P} == Fun then
	    Result = {MatchArgs {Record.toList {Record.subtract P ref}} {Record.toList Rec} Matching}
	 in
	    if Result == nil then nil else
	       {Dictionary.put Matching PropName P}  %%store PropRef of the Props to be deleted
	       Result
	    end
	 else nil end
      else
	 if {IsRecord S} then
	    if {Label S} == {Label P} then
	       Result = {MatchArgs {Record.toList {Record.subtract P ref}} {Record.toList S} Matching}
	    in
	       Result
	    else nil end
	 else nil end
      end
   end
   

   fun{MatchList Schemas ConProps TermProps FirstProps ConSymbols Matching}
%      {Debug 'matchlist'(Schemas ConProps TermProps {Dictionary.entries Matching})}
      case Schemas
      of nil then
	 Matching
      [] S|Sr then
	 IsCon = case S
		 of _#Rec then
		    if {IsRecord Rec} then
		       {Member {Label Rec} ConSymbols}
		    else false end
		 elseif {IsRecord S} then
		    {Member {Label S} ConSymbols}
		 end
      in
	 if IsCon then
	    case ConProps
	    of nil then nil
	    [] P|Pr then 
	       NewMatching = {Dictionary.clone Matching}
	       Result1 = {MatchSchemaProp S P NewMatching}
	    in
	       if Result1 == nil then
		  {MatchList Schemas Pr TermProps P|FirstProps ConSymbols Matching}
	       else
		  Result2 = {MatchList Sr {Append FirstProps Pr} TermProps nil ConSymbols NewMatching}
	       in
		  if Result2 == nil then
		     Result3 = {MatchList Schemas Pr TermProps P|FirstProps ConSymbols Matching}
		  in
		     if Result3 == nil then nil
		     else
			Result3
		     end
		  else
		     Result2
		  end
	       end
	    else nil end
	 else
	    case TermProps
	    of nil then nil
	    [] P|Pr then 
	       NewMatching = {Dictionary.clone Matching}
	       Result1 = {MatchSchemaProp S P NewMatching}
	    in
	       if Result1 == nil then
		  {MatchList Schemas ConProps Pr P|FirstProps ConSymbols Matching}
	       else
		  Result2 = {MatchList Sr ConProps {Append FirstProps Pr} nil ConSymbols NewMatching}
	       in
		  if Result2 == nil then
		     Result3 = {MatchList Schemas ConProps Pr P|FirstProps ConSymbols Matching}
		  in
		     if Result3 == nil then nil
		     else
			Result3
		     end
		  else
		     Result2
		  end
	       end
	    else nil end
	 end
      else nil end
   end
   

   fun{UnifyHelp Term1 Term2 MGU}
      {Show unifyhelp(Term1 Term2 {Dictionary.entries MGU})}
      case Term1
      of nil then (Term2 == nil)
      [] T|Tr then
	 case Term2
	 of nil then nil
	 [] S|Sr then
	    {UnifyHelp T S MGU} andthen
	    {UnifyHelp Tr Sr MGU}
	 else
	    false
	 end
	 
      [] V1#var then
	 %%                           HIER NORMALERWEISE OCCUR-CHECK!!!
	 if {Dictionary.member MGU V1} then
	    Binding = {Dictionary.get MGU V1}
	 in
	    {MyEq Binding Term2}
	 else
	    case Term2
	    of V2#var then
	       if V1 == V2 then false
	       else
		  true
	       end
	    else 
	       {Dictionary.put MGU V1 Term2}
	       true
	    end
	 end
	 
      [] _#const then
	 case Term2
	 of V2#var then 
	    if {Dictionary.member MGU V2} then
	       Binding = {Dictionary.get MGU V2}
	    in
	       {MyEq Binding Term1}
	    else
	       {Dictionary.put MGU V2 Term1}
	       true
	    end
	 else
	    {MyEq Term1 Term2}
	 end
      else
	 if {IsRecord Term1} then
	    case Term2
	    of V2#var then 
	       if {Dictionary.member MGU V2} then
		  Binding = {Dictionary.get MGU V2}
	       in
		  {MyEq Binding Term1}
	       else
		  {Dictionary.put MGU V2 Term1}
		  true
	       end
	    else
	       {Show 'casee'({Label Term1} == {Label Term2})}
		{IsRecord Term2} andthen
	       ({Label Term1} == {Label Term2}) andthen
	       {UnifyHelp
		{Record.toList Term1}
		{Record.toList Term2}
		MGU}
	    end
	 else 
	    case Term2
	    of nil then (Term1 == nil)
	    [] T|Tr then
	       case Term1
	       of nil then nil
	       [] S|Sr then
		  {UnifyHelp T S MGU} andthen
		  {UnifyHelp Tr Sr MGU}
	       else
		  false
	       end
		  
	    [] V2#var then
	       if {Dictionary.member MGU V2} then
		  Binding = {Dictionary.get MGU V2}
	       in
		  {MyEq Binding Term1}
	       else
		  {Dictionary.put MGU V2 Term1}
		  true
	       end
	       
	    [] _#const then
	       {MyEq Term1 Term2}
	    else
	       if {IsAtom Term1} then
		  {IsAtom Term2} andthen Term1 == Term2
	       else
		  {MyEq Term1 Term2}
	       end
	    end
	 end
      end
   end
      
   
   %%---------------------------------
   %%       Simple FOL-Term-Unification
   %%---------------------------------
   fun{Unify Term1 Term2}
      MGU = {NewDictionary}
   in
      if {UnifyHelp Term1 Term2 MGU} then
	 MGU
      else nil end
   end
   
   %%---------------------------------
   %%        Application of matchings an substitutions
   %%---------------------------------
   fun{ApplyMatching Obj Matching}
      {Debug 'apply-matching'#Obj}
      case Obj
      of nil then nil
      [] O|Or then
	 {ApplyMatching O Matching}|{ApplyMatching Or Matching}
      [] _#Prop then
	 {ApplyMatching Prop Matching}
      else
	 {Debug 'apply-matching2'#Obj}
	 if {IsNumber Obj} then Obj else
	    if {IsAtom Obj} then
	       if {Dictionary.member Matching Obj} then
		  {Dictionary.get Matching Obj}
	       else
		  {Show 'unknown case'#Obj}
		  Obj
	       end
	    elseif {IsRecord Obj} then
	       %{Show 'apply-matching-rec'#Obj}
	       {Record.map Obj fun{$ A} {ApplyMatching A Matching} end}
	    else
	       %{Show 'apply-matching3'#Obj}
	       if {Dictionary.member Matching Obj} then
		  {Dictionary.get Matching Obj}
	       else
		  {Show 'unknown case'#Obj}
		  Obj
	       end
	    end
	 end
      end
   end
   
   fun{ApplySubst Subst Obj}
      %{Show 'apply-subst'#{Dictionary.entries Subst}#Obj}
      case Obj
      of nil then nil
      [] _|_ then
	 {Map Obj fun{$ I} {ApplySubst Subst I} end}
      [] Name#var then
	 if {Dictionary.member Subst Name} then
	    %{Show member}
	    {Dictionary.get Subst Name}
	 else Obj end
      else
	 if {IsRecord Obj} then
	    case Obj
	    of '='(Name#var ...) then
	       %{Show member({Dictionary.member Subst Name})}
	       if {Dictionary.member Subst Name} then
		  Obj
	       else {Record.map Obj fun{$ I} {ApplySubst Subst I} end} end
	    else {Record.map Obj fun{$ I} {ApplySubst Subst I} end} end
	 else Obj end
      end
   end
   
   fun{ApplyRenaming Subst Obj}
      %{Show 'apply-subst'#{Dictionary.entries Subst}#Obj}
      case Obj
      of nil then nil
      [] _|_ then
	 {Map Obj fun{$ I} {ApplyRenaming Subst I} end}
      else
	 if {IsAtom Obj} andthen {Dictionary.member Subst Obj} then
	    %{Show member}
	    {Dictionary.get Subst Obj}
	 elseif {IsRecord Obj} then
	    {Record.map Obj fun{$ I} {ApplyRenaming Subst I} end} 
	 else Obj end
      end 
   end
   
   
end % define
/*

fun {Flatten Ls}
      {FoldL Ls
       fun {$ L R}
	  if {IsDet R} then
	     case R of nil then L
	     [] T|Rs then {Append L {Flatten R}}
	     else [R] end 
	  else
	     R|L
	  end
       end nil}
   end
   fun{PermuteRecordTerm Term T Props AlienReps} 
      Func = {Label Term}
   in
      %{Show permute#Term#T}
      if {Member Func ['+' '-' '*' '/' 'MIN' '||']} then
	 if Term.3 == T then
	    Func({ConstructRecordTerm Term.1 Props Props AlienReps}
		 {ConstructRecordTerm Term.2 Props Props AlienReps}
		)
	 else nil end
      elseif Func == '||' then
	 Func({ConstructRecordTerm Term.1 Props Props AlienReps})
      else nil end
   end
   
      
   fun{ConstructRecordTerm T Props AllProps AlienReps}
      %{Show constructRec#T}
      if {IsAtom T} then {Atom.toString T}
      elseif {IsNumber T} then
	 T
      else
	 case T
	 of nil then 'dummy'()
	 [] Name#const then   %% constants are just printed
	    Name 
	 [] Name#var then     %% problem vars are just printed.
	    Name 
	 [] Name#nested then  %% nested vars have to be replaced by the right term.
	    case Props
	    of nil then {Show 'worst case 1'} T
	    [] P|Pr then
		  %{Show matchpropa#T#P.function|P.args}
	       if {Member {Label P} ['+' '-' '*' '/' 'MIN' '||']} then
		  RealTerm = {PermuteRecordTerm {Record.subtract P ref}
			      T
			      {RemoveProps AllProps [P]}
			      AlienReps}
	       in
		  if RealTerm == nil then
		     {ConstructRecordTerm T Pr AllProps AlienReps}
		  else
		     RealTerm
		  end
	       else  {ConstructRecordTerm T Pr AllProps AlienReps}
	       end
	    else T end
	 [] Name#alien then  %% alien vars must be replaced by the alien term they stand for.
	    Replace = {Dictionary.get AlienReps Name}
	 in
	    %{Show replace#Name#Replace}
	    %{ConstructRecordTerm Replace Props AllProps AlienReps Types}
	    Replace
	 else if {IsRecord T} then
		 {Record.map T fun{$ I}
				  {ConstructRecordTerm I Props AllProps AlienReps}
			       end}
	      else T end
	 end
      end
   end
   fun{GetConstraintString Var Props AlienReps}
      Name = {Car Var}
      LBounds = {GetLowerBounds Name#var Props}
      UBounds = {GetUpperBound sName#var Props}
      Equalities = {GetEquations Name#var Props}
   in
      %%{Show Name#LBounds#UBounds#Equalities}
      if Equalities \= nil then
	 {VirtualString.toString Name#": "}
	 #{VirtualString.toString {HandleBounds Name Equalities Props AlienReps}}
      else
	 {VirtualString.toString Name#": "}
	 #{VirtualString.toString
	   if LBounds \= nil then
	      if UBounds \= nil then
		 "([AND d] "
		 #{HandleBounds Name LBounds Props AlienReps}#" "
		 #{HandleBounds Name UBounds Props AlienReps}
		 #")"
	      else {HandleBounds Name LBounds Props AlienReps} end
	   elseif UBounds \= nil then {HandleBounds Name UBounds Props AlienReps} 
	   else nil end}
      end
   end
   
   fun{HandleBounds VarName Bounds Props AlienReps}
      case Bounds
      of nil then nil
      [] B|Br then
	 case B
	 of '<'#T then
	    if Br==nil then
	       "([LESS d] "
	       #"["#VarName#" a] "
	       #{ConstructTerm T Props Props AlienReps}
	       #")"
	    else
	       "([AND d] "
	       #"([LESS d] "
	       #"["#VarName#" a] "
	       #{ConstructTerm T Props Props AlienReps}
	       #") "
	       #{HandleBounds VarName Br Props AlienReps}
	       #")"
	    end
	 [] '<='#T then
	    if Br==nil then
	       "([LEQ d] "
	       #"["#VarName#" d] "
	       #{ConstructTerm T Props Props AlienReps}
	       #")"
	    else
	       "([AND d] "
	       #"([LEQ d] "
	       #"["#VarName#" d] "
	       #{ConstructTerm T Props Props AlienReps}
	       #") "
	       #{HandleBounds VarName Br Props AlienReps}
	       #")"
	    end
	 [] '='#T then
	    if Br==nil then
	       "([= d] "
	       #"["#VarName#" d] "
	       #{ConstructTerm T Props Props AlienReps}
	       #")"
	    else
	       "([AND d] "
	       #"([= d] "
	       #"["#VarName#" d] "
	       #{ConstructTerm T Props Props AlienReps}
	       #") "
	       #{HandleBounds VarName Br Props AlienReps}
	       #")"
	    end
	 [] '>'#T then
	    if Br==nil then
	       "([LESS d] "
	       #{ConstructTerm T Props Props AlienReps}
	       #" ["#VarName#" d]"
	       #")"
	    else
	       "([AND d] "
	       #"([LESS d] "
	       #{ConstructTerm T Props Props AlienReps}
	       #" ["#VarName#" d]"
	       #") "
	       #{HandleBounds VarName Br Props AlienReps}
	       #")"
	    end
	 [] '>='#T then
	    if Br==nil then
	       "([LEQ d] "
	       #{ConstructTerm T Props Props AlienReps}
	       #" ["#VarName#" d]"
	       #")"
	    else
	       "([AND d] "
	       #"([LEQ d] "
	       #{ConstructTerm T Props Props AlienReps}
	       #" ["#VarName#" d]"
	       #") "
	       #{HandleBounds VarName Br Props AlienReps}
	       #")"
	    end
	 else
	    %{Show wrong#B}
	    nil end
      else nil end
   end

      
   old-match-schema-prop-entailment-test
in
	    case {Label S}
	    of '>' then [A1 A2] = RealArgs in
	       if {And {IsNumber A1} {IsNumber A2}} then 
		  {Debug vergleich#A1#A2#(A1>A2)}
		  if (A1 > A2) then Matching else nil end 
	       else nil end
	    [] '>=' then [A1 A2] = RealArgs in
	       if {And {IsNumber A1} {IsNumber A2}} then 
		  {Debug vergleich#A1#A2#(A1>=A2)}
		  if (A1 >= A2) then Matching else nil end
	       else nil end
	    else nil end
	 else nil end
      end

   fun{MatchSchemaProps S Props Matching}
      {Debug match_schema_props(S Props)}
      case Props 
	 of nil then nil
      [] P|Pr then
	 NewMatching = {Dictionary.clone Matching}
	 Result = {MatchSchemaProp S P NewMatching}
      in
	 if Result == nil then
	    {MatchSchemaProps S Pr Matching}
	    else
	    [P NewMatching] end 
      else nil end
   end


   */

