%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Translate.oz ; This file is part of the OMEGA system
%
%   major updates: 26.10.1999,
%
%   Authors: Juergen Zimmer
%
%   email: jzimmer@ags.uni-sb.de
%   
% For information about this program, write to:                          
%   OMEGA Project                                                        
%   AG Siekmann/FB Informatik                                            
%   Universitaet des Saarlandes                                          
%   Bau 36, 4. Stock                                                     
%   D-66041 Saarbruecken                                                 
%   Germany    
%
% For information about the newest version of Omega, see 
%   http://www.ags.uni-sb.de/~omega/
%
% This program is free software; it can be used under the terms of the GNU General
% Public License as published by the Free Software Foundation; either version 2 of
% the License, or any later version.
%
% This program is distributed in the hope that it will be useful, but
% WITHOUT ANY WARRANTY; without even the implied warranty of
% merchantibility or fitness for a particular purpose. 
% See the GNU General Public License 
% (http://www.fsf.org/copyleft/gpl.html)
% for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

functor
   %% The functor with translation functions
   %%      jzimmer 22.9.99 

import
   %System(show: Show showInfo: ShowInfo eq)
   %Browser(browse: Browse)
   
export
   List2DaVinci
   
   AList2Post
   List2Post
   List2PostConjunct
   
   Trace2Post
   
   Oz2Post

define
   
   %%-----------------
   %%  Translation to DaVinci syntax
   %%-----------------
   fun{List2DaVinci Xs}
      case Xs
      of nil then nil
      [] X|Xr then
	 {Oz2DaVinci X}
	 #if Xr==nil then
	     ""
	  else ", " end
	 #{List2DaVinci Xr}
      end
   end   

   fun{Oz2DaVinci Term}
      %{Show oz2davinci#Term}
      if {IsAtom Term} then
	 Term 
      elseif {IsInt Term} then    %% just print out integers
	 Term
      elseif {IsFloat Term} then  %% try to convert floats into integers (e.g. 2.0 -> 2)
	 Term
      else
	 case Term
	 of nil then ""
	 [] Name#const then   %% constants are just printed
	    Name
	 [] Name#var then     %% problem vars are just printed.
	    Name
	 else if {IsRecord Term} then
		 "("
		 #{Oz2DaVinci Term.1}
		 #" "
		 #{Label Term}
		 #" "
		 #{Oz2DaVinci Term.2}
		 #")"
	      else Term end
	 end
      end
   end

   %%-----------------
   %%  Translation to POST syntax
   %%-----------------

   fun{AList2Post AList FuncDict}
      case AList
      of nil then nil
      [] [Car#Cdr] then
	 "("
	 #{Oz2Post Car false FuncDict}
	 #" . "#
	 {Oz2Post Cdr false FuncDict}
	 #")"
      [] (Car#Cdr)|Lr then
	 "("
	 #{Oz2Post Car false FuncDict}
	 #" . "#
	 {Oz2Post Cdr false FuncDict}
	 #") "#{AList2Post Lr FuncDict}
      else nil end
   end
   
   fun{List2PostConjunct Cons Types FuncDict}
      case Cons
      of nil then "TRUE"
      [] C|Cr then 
	 if Cr == nil then
	    {Oz2Post C Types FuncDict}
	 else
	    "(AND "
	    #{Oz2Post C Types FuncDict}
	    #" "
	    #{List2PostConjunct Cr Types FuncDict}
	    #")"
	 end
      else nil end
   end

   fun{List2Post Xs Types FuncDict}
      case Xs
      of nil then "NIL"
      [] _|_ then
	 "("
	 #{RecList2Post Xs Types FuncDict}
	 #")"
      end
   end
   
   fun{RecList2Post Term Types FuncDict}
      case Term
      of nil then nil
      [] T|Tr then
	 {Oz2Post T Types FuncDict}
	 #if Tr == nil then ""
	  else 
	     " "
	     #{RecList2Post Tr Types FuncDict}
	  end
      else Term end
   end
   
   fun{RecTrace2Post Trace FuncDict}
      case Trace
      of nil then nil
      [] T|Tr then
	 {Entry2Post T  FuncDict}
	 #if Tr == nil then ""
	  else 
	     " "
	     #{RecTrace2Post Tr FuncDict}
	  end
      else Trace end
   end
      

   fun{Entry2Post Entry FuncDict}
      if {IsRecord Entry} then
	 "("
	 #{Label Entry}
	 #" "
	 #{List2Post if {HasFeature Entry lhs} then
			Entry.lhs else nil end
	   false FuncDict}
	 #" "
	 #{List2Post if {HasFeature Entry rhs} then
			Entry.rhs else nil end
	   false FuncDict}
	 #" "
	 #{List2Post if {HasFeature Entry args} then
			Entry.args else nil end
	   false FuncDict}
	 #")"
      else "" end
   end

   fun{Trace2Post Trace FuncDict}
      case Trace
      of nil then "NIL"
      [] _|_ then
	 "("
	 #{RecTrace2Post Trace FuncDict}
	 #")"
      end
   end
   
   fun{Oz2Post Term Types FuncDict}
      %% a record tree representing a CoSIE term and a boolean 'Types'
      %% returns: A virtual string representing the term in POST syntax
      %% (with types, if Types==true).
      
      %{Show oz2post#Term}
      if {IsAtom Term} then
	 if Types then
	    "["#Term#" d]"
	 else Term end
	 
      elseif {IsInt Term} then    %% just print out integers
	 if Types then
	    "["#Term#" n]"
	 else Term end
      elseif {IsFloat Term} then  %% try to convert floats into integers (e.g. 2.0 -> 2)
	 Num = if ({Ceil Term} == Term) then
		  {FloatToInt Term}
	       else Term end
      in
	 if Types then
	    "["#Num#" n]"
	 else Num end
      else
	 case Term
	 of nil then ""
	 [] _|_ then         %% handle lists of terms
	    {RecList2Post Term Types FuncDict}
	 [] Name#const then   %% constants are just printed
	    if Types then
	       "["#Name#" c]"
	    else Name end
	 [] Name#var then     %% problem vars are just printed.
	    if Types then
	       "["#Name#" v]"
	    else Name end
	 [] '->'(Lhs Rhs) then
	    if Types then
	       "([IMPLIES r] "
	       #{List2PostConjunct Lhs Types FuncDict}
	       #" "
	       #{List2PostConjunct Rhs Types FuncDict}
	       #")"
	    else "(IMPLIES "
	       #{List2PostConjunct Lhs Types FuncDict}
	       #" "
	       #{List2PostConjunct Rhs Types FuncDict}
	       #")"
	    end
	 else if {IsRecord Term} then
		 Func = {Label Term}
		 PostFunc = if {Dictionary.member FuncDict Func} then
			       {Dictionary.get FuncDict Func}
			    else Func end
	      in
		 if Types then
		    "(["#PostFunc#" f]"
		    #" "
		    #{Oz2Post {Record.toList Term} Types FuncDict}
		    #")"
		 else
		    "("#PostFunc
		    #" "
		    #{Oz2Post {Record.toList Term} Types FuncDict}
		    #")"
		 end
	      else
		 if Types then
		    "["#Term#" d]"
		 else Term end
	      end
	 end
      end
   end

   
   
end  % define
   /*

   fun{MakeHyps Hyps}
      case Hyps
      of nil then "TRUE"
      [] [Name#Formula] then
	 {List2PostConjunct Formula}
      [] [Name] then
	 Name
      [] Name#Formula|Hr then
	 "(AND "#{List2PostConjunct Formula}#" "#{MakeHyps Hr}#")"
      [] Name|Hr then
	 "(AND "#Name#" "#{MakeHyps Hr}#")"	 
      else Hyps end
   end

   fun{GetSolutions Vars Props AllProps AlienReps}
      case Vars
      of nil then "T"
      [] Name#var|Vr then
	 %{Show 'Var1'#Name}
	 case Props
	 of nil then nil
	 [] P|Pr then
	    if {Label P} == '=' then
	       case P.1
	       of VarName#Type then
		  if (Type == var orelse Type == nested)
		     andthen (Name == VarName)
		  then
		     RestProps = {RemoveProps AllProps [P]}
		  in
		     if Vr == nil then
			"(= "
			#Name
			#" "
			#{ConstructTerm P.2 RestProps RestProps AlienReps false}
			#")"
		     else 
			"(AND (= "#Name#" "#{ConstructTerm P.2 RestProps RestProps AlienReps false}#") "
			#{GetSolutions Vr RestProps RestProps AlienReps}#")"
		     end
		  elsecase P.2
		  of VarName2#Type2 then
		     if (Type2 == var orelse Type == nested)
			andthen (Name == VarName2)
		     then
			RestProps = {RemoveProps AllProps [P]}
		     in
			if Vr == nil then
			   "(= "#Name#" "#{ConstructTerm P.1 RestProps RestProps AlienReps false}#")"
			else 
			   "(AND (= "#Name#" "#{ConstructTerm P.1 RestProps RestProps AlienReps false}#") "
			   #{GetSolutions Vr RestProps RestProps AlienReps}#")"
			end
		     else
			{GetSolutions Vars Pr AllProps AlienReps}
		     end
		  else
		     {GetSolutions Vars Pr AllProps AlienReps}
		  end
	       else
		  {GetSolutions Vars Pr AllProps AlienReps}
	       end
	    else
	       {GetSolutions Vars Pr AllProps AlienReps}
	    end
	 else nil end
      else 'T' end
   end

    fun{List2PostConjunct Trace}
      case Trace
      of nil then "TRUE"
      [] [tell#NodeName#T] then
	 {List2PostConjunct T}
      [] [tell#T] then
	 {List2PostConjunct T}
	 
      [] tell#NodeName#T|Tr then
	 "(AND "#{List2PostConjunct T}#" "#{List2PostConjunct Tr}#")"
      [] tell#T|Tr then
	 "(AND "#{List2PostConjunct T}#" "#{List2PostConjunct Tr}#")"
      [] ask#T|Tr then
	 {List2PostConjunct Tr}
      [] ass#NodeName#Con|Tr then
	 {List2PostConjunct Tr}
      [] T|Tr then
	 "("#{List2PostConjunct Trace}#")"
      [] Name#Type then
	 {List2PostConjunct Name}
      else if {IsAtom Trace} then
	      {AtomToString Trace}
	   elseif {IsRecord Trace} then
	      Func = case {Label Trace}
		     of '<' then 'LESS'
		     [] '<=' then 'LEQ'
		     [] '>' then 'GREATER'
		     [] '>=' then 'GEQ'
		     [] '+' then 'PLUS'
		     [] '-' then 'MINUS'
		     [] '*' then 'TIMES'
		     [] '/' then 'DIV'
		     [] 'min' then 'MIN'
		     [] '||' then 'ABSVAL'
		     else {Label Trace} end
	   in
	      "("#Func#" "#{List2PostConjunct {Record.toList Trace}}#")"
	   else
	      if {IsNumber Trace} then
		 Float = if {IsFloat Trace} then
			    Trace
			 else {IntToFloat Trace} end
	      in
		 if (Float < 0.0) then
		    "-"#{Abs Float}
		 else Float end
	      else Trace end
	   end
      end
   end

    fun{PermuteTerm Term T Props AlienReps} 
      Func = {Label Term}
   in
      %{Show permute#Term#T}
      if {Member Func ['+' '-' '*' '/' 'MIN']} then
	 if Term.3 == T then
	    if Types then
	       "(["#case Func
		    of '+' then "PLUS"
		    [] '-' then "MINUS"
		    [] '*' then "TIMES"
		    [] '/' then "DIV"
		    [] 'MIN' then "MIN"
		    else "PLUS" end
	       #" f] "
	       #{ConstructTerm Term.1 Props Props AlienReps}
	       #" "
	       #{ConstructTerm Term.2 Props Props AlienReps}
	       #")"
	    else
	       "("#case Func
		   of '+' then "PLUS"
		   [] '-' then "MINUS"
		   [] '*' then "TIMES"
		   [] '/' then "DIV"
		   [] 'MIN' then "MIN"
		   else "PLUS" end
	       #" "
	       #{ConstructTerm Term.1 Props Props AlienReps}
	       #" "
	       #{ConstructTerm Term.2 Props Props AlienReps}
	       #")"
	    end
	 else nil end
      elseif Func == '||' then
	 if Term.2 == T then
	    if Types then 
	       "([ABSVAL f] "
	       #{ConstructTerm Term.1 Props Props AlienReps}
	       #")"
	    else
	       "(ABSVAL "
	       #{ConstructTerm Term.1 Props Props AlienReps}
	       #")"
	    end
	 else nil end
      else if Types then
	      "(["#Func#" f] "
	      #{ConstructTerm {Record.toList Term} Props Props AlienReps}
	      #")"
	   else
	      "("#Func
	      #{ConstructTerm {Record.toList Term} Props Props AlienReps}
	      #")"
	   end
      end %nil end
   end

       fun{List2PostConjunct List}
      case List
       of nil then nil
      [] [L] then
	 {List2PostConjunct L}
      [] L|Lr then
	 {List2PostConjunct L}#" "#{List2PostConjunct Lr}
      else nil end
   end
   

       fun{ConstructTerm T Props AllProps AlienReps Types}
      %{Show construct#T}
      if {IsAtom T} then {Atom.toString T}
      elseif {IsNumber T} then
	 if Types then
	    "["#T#" n]"
	 else T end
      else
	 case T
	 of nil then ""
	 [] H|Tr then         %% handle lists of terms
	    {ConstructTerm H Props AllProps AlienReps Types}
	    #if Tr == nil then ""
	     else 
		" "
		#{ConstructTerm Tr Props AllProps AlienReps Types}
	     end
	 [] Name#const then   %% constants are just printed
	    if Types then
	       "["#Name#" c]"
	    else Name end
	 [] Name#var then     %% problem vars are just printed.
	    if Types then
	       "["#Name#" v]"
	    else Name end
	 [] Name#nested then  %% nested vars have to be replaced by the right term.
	    case Props
	    of nil then {Show 'worst case 1'} T
	    [] P|Pr then
		  %{Show matchpropa#T#P.function|P.args}
	       if {Member {Label P} CosieFuncs} then
		  RealTerm = {PermuteTerm {Record.subtract P ref}
			      T
			      {RemoveProps AllProps [P]}
			      AlienReps
			      Types}
	       in
		  if RealTerm == nil then
		     {ConstructTerm T Pr AllProps AlienReps Types}
		  else
		     RealTerm
		  end
	       else  {ConstructTerm T Pr AllProps AlienReps Types}
	       end
	    else T end
	 [] Name#alien then  %% alien vars must be replaced by the alien term they stand for.
	    Replace = {Dictionary.get AlienReps Name}
	 in
	    %{Show replace#Name#Replace}
	    {ConstructTerm Replace Props AllProps AlienReps Types}
	 else if {IsRecord T} then
		 if Types then
		    "(["#{Label T}#" f]"
		    #" "
		    #{ConstructTerm {Record.toList T} Props AllProps AlienReps Types}
		    #")"
		 else
		    "("#{Label T}
		    #{ConstructTerm {Record.toList T} Props AllProps AlienReps Types}
		    #")"
		 end
	      else T end
	 end
      end
   end
*/
